/*************************************************************************************
    char_set_mgr.hxx
    header file for the object definitions used for character set
      management

    (c) Copyright Apollo Computer Inc., 1989 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

/**********************************************************************************
    version number of this version of the character set manager
**********************************************************************************/

// 1: base functionality

const long csm_version_number = 1;

#ifndef char_set_mgr_hxx
#define char_set_mgr_hxx


#define CHAR_SET_ALIGN_BLOCK 4   /* modulus in bytes of universal alignment boundary */

class tml_type;

#ifndef csm_character_hxx
#include <tml/csm_character.hxx>
#endif

#ifndef csm_lock_hxx
#include <tml/csm_lock.hxx>
#endif

#ifndef tml_d_ptr_hxx
#include <tml/tml_d_ptr.hxx>
#endif

#ifndef endian_hxx
#include <tml/endian.hxx>
#endif

#ifndef tml_gc_hash_hxx
#include <tml/tml_gc_hash.hxx>
#endif

#ifndef csm_kmp_hxx
#include <tml/csm_kmp.hxx>
#endif

#ifndef tml_type_hxx
#include <tml/tml_type.hxx>
#endif

#ifndef char_set_x_registry_hxx
#include <tml/char_set_x_registry.hxx>
#endif

#ifndef tml_os_depends_hxx
#include <tml/tml_os_depends.hxx>
#endif

/**************************************************************************************
    bogus type to allow passing of appropriate function pointers to tml_os_qsort
***************************************************************************************/

typedef long (*SORT_FUNC)(void*, void*);

/**********************************************************************************
    Preloaded character set types
**********************************************************************************/

extern tml_type* const tml_char_set;
extern tml_type* const tml_alpha_char_set;
extern tml_type* const tml_syll_frag_char_set;

typedef tml_type* cs_type;

/**********************************************************************************
    Enumerated and defined constants
**********************************************************************************/

/*enumerated class of directions for text rendering*/
/* c?_l?_?== <direction_to_next_character>_  ? -> { r==right, l==left, u==up, d==down };            */
/*           <direction_to_next_line>_       ? -> { r==right, l==left, u==up, d==down };            */
/*           <direction_of_next_line>        ? -> { i==identical, b==bustrophodon (alternating) };  */
/* In bustrophodon the c? determines direction of first line                                        */
enum char_set_dir  { csd_undefined = -1,
                     cl_ll_i, cl_lr_i, cl_lu_i, cl_ld_i, cl_ll_b, cl_lr_b, cl_lu_b, cl_ld_b,
                     cr_ll_i, cr_lr_i, cr_lu_i, cr_ld_i, cr_ll_b, cr_lr_b, cr_lu_b, cr_ld_b, 
                     cu_ll_i, cu_lr_i, cu_lu_i, cu_ld_i, cu_ll_b, cu_lr_b, cu_lu_b, cu_ld_b, 
                     cd_ll_i, cd_lr_i, cd_lu_i, cd_ld_i, cd_ll_b, cd_lr_b, cd_lu_b, cd_ld_b };


enum lrud_dir      { lrud_undefined = -1,
                     left, right, up, down };

/*enumerated class of C Language escape characters */
enum esc_char {esc_alert,        esc_backspace,       esc_formfeed,
               esc_newline,      esc_carriage_return, esc_horizontal_tab,
               esc_vertical_tab, esc_backslash,       esc_question_mark,
               esc_single_quote, esc_double_quote,    num_esc_chars};

const char_set_heap_incr = 1024;

/**********************************************************************************
    Objects defined elsewhere
**********************************************************************************/

class csm_character;
class csm_string;
class xregister;

/**********************************************************************************
    Objects defined in this file
**********************************************************************************/

class char_range;
class char_range_set;
class collation_map;
class char_set;
class alpha_char_set;
class syll_frag_char_set;
class union_char_set;
class char_set_heap;
class cs_heap;
class d_csm_character;
class d_char_range_set;
class d_kmp_node;
class d_collation_map;
class d_char_set;
class d_syll_frag_char_set;
class d_union_char_set;
class d_union_header;

/**********************************************************************************
    Objects used to define and manage character sets
**********************************************************************************/

class char_range {
    csm_character high;
    csm_character low;

    friend class char_range_set;
    friend class collation_map;
    friend class char_set;
    friend int col_map_cmp_for_qsort(char_range* r1, char_range* r2);
public:
    char_range();
    // lo and hi must be terminated
    char_range(unsigned char* lo, unsigned char* hi, char_set* for_char_set,
               unsigned long lo_map = MAX_CHARS_IN_SET, unsigned long hi_map = MAX_CHARS_IN_SET);
    // lo and hi need not be terminated
    // lo_len, hi_len are length in CHARACTERS with respect to for_char_set
    char_range(unsigned char* lo, unsigned long lo_len,
               unsigned char* hi, unsigned long hi_len,
               char_set* for_char_set,
               unsigned long lo_map = MAX_CHARS_IN_SET,
               unsigned long hi_map = MAX_CHARS_IN_SET);

    csm_character* get_high(){return &high;}
    csm_character* get_low(){return &low;}
    unsigned long get_high_index(){return high.collation;}
    unsigned long get_low_index(){return low.collation;}
    unsigned long get_size(); // returns number of characters in range
    // candidate must be terminated
    long          contains(unsigned char* candidate)
                          {return(((high >= candidate) &&
                                   (low  <= candidate))
                                    ? -1 : 0);}
    long          contains(unsigned long collation)
                          {return(((high.collation >= collation) &&
                                   (low.collation  <= collation))
                                    ? -1 : 0);}
    // candidate need not be terminated
    // length in CHARACTERS with respect to in_char_set of high and low
    long          contains(unsigned char* candidate, unsigned long len)
                          {return(((high.ge(candidate, len)) &&
                                   (low.le(candidate, len))
                                    ? -1 : 0));}
    void          save(cs_heap* heaps, gc_hash_tbl& hash_tbl);
    void          restore(cs_heap* heaps, gc_hash_tbl& hash_tbl);
};

class char_range_set {
    char_range*   ranges;
    unsigned long num_ranges;

    friend class alpha_char_set;
    friend class char_set;
public:
    char_range_set();
    ~char_range_set();
    long contain(unsigned char* candidate);
    char_range*    get_range(unsigned long i){return ranges + i;}
    char_range*    get_ranges(){return ranges;}
    unsigned long  get_num_ranges(){return num_ranges;}
    void           make_ranges(long new_num){ranges = new char_range[new_num]; num_ranges = new_num;}
    void           init_range(unsigned long slot,
                              unsigned char* lo, unsigned char* hi,
                              char_set* for_char_set,
                              unsigned long spec_size = 1,
                              unsigned long lo_map = MAX_CHARS_IN_SET, unsigned long hi_map = MAX_CHARS_IN_SET);
    void           set_range(char_range* from_range, char_set* for_char_set, unsigned long slot);
    // lo and hi must be terminated
    void           set_range(unsigned char* lo, unsigned char* hi,
                             char_set* for_char_set,
                             unsigned long slot,
                             unsigned long lo_map = MAX_CHARS_IN_SET, unsigned long hi_map = MAX_CHARS_IN_SET);
    // lo and hi need not be terminated
    void           set_range(unsigned char* lo, unsigned long lo_len,
                             unsigned char* hi, unsigned long hi_len,
                             char_set* for_char_set,
                             unsigned long slot,
                             unsigned long lo_map = MAX_CHARS_IN_SET, unsigned long hi_map = MAX_CHARS_IN_SET);
    unsigned long  lookup(unsigned char* sought);
    unsigned char* lookup(unsigned long collation, unsigned char* c);
    unsigned char* corresponding(char_range_set& other_set, unsigned char* c);
    void           save(cs_heap* heaps, gc_hash_tbl& hash_tbl);
    void           restore(cs_heap* heaps, gc_hash_tbl& hash_tbl);
};

class collation_map {
    kmp_node*      substitute;  // for POSIX LC_COLLATE "substitute" keyword
    kmp_node*      many_to_one; // for e.g. Spanish "ch"
    char_set*      in_char_set;
public:
    char_range_set primaries;   // ordinary collation
    char_range_set secondaries; // for e.g. French <e> and <e-acute>
    char_range_set one_to_many; // for e.g. German <sharp s>

    collation_map();
    collation_map(char_range* sequences, unsigned long num_sequences, char_set* for_char_set, unsigned long start_at = 0);
    ~collation_map();
    kmp_node*     get_many_to_one(){return many_to_one;}
    // the following three functions may be called only for maps created by constructor with arguments
    void          make_secondaries(char_range* sequences, unsigned long num_sequences);
    void          make_one_to_many(char_range* sequences, unsigned long num_sequences);
    void          make_substitutes(char** find, char** sub, unsigned long num_subs);
    // returns length of single- or multicharacter collation sequence beginning at sought; 
    //   advances sought; reports primary collation only
    unsigned long lookup(unsigned char*& sought);
    // both strings must be terminated - like strcmp()
    // -- performs primary/secondary/many_to_one/one_to_many collation as needed
    long          compare(unsigned char* s1, unsigned char* s2);
    // neither string need be terminated - like strncmp()
    // -- performs primary/secondary/many_to_one/one-to-many collation as needed
    long          ncompare(unsigned char* s1, unsigned char* s2, unsigned long n);
    void          save(cs_heap* heaps, gc_hash_tbl& hash_tbl);
    void          restore(cs_heap* heaps, gc_hash_tbl& hash_tbl);
};

extern char* tml_os_byte_copy(char* from, char* to, long count);

class char_set {
protected:
    cs_type         char_set_type;    // alphabet, syllable_fragments, syllabary, ideograms
    unsigned char*  char_set_name;    // name of this character set
    xregister*      xreg;             // pointer to X Consortium designation
    unsigned long   char_size;        // number of bytes in a character in this character set
    unsigned long   num_chars;        // number of characters in this character set - maximum MAX_CHARS_IN_SET
    unsigned long   ref_count;        // number of references to this char_set in the work space
    char_set*       next;             // hook for chaining char_set objects
    collation_map*  collate_map;      // map from characters to collation sequence
    char_set_dir    predom_direc;     // predominant direction - one of char_set_dir
    lrud_dir        char_direc;       // if cs_undefined, use direction ranges to determine direction per character - one of lrud_dir
    char_range_set  direction_ranges; // text formatting/rendering directions 
    char_range_set  legal_ranges;     // range(s) of characters instantiated in this set
    char_range_set  alpha_ranges;     // range(s) of characters true for isalpha()
    char_range_set  digit_ranges;     // range(s) of characters true for isdigit()
    char_range_set  xdigit_ranges;    // range(s) of characters true for isxdigit()
    char_range_set  space_ranges;     // range(s) of characters true for isspace()
    char_range_set  punct_ranges;     // range(s) of characters true for ispunct()
    char_range_set  print_ranges;     // range(s) of characters true for isprint()
    char_range_set  graph_ranges;     // range(s) of characters true for isgraph()
    char_range_set  cntrl_ranges;     // range(s) of characters true for iscntrl()
    unsigned long   null_term;        // 1 (default) if null terminator is char_size zero bytes, 0 otherwise
    unsigned long   standard;         // 1 (default) if (a) char_size==1, (b) null terminated, (c) standard UNIX escape characters
                                      // NOTE: must be set to 0 for nonstandard character sets or results are unpredictible
    unsigned long   hibit_clear;      // For 1-byte character sets, 1 if 8th bit not used; 0 otherwise
    csm_character   null_char;        // string terminating character
    csm_character   escape_chars[num_esc_chars]; // C language escape characters
    csm_character*  regexp_chars;                // alternative regular expression characters
    friend class collation_map;
    friend class union_char_set;
public:

    // Public function-pointer members

    // For your convenience, please use the wrapper functions below. These functions take a character
    //   set argument. The wrapper functions supply it automatically. There should be no performance
    //   differences between the two sets of calls because the wrapper functions are inline.

    // Pointers to equivalents of C library string-handling functions for terminated strings of CHARACTERS.
    //   No checks for receiver overflow. See <string.h>.
    // NOTE: Default assignment will be appropriate for 1 <= char_size <= 4, with two-byte characters
    //   assumed to be aligned at 16-bit boundaries, three-byte characters unaligned, and four-byte
    //   characters aligned at 32-bit boundaries. For char_size > 4 or different alignment assumptions
    //   user must assign appropriate function pointers to these members. For variable-size characters,
    //   use class csm_character.

    unsigned char* (*strcat_p) (unsigned char* s1, unsigned char* s2, char_set* for_char_set = 0);         // append s2 including terminator to s1, return s1
    unsigned char* (*strncat_p)(unsigned char* s1, unsigned char* s2, long n, char_set* for_char_set = 0); // append up to n CHARACTERS of s2 to s1, return result
    long           (*strcmp_p) (unsigned char* s1, unsigned char* s2, char_set* for_char_set = 0);         // compare lexicographic values of CHARACTERS in two strings
    long           (*strncmp_p)(unsigned char* s1, unsigned char* s2, long n, char_set* for_char_set = 0); // compare lexicographic values of up to n CHARACTERS in two strings
    unsigned char* (*strcpy_p) (unsigned char* s1, unsigned char* s2, char_set* for_char_set = 0);         // copy s2 to s1 including null terminator, return s1
    unsigned char* (*strncpy_p)(unsigned char* s1, unsigned char* s2, long n, char_set* for_char_set = 0); // copy n CHARACTERS of s2 to s1, return s1
    long           (*strlen_p) (unsigned char*, char_set* for_char_set = 0);                               // return number of CHARACTERS in a string

    // Pointers to equivalents of C language index() and rindex(). Note type of second argument (char*)
    //   differs from type of corresponding argument in C language functions (char). See <string.h>.

    unsigned char* (*index_p)  (unsigned char* s, unsigned char* c, char_set* for_char_set);           // return pointer to first instance of c in s or 0 if not there
    unsigned char* (*rindex_p) (unsigned char* s, unsigned char* c, char_set* for_char_set);           // return pointer to last instance of c in s or 0 if not there

    // Member functions

    char_set(unsigned long per_char_size = 1);
    ~char_set();
    virtual char_set* make_another(unsigned long c_size){return new char_set(c_size);}
    unsigned long     is_std(){return standard;}

    // Wrappers calling function-pointer members

    unsigned char* cs_strcat (unsigned char* s1, unsigned char* s2)        {return (*strcat_p) (s1, s2, this);}
    unsigned char* cs_strncat(unsigned char* s1, unsigned char* s2, long n){return (*strncat_p)(s1, s2, n, this);}
    long           cs_strcmp (unsigned char* s1, unsigned char* s2)        {return (*strcmp_p) (s1, s2, this);}
    long           cs_strncmp(unsigned char* s1, unsigned char* s2, long n){return (*strncmp_p)(s1, s2, n, this);}
    unsigned char* cs_strcpy (unsigned char* s1, unsigned char* s2)        {return (*strcpy_p) (s1, s2, this);}
    unsigned char* cs_strncpy(unsigned char* s1, unsigned char* s2, long n){return (*strncpy_p)(s1, s2, n, this);}
    long           cs_strlen (unsigned char* s)                            {return (*strlen_p) (s, this);}
    unsigned char* cs_index  (unsigned char* s, unsigned char* c)          {return (*index_p)  (s, c, this);}
    unsigned char* cs_rindex (unsigned char* s, unsigned char* c)          {return (*rindex_p) (s, c, this);}

    // Functions to get data
    cs_type         get_type(){return char_set_type;}
    unsigned char*  get_name(){return char_set_name;}
    xregister*      get_xreg(){return xreg;}
    unsigned long   get_char_size(){return char_size;}
    unsigned long   get_num_chars(){return num_chars;}
    unsigned long   get_ref_count(){return ref_count;}
    char_set_dir    get_predom_direc(){return predom_direc;}
    lrud_dir        get_dir_to_char(unsigned char* c = 0);
    lrud_dir        get_dir_to_line();
    lrud_dir        get_dir_of_line(lrud_dir dir = lrud_undefined); // default gives direction of first line
    char_set*       get_next(){return next;}
    collation_map*  get_collate_map(){return collate_map;}
    char_range_set* get_legal_ranges() {return &legal_ranges;}
    char_range_set* get_alpha_ranges() {return &alpha_ranges;}
    char_range_set* get_digit_ranges() {return &digit_ranges;}
    char_range_set* get_xdigit_ranges(){return &xdigit_ranges;}
    char_range_set* get_space_ranges() {return &space_ranges;}
    char_range_set* get_punct_ranges() {return &punct_ranges;}
    char_range_set* get_print_ranges() {return &print_ranges;}
    char_range_set* get_graph_ranges() {return &graph_ranges;}
    char_range_set* get_cntrl_ranges() {return &cntrl_ranges;}
    void            get_terminator(unsigned char* c){tml_os_byte_copy((char*)null_char.value, (char*) c, char_size);} // write string-terminating character into c
    unsigned long   zero_term(){return null_term;}
    unsigned long   hibit_is_clear(){return hibit_clear;}

    // Functions to change data
    void set_type(cs_type new_cs_type){char_set_type = new_cs_type;}
    void set_name(unsigned char* new_name){char_set_name = new_name;}
    void set_xreg(xregister* new_xreg){xreg = new_xreg;}
    void set_char_size(unsigned long new_char_size){char_size = new_char_size;}
    void set_num_chars(unsigned long new_num_chars){num_chars = new_num_chars;}
    void set_ref_count(unsigned long new_ref_count){ref_count = new_ref_count;}
    void incr_ref_count(){++ref_count;}
    void decr_ref_count(){--ref_count;}
    void set_predom_direc(char_set_dir new_direc){predom_direc = new_direc;}
    void set_char_direc(lrud_dir new_direc){char_direc = new_direc;}
    void set_next(char_set* new_next){next = new_next;}
    void set_map(collation_map* new_map){collate_map = new_map;}
    void set_null(unsigned char* new_null);
    void init_esc_chars(unsigned char** esc_chars, long spec_size = 1);
    char_set* prepend(char_set* new_sets);
    char_set* append(char_set* new_sets);
    char_set* detach(char_set* from_vec);

    // return nonzero if character is the terminator, zero otherwise
    long is_null(unsigned char* c);
     // return nonzero if both characters are terminators, zero otherwise
    long both_null(unsigned char* c1, unsigned char* c2){return is_null(c1) && is_null(c2);}

    // NOTE: For comparisons by COLLATING CHARACTERS of strings in different
    //   character sets, use general functions
    //      long char_arr_cmp(csm_character* cc1, csm_character* cc2)
    //      long char_arr_ncmp(csm_character* cc1, csm_character* cc2, unsigned long n)
    //   For comparisons within character sets, use the following two member
    //     functions.

    // For comparisons by COLLATING CHARACTERS of strings in this character set
    // -- both strings must be terminated - like strcmp()
    long char_arr_cmp(unsigned char* s1, unsigned char* s2);
    // -- neither string need be terminated - like strncmp()
    long char_arr_ncmp(unsigned char* s1, unsigned char* s2, unsigned long n);

    // Sort strings, ignoring given prefixes; if reentrant set, use slower reentrant code
    void ignored_prefix_sort(char** strings, unsigned long num_strings, 
                             char** prefixes, unsigned long num_prefixes);

    // if char_size <= 4 bytes, return character right-justified in a long
    // else return MAX_CHARS_IN_SET
    unsigned long as_ulong(unsigned char* s);

    virtual void         advance(unsigned char*& c){c += char_size;} // advance to next CHARACTER start
    virtual void         retreat(unsigned char*& c){c -= char_size;} // advance to previous CHARACTER start

    // character at c is instantiated in this set
    virtual long is_char  (unsigned char* c){return legal_ranges.contain(c);}

    // Equivalents of C language  "is..." functions. Note type of second argument (char*)
    //   differs from type of corresponding argument in C language functions (char).
    //   See <ctype.h>. Names are prefixed with "cs_" to prevent macro expansion
    //   by simultaneous inclusion of <ctype.h>.

    virtual long cs_isalpha (unsigned char* c){return alpha_ranges.contain(c);}          // character at c is a letter/syllable fragment/syllable/ideogram
    virtual long cs_isdigit (unsigned char* c){return digit_ranges.contain(c);}          // character at c is a digit
    virtual long cs_isxdigit(unsigned char* c){return xdigit_ranges.contain(c);}         // character at c is a hexadecimal digit
    virtual long cs_isalnum (unsigned char* c){return cs_isalpha(c) || cs_isdigit(c);}   // character at c is an "alpha"numeric character
    virtual long cs_isspace (unsigned char* c){return space_ranges.contain(c);}          // character at c is a space, tab, carriage return, newline,
                                                                                //   vertical tab, or formfeed.
    virtual long cs_ispunct (unsigned char* c){return punct_ranges.contain(c);}          // character at c is a punctuation character (neither control
                                                                                //   nor alnum)
    virtual long cs_isprint (unsigned char* c){return print_ranges.contain(c);}          // character at c is a printing character (includes space)
    virtual long cs_isgraph (unsigned char* c){return graph_ranges.contain(c);}          // character at c is a printing character (excludes space)
    virtual long cs_iscntrl (unsigned char* c){return cntrl_ranges.contain(c);}          // character at c is a control character

    // Check for identity with escape characters (default to C language compatible values)

    virtual long is_alert    (unsigned char* c){return escape_chars[esc_alert          ].eq(c, char_size);}  // character at c is an alert (bell) character
    virtual long is_backspace(unsigned char* c){return escape_chars[esc_backspace      ].eq(c, char_size);}  // character at c is a backspace character
    virtual long is_formfeed (unsigned char* c){return escape_chars[esc_formfeed       ].eq(c, char_size);}  // character at c is a formfeed character
    virtual long is_newline  (unsigned char* c){return escape_chars[esc_newline        ].eq(c, char_size);}  // character at c is a newline character
    virtual long is_cr       (unsigned char* c){return escape_chars[esc_carriage_return].eq(c, char_size);}  // character at c is a carrriage return character
    virtual long is_htab     (unsigned char* c){return escape_chars[esc_horizontal_tab ].eq(c, char_size);}  // character at c is a horizontal tab character
    virtual long is_vtab     (unsigned char* c){return escape_chars[esc_vertical_tab   ].eq(c, char_size);}  // character at c is a vertical tab character
    virtual long is_backslash(unsigned char* c){return escape_chars[esc_backslash      ].eq(c, char_size);}  // character at c is a backslash character
    virtual long is_question (unsigned char* c){return escape_chars[esc_question_mark  ].eq(c, char_size);}  // character at c is a question mark character
    virtual long is_squote   (unsigned char* c){return escape_chars[esc_single_quote   ].eq(c, char_size);}  // character at c is a single quote character
    virtual long is_dquote   (unsigned char* c){return escape_chars[esc_double_quote   ].eq(c, char_size);}  // character at c is a double quote character
    virtual void save(cs_heap* heaps, gc_hash_tbl& hash_tbl, d_char_set* savee = 0);
    virtual void restore(cs_heap* heaps, gc_hash_tbl& hash_tbl, d_char_set* restoree = 0);

    // Get value of escape characters
    virtual void get_alert    (unsigned char* c){(*strcpy_p)(c, escape_chars[esc_alert].value,  this);}
    virtual void get_backspace(unsigned char* c){(*strcpy_p)(c, escape_chars[esc_backspace].value, this);}
    virtual void get_formfeed (unsigned char* c){(*strcpy_p)(c, escape_chars[esc_formfeed].value, this);}
    virtual void get_newline  (unsigned char* c){(*strcpy_p)(c, escape_chars[esc_newline].value, this);}
    virtual void get_cr       (unsigned char* c){(*strcpy_p)(c, escape_chars[esc_carriage_return].value, this);}
    virtual void get_htab     (unsigned char* c){(*strcpy_p)(c, escape_chars[esc_horizontal_tab].value, this);}
    virtual void get_vtab     (unsigned char* c){(*strcpy_p)(c, escape_chars[esc_vertical_tab].value, this);}
    virtual void get_backslash(unsigned char* c){(*strcpy_p)(c, escape_chars[esc_backslash].value, this);}
    virtual void get_question (unsigned char* c){(*strcpy_p)(c, escape_chars[esc_question_mark].value, this);}
    virtual void get_squote   (unsigned char* c){(*strcpy_p)(c, escape_chars[esc_single_quote].value, this);}
    virtual void get_dquote   (unsigned char* c){(*strcpy_p)(c, escape_chars[esc_double_quote].value, this);}
};

class alpha_char_set : public char_set {
    char_range_set  upper_ranges;     // range(s) of characters true for isupper();
    char_range_set  lower_ranges;     // range(s) of characters true for islower();

    friend union_char_set* get_dflt_union_char_set();
public:
    alpha_char_set(unsigned long per_char_size = 1);
    ~alpha_char_set();
    char_set* make_another(unsigned long c_size){return new alpha_char_set(c_size);}

    // Functions specific to alphabetic character sets. Note that to_lower() and
    //   to_upper return argument (char*) differ from corresponding C library
    //   return arguments (long)

    virtual long     cs_isupper(unsigned char* c){return upper_ranges.contain(c);}
    virtual long     cs_islower(unsigned char* c){return lower_ranges.contain(c);}
    virtual unsigned char* cs_tolower(unsigned char* c){return lower_ranges.corresponding(upper_ranges, c);}
    virtual unsigned char* cs_toupper(unsigned char* c){return upper_ranges.corresponding(lower_ranges, c);}

    void          become_iso_latin_1(); // become a character set equivalent to ISO Latin 1 at this character size
    void          become_ascii();       // become an ASCII-equivalent character set at this character size
    void          save(cs_heap* heaps, gc_hash_tbl& hash_tbl, d_char_set* savee = 0);
    void          restore(cs_heap* heaps, gc_hash_tbl& hash_tbl, d_char_set* restoree = 0);
};

// suitable e.g. for translating East-Indian and Korean
//   phonemic input to scripted syllables
class syll_frag_char_set : public char_set {
    collation_map*  syllable_map;    // collation is syllable character ID
                                     //   in output character set
    char_set*       output_char_set; // char_set of complete syllable IDs
public:
    syll_frag_char_set(unsigned long per_char_size = 1);
    ~syll_frag_char_set();
    char_set* make_another(unsigned long c_size){return new syll_frag_char_set(c_size);}

    // returns nonzero if a legal multichar sequence, 0 otherwise
    unsigned long is_syll(unsigned char* start);
    // returns syllable ID in output character set if is_syll(),
    //   MAX_CHARS_IN_SET otherwise, advances syllable
    unsigned long get_syll(unsigned char*& start);
    // returns length of syllable
    unsigned long get_syll_len(unsigned char* start);
    void          save(cs_heap* heaps, gc_hash_tbl& hash_tbl, d_char_set* savee = 0);
    void          restore(cs_heap* heaps, gc_hash_tbl& hash_tbl, d_char_set* restoree = 0);
};

class union_char_set {
    char_set*       char_sets;      // character sets that make up this union_char_set
    unsigned long   num_sets;       // number of character sets in this union_char_set
    union_char_set* next;           // for linking on model
    unsigned long   ref_count;      // reference counter for this union_char_set

    // create union_char_sets from binary file indicated by file handle, using rfp to read the file
    friend union_char_set* read_ucs_binary_file(void *file_handle, gc_hash_tbl*& hash_tbl, tml_ext_type*& types, endian*& their_endian);
    // create union_char_sets from human-readable file indicated by file handle, using rfp to read the file
    friend union_char_set* read_ucs_config_file(void *file_handle);
public:
    union_char_set();
    union_char_set(char_set* new_char_sets);
    ~union_char_set();
    char_set*       get_sets(){return char_sets;}
    unsigned long   get_num_sets(){return num_sets;}
    union_char_set* get_next(){return next;}
    unsigned long   get_ref_count(){return ref_count;}
    void            set_sets(char_set* new_sets){char_sets = new_sets;}
    void            set_num_sets(unsigned long new_num){num_sets = new_num;}
    void            set_next(union_char_set* new_next){next = new_next;}
    void            set_ref_count(unsigned long new_ref_count){ref_count = new_ref_count;}
    void            incr_ref_count(){++ref_count;}
    void            decr_ref_count(){--ref_count;}
    union_char_set* prepend(union_char_set* new_unions);
    union_char_set* append(union_char_set* new_unions);
    union_char_set* detach(union_char_set* from_vec);
    char_set*       prepend(char_set* new_sets);
    char_set*       append(char_set* new_sets);
    char_set*       detach(char_set* old_set);
    char_set*       contains(unsigned char* c, unsigned long char_size);
    long            contains(unsigned long char_set_index, unsigned char* c);
    // returns nonzero if this is a default union_char_set (ISO) with no "next"
    long            default_only();
    void            save(cs_heap* heaps, gc_hash_tbl& hash_tbl);
    void            restore(cs_heap* heaps, gc_hash_tbl& hash_tbl);
    // save in human-readable form to file indicated by file handle, using wfp to
    // write the file
    long            write_ucs_config_file(void *file_handle);
    // prepare for binary write - used by code that tracks character sets elsewhere
    // in structures that are also saved to disk - called by ::write_binary()
    // if not called previously
    void            prepare_ucs_binary(gc_hash_tbl*& hash_tbl, cs_heap*& heaps, tml_ext_type*& types);
    // save in binary form to file indicated by file handle, using wfp to write the file.
    // Calls ::prepare_binary() if argument heaps is NULL
    long            write_ucs_binary_file(void *file_handle, gc_hash_tbl*& hash_tbl, cs_heap*& heaps, tml_ext_type*& types);
};

/**********************************************************************************
    Classes for save/restore to/from disk
**********************************************************************************/

class d_csm_character {
    unsigned long num_chars;
    unsigned long collation;

    friend class csm_character;
public:
    void unscramble(endian* their_endian);
};

class d_char_range_set {
    unsigned long num_ranges;

    friend class char_range_set;
public:
    void unscramble(endian* their_endian);
};

class d_kmp_node {
    void* right;
    void* down;
    void* failure;

public:
    void  unscramble(endian* their_endian);
    void* get_down(){return down;}
    void* get_right(){return right;}
    void* get_failure(){return failure;}
    void  set_down(d_ptr *d_ptr_tmp);
    void  set_right(d_ptr *d_ptr_tmp);
    void  set_failure(d_ptr *d_ptr_tmp);
};

class d_collation_map {
    void* substitute;
    void* many_to_one;

    friend class collation_map;
public:
    void unscramble(endian* their_endian);
};

#define cs_null_term   1
#define cs_standard    2
#define cs_hibit_clear 4

class d_char_set {
    cs_type           char_set_type;
    unsigned long     name_chars;
    unsigned long     char_size;
    unsigned long     num_chars;
    void*             collate_map;
    char_set_dir      predom_direc;
    lrud_dir          char_direc;
    unsigned long     flags;
    void*             next;

    friend class char_set;
public:
    void          unscramble(endian* their_endian);
    void          set_type_index(char_set* from_set);
    void*         get_char_set_type(){return char_set_type;}
    unsigned long get_char_size(){return char_size;}
};

class d_syll_frag_char_set : public d_char_set {
    void* syllable_map;
    void* output_char_set;

    friend class syll_frag_char_set;
public:
    void  unscramble(endian* their_endian);
};

class d_union_char_set {
    unsigned long   num_sets;
    unsigned long   ref_count;
    void*           sets;
    void*           next;

    friend class union_char_set;
public:
    void unscramble(endian* their_endian);
};
                                  
class d_union_header {
    unsigned long  num_heaps;
    unsigned long  num_unions;
    unsigned long  num_maps;
    unsigned long  num_kmp_nodes;
    unsigned long  num_types;
    unsigned long  version_number;

    friend class kmp_node;
    friend class collation_map;
    friend class char_set;
    friend class alpha_char_set;
    friend class syll_frag_char_set;
    friend class union_char_set;
    friend class cs_heap;
    friend void kmp_tree_save(kmp_node* root, cs_heap* heaps, gc_hash_tbl& hash_tbl);
    friend void kmp_tree_restore(kmp_node*& root, cs_heap* heaps, gc_hash_tbl& hash_tbl);
    friend union_char_set* read_ucs_binary_file(void *file_handle, gc_hash_tbl*& hash_tbl, tml_ext_type*& types, endian*& their_endian);
public:
    d_union_header();
    unsigned long get_num_unions(){return num_unions;}
    unsigned long get_num_maps(){return num_maps;}
    unsigned long get_num_kmp_nodes(){return num_kmp_nodes;}
    void          set_num_unions(unsigned long num){num_unions = num;}
    void          set_num_maps(unsigned long num){num_maps = num;}
    void          set_num_kmp_nodes(unsigned long num){num_kmp_nodes = num;}
    void unscramble(endian* their_endian);
};

class char_set_heap {
    unsigned long  next_byte;
    unsigned long  bytes_used;
    unsigned long  size;
    unsigned char *heap_start;
    char_set_heap *next;

    unsigned char* obj_alloc(long num_bytes);
    unsigned char* obj_realloc(long num_bytes);
friend class cs_heap;
public:
    char_set_heap(long new_size = 0);
    ~char_set_heap();
    unsigned long  get_bytes_used(){return bytes_used;}
    char_set_heap* get_next(){return next;}
    unsigned char* get_start(){return heap_start;}
    void           set_bytes_used(unsigned long n){bytes_used = n;}
};

class cs_heap {
public:
    char_set_heap  *head_heap;
    unsigned long  *bytes_used;
    unsigned long   num_char_sets;
    char_set*       cur_char_set;
    endian         *their_endian;
    d_union_header  header;
    tml_ext_type   *types;

    // when creating cs_heap for reading a model,
    //   endian argument should be chosen by application
    //   based on order of bytes in magic number at head
    //   of file for cross-architecture compatibility
    // NULL new_endian means application doesn't write magic
    //   number or always assumes endian_3210. See
    //   arch_neutral.hxx, endian.hxx.
    cs_heap(endian* new_endian = 0);
    ~cs_heap();
    unsigned char* obj_alloc(long num_bytes);
    unsigned char* obj_realloc(long num_bytes);
    unsigned char* obj_noalloc();
    void  queue_heap(char_set_heap* new_heap);
    void load_types();
    char_set*      get_cur_char_set(){return cur_char_set;}
    void           set_cur_char_set(char_set* new_char_set){cur_char_set = new_char_set;}
};

/*************************************************************************************
    Declaration of general function to get default union_char_set
*************************************************************************************/

/* return default union char set, initializing global if necessary */
extern union_char_set* get_dflt_union_char_set();

/*************************************************************************************
    Declaration of general functions for global type queue
*************************************************************************************/

/* return head of type queue */
extern tml_type* csm_type_q_get_head();

/* queue a new type */
extern void      csm_type_q_queue_type(tml_type* new_type);

/* utility functions for save/restore mechanism */

tml_type*        csm_type_q_get_type(long index, tml_ext_type* types);
long             csm_type_q_get_index(tml_type* this_type);
void*            csm_type_q_get_instance(tml_type* this_type);

/*************************************************************************************
    Declarations of general functions to read character sets recorded
      in binary and human-readable format respectively
*************************************************************************************/

/* create union_char_sets from binary file indicated by file handle, using rfp to read the file */
extern union_char_set* read_union_char_sets(void *file_handle, gc_hash_tbl*& hash_tbl, tml_ext_type*& types, endian*& their_endian);

/* create union_char_sets from human-readable file indicated by file handle, using rfp to read the file */
extern union_char_set* read_config_file_union_char_sets(void *file_handle);

/*************************************************************************************
    Declarations of general functions for comparison of strings
      in two different character sets whose collation maps have been
      coordinated
*************************************************************************************/

extern long char_arr_cmp(csm_character* c1, csm_character* c2);
extern long char_arr_ncmp(csm_character* c1, csm_character* c2, unsigned long n);

/*************************************************************************************
    Declarations of some general functions for sorting
*************************************************************************************/

/* set or reset a character set pointer to be used for csm_cmp_for_qsort() */
extern void set_csm_char_set_for_qsort(char_set* new_char_set);

/* nonreentrant comparison function to be used in bound libraries */
/*   with UNIX qsort() when character sets are involved */
extern int csm_cmp_for_qsort(char** sp1, char** sp2);

/*************************************************************************************
    Declarations of general functions for preserving architecture neutrality
*************************************************************************************/

extern endian* get_my_endian();

extern endian* get_their_endian(unsigned long their_magic_number);

#endif
// char_set_mgr_hxx

