/*************************************************************************************
    csm_character.hxx
    header file for the object definition of the lowest-level
        object used by the character set manager

    (c) Copyright Apollo Computer Inc., 1989 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#ifndef csm_character_hxx
#define csm_character_hxx

/**********************************************************************************
    Objects defined elsewhere
**********************************************************************************/

class char_set;
class csm_string;
class posix_re;
class kmp_node;
class kmp_sequence_node;
class kmp_range_node;
class char_range;
class cs_heap;
class gc_hash_tbl;

/**********************************************************************************
    Declarations establishing independence from pointer/integer conversion
      peculiarities of heterogeneous compiler architectures
**********************************************************************************/

#ifdef ptr_32_int_16
#else
#ifdef ptr_16_int_16
#else
#ifdef ptr_16_int_32
#else                    // ptr_32_int_32
const long MAX_BYTES_IN_CHAR = 4294967295;  // maximum bytes in a character 
const long MAX_CHARS_IN_SET  = 4294967295;  // maximum characters in a character set and
#endif                                      //   sentinel value for illegal collation value
#endif                                      //   and illegal numeric value of character
#endif

/**********************************************************************************
    Objects defined in this file
**********************************************************************************/

class csm_character;

/**********************************************************************************
    Object used where necessary to represent a single character or sequence
      of characters for character-set-transparent processing
**********************************************************************************/

class csm_character {         // collating character (e.g. "a", "ch" in Spanish)
    unsigned char* value;     // null-terminated string
    unsigned long  num_chars; // length of value in CHARACTERS
    unsigned long  collation; // used for collation order of this text value
    char_set* in_char_set;    // character set associated with this character

    friend class char_range;
    friend class char_range_set;
    friend class collation_map;
    friend class char_set;

    friend long      cmp_csm_str(csm_string* cs1, csm_string* cs2);
    friend kmp_node* kmp_fsm(unsigned char**, unsigned long, unsigned long*, char_set* for_char_set);
    friend kmp_node* kmp_re_fsm(posix_re* expr);
    friend class kmp_node;
    friend class kmp_sequence_node;
    friend class kmp_range_node;
    friend int col_map_cmp_for_qsort(char_range* r1, char_range* r2);

public:
    // value = NULL; collation = MAX_CHARS_IN_SET; in_char_set = NULL
    csm_character();
    // initialized for a given character set
    csm_character(char_set* for_char_set);
    // val is null-terminated in given character set
    csm_character(unsigned char* val, char_set* for_char_set);
    // val need not be null-terminated; length in CHARACTERS for given char_set
    csm_character(unsigned char* val, unsigned long len, char_set* for_char_set);
    // initialize collation as well; length in CHARACTERS for given char_set
    csm_character(unsigned char* val, unsigned long len, unsigned long collation_order, char_set* for_char_set);
    ~csm_character();
    void           init(unsigned char* val, unsigned long len, char_set* for_char_set = 0, unsigned long index = MAX_CHARS_IN_SET);
    unsigned char* get_value(){return value;}
    unsigned long  get_num_chars(){return num_chars;}  //length in CHARACTERS
    unsigned char* get_byte_at(long byte_index){return value + byte_index;}
    unsigned long  get_collation(){return collation;}
    void           set_collation(unsigned long new_order){collation = new_order;}
    char_set*      get_char_set(){return in_char_set;}
    void           save(cs_heap* heaps, gc_hash_tbl& hash_tbl);
    void           restore(cs_heap* heaps, gc_hash_tbl& hash_tbl);

    // comp_val need not be terminated
    // len is length in CHARACTERS for size of characters in in_char_set
    // NOTE: in_char_set member must be non-NULL
    long eq (unsigned char* comp_val, unsigned long len);
    long ne (unsigned char* comp_val, unsigned long len);
    long lt (unsigned char* comp_val, unsigned long len);
    long le (unsigned char* comp_val, unsigned long len);
    long gt (unsigned char* comp_val, unsigned long len);
    long ge (unsigned char* comp_val, unsigned long len);

    // comp_val must be terminated with respect to in_char_set
    // NOTE: in_char_set member must be non-NULL
    long operator== (unsigned char* comp_val);
    long operator!= (unsigned char* comp_val);
    long operator<  (unsigned char* comp_val);
    long operator<= (unsigned char* comp_val);
    long operator>  (unsigned char* comp_val);
    long operator>= (unsigned char* comp_val);

    // compare string values of two csm_character objects
    long operator== (csm_character comp_char){return operator==(comp_char.value);}
    long operator!= (csm_character comp_char){return operator!=(comp_char.value);}
    long operator<  (csm_character comp_char){return operator< (comp_char.value);}
    long operator<= (csm_character comp_char){return operator<=(comp_char.value);}
    long operator>  (csm_character comp_char){return operator> (comp_char.value);}
    long operator>= (csm_character comp_char){return operator>=(comp_char.value);}
};

#endif 
//csm_character_hxx

