/**************************************************************************************
    File extend.cxx
    Utility functions and global objects for the extension mechanism

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

/**********************************************************************************
    NOTE: WHAT THE TYPES ARE

          These types are based on Apollo NCS uuid identifiers generated by
          the program uuid_gen. They are guaranteed to be unique for all times
          and places when compared with other uuid objects. The algorithm for
          generating these identifiers is freely available.

          HOW THE TYPES ARE USED

          The types are used principally in three ways. Since they are unique
          they can function as type identifiers where that is necessary. They
          play an important role in the save/restore mechanism. They are also
          essential to the extension mechanism.

          In the save/restore mechanism, typed objects that contain virtual
          function pointers are required to carry a type that identifies them.
          This type (the tml_ext_type associated with the object's type) is
          written to disk in an array of tml_ext_type objects. The type field
          in the memory format of an object is then replaced by the index of
          the appropriate tml_ext_type object in the array. When the file is
          restored, this index yields a pointer to the proper tml_ext_type in
          the restoring program.

          A global queue of tml_type objects exists in all TMLib programs. 
          (This file declares and initializes that queue, the global object
          type_q.) When a TMLib program writes a file to disk, it also writes
          an array of the tml_ext_type objects which are members of the tml_type
          objects on the queue. This permits user extensions: Users may queue
          their own tml_type objects on the global queue. TMLib programs that
          are built with appropriate code for saving and restoring these extended
          objects can then recognize extensions as they are encountered.
          
          For objects with virtual function pointer members the presence of a
          tml_type object on the queue is vital to the success of the save/restore
          mechanism. Since the value of a virtual function pointer is runtime-
          dependent, these values cannot be written to and read from disk with
          consistent results. Therefore, a virtual function pointer is saved as
          an index into the array of tml_ext_type objects. This index points out
          the tml_ext_type in the restored array that corresponds to the object
          type being restored. TMLib can then search its queue of tml_type objects
          to find an instance of the object being restored. TMLib then assigns
          the virtual function pointer member of that instance object to the virtual
          function pointer member of the restored object.

          The save/restore mechanism works through virtual functions called 
          save() and restore(). Each object that is to be saved to disk must define
          these functions.
          
          PROGRAMS THAT BIND IN THE CSM OR TMLib

          The bracket-list initializers apply only to builds where TMLib is to be
          bound into the program. They are ignored when TMLib is installed via
          the Apollo inlib facility, and they must not be present for Apollo
          global library builds.

          PROGRAMS THAT INLIB THE CSM OR TMLib (Apollo only)

          In the case of Apollo inlib'd libraries, global variables of types that
          have constructors are not initialized, because the libraries are not 
          bound in with the executables. An alternate entry point initializes 
          these variables when an inlib'd library loads. This entry point was
          specified in the linker command that built the library. It is called
          csm_inlib_entry() for the CSM and tml_inlib_entry() for TMLib. The
          alternate entry point makes explicit constructor calls to create
          analogues for the "sample" objects declared below.

          THE CSM AND TMLib AS GLOBAL LIBRARIES (Apollo only)

          No automatic alternate entry point is specified when Apollo global 
          libraries are built. Users of global libraries must call the initialization
          function before calling any other function that directly accesses this data. 

          When an inlib'd library loads, the procedures and bounds for data allocation 
          are already established, but a global library loads at boot time: The dynamic 
          allocators it might call and their working space do not yet exist. Hence
          the requirement for user initialization.

          Multiple calls to the initialization function are not necessary but they
          are not harmful. The initialization function names are given above.

          WHY TML_TYPE AND TML_EXT_TYPE HAVE ONLY PUBLIC MEMBERS

          For applications that bind the library, the initializers given here in the
          nonglobal case do take effect. Hence the data members of the type objects 
          (tml_type and tml_ext_type) are all public, so they can be initialized with 
          initializer lists. 

          WHERE ORDER OF DECLARATIONS IS IMPORTANT

          Since the initializer lists for objects of class tml_type must contain a 
          pointer to "next" type in order to be present on the global object type_q 
          of class type_queue, they must be declared in reverse order of their queuing. 
          This guarantees that the "next" member of an object of class tml_type will 
          be defined before that object is in turn defined.

          In other words, the order given here is required only to queue the objects
          properly given the values in the initializer lists. Otherwise the order 
          of declarations is not important except for those types mentioned below
          under the note on restoring TMLib Version 1 objects. (The version number
          is the value of tml_version_number, declared in tml_objects.hxx.)

**********************************************************************************/

/**********************************************************************************
    global queue of extensible types to which extended types are added by user
**********************************************************************************/

#ifdef apollo_global
type_queue type_q;
#else
#ifdef apollo
// this initialization has no effect for global or inlib'd builds
// for inlib'd builds, tml_inlib_entry initializes this variable
// for global builds, test for NULL head in type_queue::queue_type 
//   performs initialization
type_queue type_q __attribute((section(type_q))) = {&tml_char_set_t, &tml_syll_frag_char_set_t};
#else
type_queue type_q = {&tml_char_set_t, &tml_syll_frag_char_set_t};
#endif
#endif

    // New types must be queued at the end of the type queue
    //   (a) by the mechanism outlined in NOTE above 
    //   (b) by user calls to csm_queue_new_type() for user extended types.

    // For TML Version 2 restore of Version 1 objects to work properly,
    //   members bobj_index_v1 and gobj_index_v1 must be set as per
    //   above initialization. These correspond to the queue positions
    //   of the bobj and gobj types respectively at both TML Version 1
    //   and TML Version 2.
    // For TML Version 2 restore of Version 1 objects to work properly,
    //   member first_prop_index must be set to queue position of first
    //   property type in queue. Do not change either queue position of
    //   tml_wrap_prop or relative order of tml_wrap_prop through
    //   tml_pgsize_prop or Version 2 restore of Version 1 objects
    //   will not work.
    // In general, new types must be queued at the end of the type queue
    //   (a) by the mechanism outlined in NOTE above (b) by user calls
    //   to type_queue::queue_type() for user extended types.

void type_queue::queue_type(tml_type* new_type)
{
    if (NULL == head) {
        // global or inlib build -- not initialized
        head = &tml_char_set_t;
        tail = &tml_syll_frag_char_set_t;
        }
    tail->next = new_type;
    tail = new_type;
}

// Interface function to queue extended type
void csm_queue_new_type(tml_type* new_type)
{
    type_q.queue_type(new_type);
}

// Interface function to get head of type queue
tml_type* csm_type_q_get_head()
{
    return type_q.head;
}

// Utility function to return a type from an array of
//   persistent extension types and an index into the array
tml_type* csm_type_q_get_type(long index, tml_ext_type* types)
{
    tml_ext_type *cur_ext_type;
    tml_type     *cur_type;

    cur_ext_type = types + index;
    for (cur_type = csm_type_q_get_head(); cur_type; cur_type = cur_type->next)
        if (*cur_ext_type == *(cur_type->type))
            break;
    return cur_type;
}

// Utility function to return the queue position of a 
//   type in the queue of types
long csm_type_q_get_index(tml_type* this_type)
{
    long i;
    tml_type* next_type;

    for (i = 0, next_type = csm_type_q_get_head(); next_type; i++, next_type = next_type->next)
        if (next_type == this_type)
            break;
    return i;
}

