/**************************************************************************************
    heaps_traits.cxx
    functions for the heap classes of the tml modelling component


    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <stdio.h>
#include <string.h>
#include <tml/tml_os_depends.hxx>
#include <tml/tml_objects.hxx>

/****************************************************
    Class functions for heap (base class for heaps associated
    with a given model

****************************************************/
/*base class constructor for heaps -- should never be accessed directly */
heap::heap()
{
    next_byte = 0;
    in_model = NULL;
}

/*base class destructor for heaps -- should never be accessed directly */
heap::~heap()
{
    heap_start = NULL;
    in_model = NULL;
}


/****************************************************
    Object  heap functions--the object heap is where all perminent
    non-character objects in a tml model are located.

*****************************************************/
/*constructor for object heaps--should only be accessed via calls to new*/
obj_heap::obj_heap(long oheap_size)
{

    tf_node = NULL;
    tf_pvec = NULL;
    tf_prop = NULL;
    tf_arg = NULL;
    tf_p_trace = NULL;
    if (oheap_size > MAX_HEAP_INCR) {
        oheap_size = MAX_HEAP_INCR;
        printf("obj_heap::obj_heap: size requested exceeded maximum; maximum allocated\n");
        }
    size = oheap_size ? oheap_size : obj_heap_incr;
    heap_start = new char[size];
}
                                                                                                       
/*destructor for object heaps*/
obj_heap::~obj_heap()
{
    if (in_model)
        in_model->detach(this);
    delete [size] heap_start;
}
                                        
/*allocation of a object of size count bytes from a object heap*/
char *obj_heap::alloc(long count)
{
    register obj_heap* new_heap;
    register char*     new_obj;
    long               alloc_size;

    if ((next_byte + count) > size) {    // if there is no more room in this heap, make a new one
        if (count > in_model->get_oheap_size()) {
            // actual size allocated must include longword alignment pad if any
            alloc_size = (count
                         + ((count % TML_ALIGN_BLOCK)
                           ? (TML_ALIGN_BLOCK - (count % TML_ALIGN_BLOCK))
                           : 0));
            if (alloc_size > MAX_HEAP_INCR) {
                alloc_size = MAX_HEAP_INCR;
                printf("obj_heap::alloc: size requested exceeded maximum; maximum allocated\n");
                }
            new_heap = new obj_heap(alloc_size);
            }
        else
            new_heap = new obj_heap(in_model->get_oheap_size());
        new_heap->tf_node = tf_node;
        new_heap->tf_pvec = tf_pvec;
        new_heap->tf_prop = tf_prop;
        new_heap->tf_p_trace = tf_p_trace;
        in_model->attach(new_heap);
        new_obj = new_heap->alloc(count);
        in_model->current_oheap = new_heap;
        }
    else {                             // there is room in the current heap
        new_obj = heap_start + next_byte;
        next_byte += (count
                  += ((count % TML_ALIGN_BLOCK)
                     ? (TML_ALIGN_BLOCK - (count % TML_ALIGN_BLOCK))
                     : 0));
        }
    return(new_obj);
}

/*routine to free a node within the object heap*/
void obj_heap::free(node *free_node)
{
    register node  *find_node,
                        *last_node;

    free_node->sib_order = free_node->get_osize();
    free_node->n_type = is_free;
    /* free the property vector*/
    free_node->prop_detach();

    if (!tf_node) {
        free_node->up = free_node->left = 0;
        free_node->right = tf_node;
        if (tf_node)
            tf_node->left = free_node;
        tf_node = free_node;
        return;
        }

    for (find_node = tf_node;
          find_node;
          find_node = find_node->get_node(n_right)) {
        if (find_node->sib_order > free_node->sib_order)
            break;
        last_node = find_node;
        if (find_node->sib_order == free_node->sib_order) {
            free_node->up = find_node;
            free_node->right = find_node->right;
            if (find_node->right)
                find_node->right->left = free_node;
            free_node->left = find_node->left;
            if (find_node->left)
                find_node->left->right = free_node;
            find_node->right = find_node->left = 0;
            return;
            }
        }

    if (find_node){
        free_node->left = find_node->left;
        if (free_node->left)
            free_node->left->right = free_node;
        else
            tf_node = free_node;
        free_node->right = find_node;
        find_node->left = free_node;
        free_node->up = 0;
        }
    else {
        free_node->up = free_node->right = 0;
        free_node->left = last_node;
        last_node->right = free_node;
        }

}
                                                                          
/*routine to free a property vector within the object heap*/
void obj_heap::free(prop_vec *free_vec)
{
    free_vec->next = tf_pvec;
    tf_pvec = free_vec;
}
                                                
/*routine to free a property vector within the object heap*/
void obj_heap::free(property *free_prop)
{
    /*first, get rid of any argument vectors associated with the property--
        uncomment when implemented*/
    if (free_prop->entry_arg) free_prop->detach_entry();
    if (free_prop->exit_arg) free_prop->detach_exit();

    free_prop->ref_count = 0;
    free_prop->index = 0;
    free_prop->entry_arg = free_prop->exit_arg = 0;
    free_prop->locator.next = tf_prop;
    tf_prop = free_prop;
}

/*routine to free an argument vector element within the object heap*/
void obj_heap::free(argument *free_arg)
{
    free_arg->next = tf_arg;
    tf_arg = free_arg;
}

node* obj_heap::get_tf_node(long node_size)
{
    register node  *return_node;

    for (return_node = tf_node;
          return_node && (return_node->get_sib_order() <= node_size);
          return_node = return_node->get_node(n_right))
          if (return_node->get_sib_order() == node_size) return return_node;

    return 0;
}

property* obj_heap::get_tf_prop(long property_size, property*& prev_prop)
{
    register property *return_prop;

    prev_prop = NULL;

    for (return_prop = tf_prop;
           return_prop;
           prev_prop = return_prop, return_prop = return_prop->locator.next)
        if (property_size == return_prop->get_osize())
            return return_prop;

    return NULL;
}


argument* obj_heap::get_tf_arg(long arg_size)
{
    register argument *return_arg;

    for (return_arg = tf_arg;
           return_arg;
           return_arg = return_arg->next)
        if (arg_size == return_arg->get_osize())
            return return_arg;

    return NULL;
}

/**********************************************************************************
    Goo heap functions

**********************************************************************************/
                                                              
/*constructor for a goo heap--should only be called via new*/
goo_heap::goo_heap(long gheap_size)
{
    if (gheap_size > MAX_HEAP_INCR) {
        gheap_size = MAX_HEAP_INCR;
        printf("goo_heap::goo_heap: size requested exceeded maximum; maximum allocated\n");
        }
    size = gheap_size ? gheap_size : goo_heap_incr;
    heap_start = new char[size];
}
                                                                                       
/*destructor for the goo heap*/
goo_heap::~goo_heap()
{
    if (in_model)
        in_model->detach(this);
    delete [size] heap_start;
}
                                           
/*allocates an area of count bytes in the current goo heap*/
char* goo_heap::string_alloc(long count, char* copy_string)
{
    register goo_heap* new_heap;
    register char*     string_start;

    if ((next_byte + count) > size) {
    /*if there is no room in the current heap, allocate a new heap and make it current*/
        if (count > in_model->get_gheap_size()) {
            if (count > MAX_HEAP_INCR) {
                count = MAX_HEAP_INCR;
                printf("goo_heap::string_alloc: size requested exceeded maximum; maximum allocated\n");
                }
            new_heap = new goo_heap(count);
            }
        else
            new_heap = new goo_heap(in_model->get_gheap_size());
        in_model->attach(new_heap);
        in_model->current_gheap = new_heap;
        string_start = new_heap->string_alloc(count, copy_string);
        }
    else {
    /*if there is room in the current heap, allocate the needed size area*/
        string_start = heap_start + next_byte;
        if (copy_string)
            tml_os_byte_copy(copy_string, string_start, count);
        next_byte += count;
        }
    return(string_start);
}
                                                                                                   
/*allocate a single character--note that characters may take more than one byte of storage*/
char* goo_heap::char_alloc(char* copy_char, long char_size)
{
    register char*    char_start;
    register goo_heap * new_heap;

    if ((next_byte + char_size) > size) {
    /*if the heap is full, allocate a new character heap and make it current*/
        if (char_size > in_model->get_gheap_size()) {
            if (char_size  > MAX_HEAP_INCR) {
                char_size = MAX_HEAP_INCR;
                printf("goo_heap::char_alloc: size requested exceeded maximum; maximum allocated\n");
                }
            new_heap = new goo_heap(char_size);
            }
        else
            new_heap = new goo_heap(in_model->get_gheap_size());
        in_model->attach(new_heap);
        in_model->current_gheap = new_heap;
        char_start = new_heap->string_alloc(char_size, copy_char);
        }
    else {
    /*there is room in the current heap--allocate it*/
        char_start = &(heap_start[next_byte]);
        tml_os_byte_copy(copy_char, char_start, char_size);
        next_byte += char_size;
        }
    return(char_start);
}


/**********************************************************************************
    String heap functions

**********************************************************************************/
                                                              
/*constructor for a string heap--should only be called via new*/
st_heap::st_heap(long sheap_size)
{
    if (sheap_size > MAX_HEAP_INCR) {
        sheap_size = MAX_HEAP_INCR;
        printf("size requested exceeded maximum; maximum allocated\n");
        }
    size = sheap_size ? sheap_size : st_heap_incr;
    heap_start = new char[size];
}
                                                                                       
/*destructor for the string heap*/
st_heap::~st_heap()
{
    if (in_model)
        in_model->detach(this);
    delete [size] heap_start;
}
                                           
/*allocates an area of count bytes in the current string heap*/
char* st_heap::alloc(char* copy_string, long count)
{
    register st_heap* new_heap;
    register char*    string_start;

    if ((next_byte + count) > size) {
    /*if there is no room in the current heap, allocate a new heap and make it current*/
        if (count > in_model->get_sheap_size()) {
            if (count > MAX_HEAP_INCR) {
                count = MAX_HEAP_INCR;
                printf("st_heap::alloc: size requested exceeded maximum; maximum allocated\n");
                }
            new_heap = new st_heap(count);
            }
        else
            new_heap = new st_heap(in_model->get_sheap_size());
        in_model->attach(new_heap);
        in_model->current_sheap = new_heap;
        string_start = new_heap->alloc(copy_string, count);
        }
    else {
    /*if there is room in the current heap, allocate the needed size area*/
        string_start = heap_start + next_byte;
        tml_os_byte_copy(copy_string, string_start, count);
        next_byte += count;
        }
    return(string_start);
}


/**********************************************************************************
    Squeeze heap functions

**********************************************************************************/
                                                              
/*constructor for squeeze heaps--should only be accessed via calls to new*/
squeeze_heap::squeeze_heap(long sqheap_size)
{

    if (sqheap_size > MAX_HEAP_INCR) {
        sqheap_size = MAX_HEAP_INCR;
        printf("size requested exceeded maximum; maximum allocated\n");
        }
    size = sqheap_size ? sqheap_size : obj_heap_incr;
    heap_start = new char[size];
}
                                                                                                       
/*destructor for object heaps*/
squeeze_heap::~squeeze_heap()
{
    // in_model element should never be other than NULL
    //   but set to NULL here for safety
    if (in_model)
        in_model = 0;
    delete [size] heap_start;
}
                                        
char *squeeze_heap::alloc(long count)
{
    register char * new_obj;

    // there is always room in the current heap -- identification of
    //   current heap (either by allocation or by following a pointer)
    //   is taken care of by an associated d_model_parts object
    new_obj = heap_start + next_byte;
    next_byte += (count
              += ((count % TML_ALIGN_BLOCK)
                 ? (TML_ALIGN_BLOCK - (count % TML_ALIGN_BLOCK))
                 : 0));
    return(new_obj);
}

/*********************************************************************************
    Functions for heaps containing heterogeneous data (e.g. graphics)

**********************************************************************************/
                                                              
/* constructor for bit_heap class - should never be called directly */
bit_heap::bit_heap()
{
    in_model = NULL;
    heap_start = NULL;
    next_byte = 0;
    size = 0;
    lock = new csm_lock((void*) this);
}

/* destructor for bit_heap class */
bit_heap::~bit_heap()
{
    delete lock;
    lock = NULL;
    delete [size] heap_start;
}


/* constructor for write_bit_heap, the heap used during model save */
/*   for heterogeneous data (e.g. graphics) */
/* should be called only via new */
/* the data argument is user-supplied data for the heap */
write_bit_heap::write_bit_heap(void* data, long count)
{
    next_byte = count;
    size = count;
    heap_start = (char*) data;
    in_model = NULL;
}
                                                                                      

/* constructor for read_bit_heap, the graphics heap used during model restore; */
/*   should be called only via new. */
/* allocates count bytes of data from free space */
read_bit_heap::read_bit_heap(long count)
{
    next_byte = count;
    size = count;
    heap_start = new char[size];
}


