/**************************************************************************************
    history.cxx

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <stdio.h>
#include <tml/tml_objects.hxx>
#include <tml/tml_os_depends.hxx>

// prefix "ao_" refers to atom object heaps

/**********************************************************************************

        Functions for history and journalling, undo and redo

**********************************************************************************/

/* constructor for history */
history::history()
{
    ao_store = new ao_list;
    top_free_molecule = NULL;
    top_free_atom = NULL;
    top_free_this_arg = NULL;
    top_free_ptr_arg = NULL;
    top_free_val_arg = NULL;
    head_atom = NULL;
    tail_atom = NULL;
    hist = NULL;
    hist_size = 0;
    hist_array_size = 0;
    active = -1;            // ready to record when and if hist allocated
    recursing = 0;
    next_op = 0;
    undo_op = 0;
    in_effect = 0;
    filled = 0;
    uninitialized = -1;     // no molecule array yet allocated
}

/* destructor for history */
history::~history()
{
    ao_list *ao_current;
    ao_list *ao_next;

    if (head_atom)
        end_molecule();                     // end the last molecular operation
    if (hist_size)
        set_hist_size(0);                   // delete history
    for (ao_current = ao_store; ao_current != NULL; ao_current = ao_next) {
    // get rid of the atom object heaps associated with the model's history
        ao_next = ao_current->get_next();
        delete ao_current;
        }
    ao_store = NULL;
    top_free_molecule = NULL;
    top_free_atom = NULL;
    top_free_this_arg = NULL;
    top_free_ptr_arg = NULL;
    top_free_val_arg = NULL;
    head_atom = NULL;
    tail_atom = NULL;
    hist = NULL;
}

/*free storage associated with a history to be deleted */
void history::hist_free()
{
    long i;

    for (i = 0; i < hist_size; i++)
        *(hist + i) = NULL;
}
        
/*set number of molecular operations recorded in history */
/*  return actual new size (see shrink, below) */
long history::set_hist_size(long size)
{
    long num_hist_bytes;
    long i;
    long j;
    long min_size;
    molecule* *new_hist;
    molecule *cur_molecule;

    if (uninitialized && (size != 0)) {
        uninitialized = 0;
        num_hist_bytes = size * (long) sizeof(molecule*);
        hist = (molecule**) alloc_obj_bytes(num_hist_bytes);
        for (i = 0; i < size; i++)
            *(hist + i) = NULL;
        hist_size = size;
        hist_array_size = size;
        next_op = 0;
        undo_op = 0;
        in_effect = 0;
        filled = 0;
        return(hist_size);
        }

    if (0 > size) {                                         // bad argument
        printf("history::set_hist_size: illegal size (negative)\n");
        return(hist_size);
        }

    else if (0 == size) {                                   // cancel history
        head_atom = NULL;
        tail_atom = NULL;
        for (i = 0; i < hist_size; i++) {
            if (NULL != (cur_molecule = *(hist + i)))
                cur_molecule->free();
            }
        if (NULL != hist) {
            hist_free();
            hist = NULL;
            hist_size = 0;
            hist_array_size = 0;
            }
        return(0);                                          // notify that history is now off
        }

    else if (size == hist_size)                             // do nothing
        return(hist_size);                                  //   return current history size

    else if (size < hist_size) {                            // shrink history - doesn't shrink array
        if (size >= filled)                                 // can preserve all information
            i = filled - 1;
        else {                                              // will lose some history of jobs not undone
            if (next_op > undo_op) {                        // min size is the greater of 1 or the
                min_size = next_op - undo_op;               //   sum of redoable operations plus one
                filled = size;
                }                                           //   remaining operation to undo
            else if (next_op == undo_op) {
                min_size = 1;
                filled = 0;
                }
            else {
                min_size = next_op + hist_size - undo_op;
                filled = min_size;
                }
            in_effect = 1;
            if (size < min_size)
                size = min_size;
            i = size - 1;
            filled = size;
            }
        j = next_op - 1;
        if (j < 0)
            j = hist_size - 1;
        for (; i >= 0; i--) {
            if (undo_op == j)
                undo_op = i;
            *(hist + i) = *(hist + j--);
            if (j < 0)
                j = hist_size - 1;
            }
        hist_size = size;
        next_op = 0;
        }

    else {                                                  // expand history
        if (size > hist_array_size) {                       // must expand hist array
            num_hist_bytes = size * (long) sizeof(molecule*);
            new_hist = (molecule**) alloc_obj_bytes(num_hist_bytes);
            for (i = size; i > filled; i--)
                *(new_hist + i) = NULL;
            for (i = filled; i; i--)
                *(new_hist + i) = new molecule(this);
            hist_array_size = size;
            if (filled == hist_size)
                j = next_op;
            else if (next_op - 1 >= 0)
                j = next_op - 1;
            else
                j = hist_size - 1;
            for (i = filled; i; i--) {
                (*(new_hist + i))->copy_free(*(hist + j));
                j = (--j >= 0) ? j : (size - 1);
                }
            hist_free();
            hist = new_hist;
            }
        hist_size = size;
        }
}

char* history::alloc_obj_bytes(unsigned long nbytes) {

    ao_list *new_ao_list;
    char *retval;

    if (ao_store->this_heap->heap_size < (nbytes + ao_store->this_heap->next_byte)) {
        new_ao_list = new ao_list((nbytes > tml_ao_heap_incr) ? nbytes : tml_ao_heap_incr);
        new_ao_list->next = ao_store;
        ao_store = new_ao_list;
        }
    retval = ao_store->this_heap->bytes + ao_store->this_heap->next_byte;
    ao_store->this_heap->next_byte += (nbytes
                                   += ((nbytes % TML_ALIGN_BLOCK)
                                      ? TML_ALIGN_BLOCK - (nbytes % TML_ALIGN_BLOCK)
                                      : 0));

    return(retval);
}

/* add atom to current molecule; if new_atom is null, add molecule to history */
long history::add_atom(atom *new_atom)
{
    long retval;

    if (NULL == hist) {
        printf("history::add_atom: history is not enabled\n");
        return(-1);
        }
    else if ((NULL == head_atom) && (NULL != new_atom)) {  // first atom in molecule
        head_atom = tail_atom = new_atom;
        return(0);
        }
    else {
        if (NULL == new_atom) {                            // signal to end and add molecule
            retval = add_molecule(head_atom, tail_atom);   // NOTE: one could add NULL molecules
            head_atom = NULL;                              //   for per-time work record
            tail_atom = NULL;
            return(retval);
            }
        else {
            new_atom->set_prev(NULL); // should already be NULL from atom::free or atom::atom
            new_atom->set_next(head_atom);
            if (head_atom)
                head_atom->set_prev(new_atom);
            head_atom = new_atom;
            return(0);
            }
        }
}

/* terminate a molecule */
void history::end_molecule()
{
    add_atom(NULL);
}

/* add a molecular operation to history */
long history::add_molecule(atom *head, atom *tail)
{
    molecule *cur_molecule;

    if (NULL == hist) {
        printf("history::add_molecule: history is not enabled\n");
        return(-1);
        }

    if (filled) {

        // some history has already been recorded

        do {

            // clear "next_op" slot and any slots containing
            //   undone molecules

            cur_molecule = *(hist + next_op);
            if (NULL != cur_molecule) {
                cur_molecule->free();
                --filled;
                }
            *(hist + next_op) = NULL;
            --next_op;
            if (next_op < 0)
                next_op = hist_size - 1;
            } while (next_op != undo_op);

        // if any molecules remain in effect, set next_op past them

        if (in_effect)
            next_op = (++undo_op) % hist_size;

        // else if a previously undone molecule lurks at undo_op,
        //   free it

        else if (filled == 1) {
            cur_molecule = *(hist + undo_op);
            if (NULL != cur_molecule) {
                cur_molecule->free();
                --filled;
                }
            *(hist + undo_op) = NULL;
            }

        // else if filled is not zero, error

        else if (filled != 0) {
            printf("history::add_molecule: hist_array indexes impossibly out of synch\n");
            return(-1);
            }
        }

    ++filled;
    ++in_effect;
    cur_molecule = new molecule(this);
    cur_molecule->set(head, tail);
    cur_molecule->in_history = this;
    *(hist + next_op) = cur_molecule;
    undo_op = next_op;
    next_op = (++next_op) % hist_size;
    return(0);
}


/* undo a molecular operation */
long history::undo_molecule()
{
    active = 0;              // turn off recording of history during undo
    if (NULL == hist) {
        printf("history::undo_molecule: history is not enabled\n");
        active = -1;         // resume recording history
        return(-1);          // signal error
        }
    if (head_atom)
        end_molecule();
    if (in_effect) {
        --in_effect;
        (*(hist + undo_op))->undo();
        if (in_effect) {
            if (--undo_op < 0)
                undo_op = hist_size - 1;
            }
        active = -1;        // resume recording history
        return(0);
        }
    else {
        printf("history::undo_molecule: nothing to undo\n");
        active = -1;        // resume recording history
        return(-1);         // signal error
        }
}

/* redo a molecular operation */
long history::redo_molecule()
{
    active = 0;             // turn off recording of history during redo
    if (NULL == hist) {
        printf("history::redo_molecule: history is not enabled\n");
        active = -1;        // resume recording history
        return(-1);         // signal error
        }
    if (in_effect == filled) {
        printf("history::redo_molecule: nothing to redo\n");
        active = -1;        // resume recording history
        return(-1);         // signal error
        }
    else if (in_effect < filled) {
        if (in_effect)
            undo_op = (undo_op + 1) % hist_size;
        (*(hist + undo_op))->redo();
        ++in_effect;
        active = -1;        // resume recording history
        return(0);
        }
    else {
        printf("history::redo_molecule: hist_array indexes impossibly out of synch\n");
        return(-1);
        }
}


/************************************************************************

    Molecule traits

************************************************************************/

/* make a molecule */
molecule::molecule(history* to_history)
{
    molecule* cur_molecule;

    if (cur_molecule = to_history->get_top_free_molecule()) {
        to_history->set_top_free_molecule(cur_molecule->next);
        cur_molecule->set_next(NULL);
        this = cur_molecule;
        }
    else
        this = (molecule*) to_history->alloc_obj_bytes((long) sizeof(molecule));
    head_atom = NULL;
    tail_atom = NULL;
    in_history = to_history;
    next = NULL;
}

molecule::~molecule()
{
    this = NULL;
}

/* undo molecular operation */
void molecule::undo()
{
    atom *cur_atom;

    for (cur_atom = head_atom; cur_atom; cur_atom = cur_atom->get_next())
        cur_atom->undo();
}

/* redo molecular operation */
void molecule::redo()
{
    atom *cur_atom;

    for (cur_atom = tail_atom; cur_atom; cur_atom = cur_atom->get_prev())
        cur_atom->redo();
}

/* copy head_atom and tail_atom from the argument molecule and free it */
void molecule::copy_free(molecule* from_molecule)
{
    head_atom = from_molecule->head_atom;
    tail_atom = from_molecule->tail_atom;
    from_molecule->free();
    }

/* free atoms associated with a molecule, then free the molecule */
void molecule::free()
{
    atom          *cur_atom;
    atom          *next_atom;
    molecule      *top_free_molecule;

    if (head_atom) {

        // atom::free resets cur_atom->next to point to former
        //   top of freelist, so must preserve cur_atom->next
        //   in next_atom before freeing cur_atom

        cur_atom = next_atom = head_atom;
        while (next_atom) {
            next_atom = cur_atom->get_next();
            cur_atom->free(in_history);
            cur_atom = next_atom;
            }
        }

    top_free_molecule = in_history->get_top_free_molecule();
    in_history->set_top_free_molecule(this);
    next = top_free_molecule;
    head_atom = NULL;
    tail_atom = NULL;
    in_history = NULL;
}

/************************************************************************

    Atom traits

************************************************************************/

/* atom argument constructor */
atom_arg::atom_arg(history *to_history, atom_arg_type type)
{
    atom_arg *top_arg;

    switch (type) {
    case this_arg:
        if (NULL == (top_arg = to_history->get_top_free_this_arg()))
            this = (atom_this_arg*) to_history->alloc_obj_bytes((long) sizeof(atom_this_arg));
        else {
            this = (atom_this_arg*) top_arg;
            to_history->set_top_free_this_arg((atom_this_arg*) this->get_next());
            this->set_next(NULL);
            }
        break;
    case ptr_arg:
        if (NULL == (top_arg = to_history->get_top_free_ptr_arg()))
            this = (atom_ptr_arg*) to_history->alloc_obj_bytes((long) sizeof(atom_ptr_arg));
        else {
            this = (atom_ptr_arg*) top_arg;
            to_history->set_top_free_ptr_arg((atom_ptr_arg*) this->get_next());
            this->set_next(NULL);
            }
        break;
    case val_arg:
        if (NULL == (top_arg = to_history->get_top_free_val_arg()))
            this = (atom_val_arg*) to_history->alloc_obj_bytes((long) sizeof(atom_val_arg));
        else {
            this = (atom_val_arg*) top_arg;
            to_history->set_top_free_val_arg((atom_val_arg*) this->get_next());
            this->set_next(NULL);
            }
        break;
    }
}

atom_arg::~atom_arg()
{
    this = NULL;
}

/* make an atom argument that indicates whether the undo/redo function is a */
/*   regular function call or a member function. If the former, member "origin" */
/*   is NULL; if the latter, "origin" is a pointer to the object for which the */
/*   member undo/redo function is to be called. */
atom_this_arg::atom_this_arg(history *to_history, void* source) : (to_history, this_arg)
{
    set_contents(source);
}

atom_this_arg::~atom_this_arg()
{
    set_contents(NULL);
}

void atom_this_arg::set_top_free(history* from_history)
{
    atom_this_arg *was_top;

    if (was_top = from_history->get_top_free_this_arg())
        this->set_next(was_top);
    from_history->set_top_free_this_arg(this);
}

/* make an atom pointer argument - origin member is a pointer to an object */
/*   with a reference count, i.e. an object that may the subject of a call */
/*   of delete without having its storage deallocated, depending on the */
/*   reference count value */
atom_ptr_arg::atom_ptr_arg(history *to_history, void* source) : (to_history, ptr_arg)
{
    set_contents(source);
}

atom_ptr_arg::~atom_ptr_arg()
{
    set_contents(NULL);
}

void atom_ptr_arg::set_top_free(history* from_history)
{
    atom_ptr_arg *was_top;

    if (was_top = from_history->get_top_free_ptr_arg())
        this->set_next(was_top);
    from_history->set_top_free_ptr_arg(this);
}

/* make an atom value argument. Member "origin" is a value not to be deleted, */
/*   e.g. a number or a pointer to an object that does not have a reference count. */
/*   Value is stored in a void* and will be pushed on the stack in a void* as an */
/*   argument to undo and redo functions by atom::undo() and atom::redo() respectively. */
/*   The value will be pushed as an argument to its undo/redo functions in a stack */
/*   object of the same size as a void*. Overloaded function. Note that float values */
/*   are explicitly cast to type long before being cast to void*. Note also that char, */
/*   short, and int values are converted to long by C++ standard type conversion, which */
/*   then selects the instance with (history*, long) argument list. */
atom_val_arg::atom_val_arg(history *to_history, void* source) : (to_history, val_arg)
{
    set_contents(source);
}

atom_val_arg::atom_val_arg(history* to_history, long  source) : (to_history, val_arg)
{
    set_contents((void*) source);
}

atom_val_arg::atom_val_arg(history* to_history, float source) : (to_history, val_arg)
{
    set_contents((void*) ((long) source));
}

atom_val_arg::~atom_val_arg()
{
    set_contents(NULL);
}

void atom_val_arg::set_top_free(history* from_history)
{
    atom_val_arg *was_top;

    if (was_top = from_history->get_top_free_val_arg())
        this->set_next(was_top);
    from_history->set_top_free_val_arg(this);
}

/* make an atom - arguments are made and passed by caller - first NULL argument */
/*   found signals end of undo arguments; second NULL argument found signals end of redo */
/*   arguments. Maximum 20 non-NULL arguments. First undo and first redo argument signal */
/*   whether corresponding function pointer is a function call or a member function. If */
/*   the contents of the first argument is NULL, the corresponding function pointer is */
/*   a function call, otherwise it is a member function. See atom::undo() and atom::redo() */
/*   for the significance of this. */
atom::atom(history *to_history, WHOLE_PTR undo_funcp, WHOLE_PTR redo_funcp,
           atom_arg* arg1 ,  atom_arg* arg2 ,  atom_arg* arg3 ,
           atom_arg* arg4 ,  atom_arg* arg5 ,  atom_arg* arg6 ,
           atom_arg* arg7 ,  atom_arg* arg8 ,  atom_arg* arg9 ,
           atom_arg* arg10,  atom_arg* arg11,  atom_arg* arg12,
           atom_arg* arg13,  atom_arg* arg14,  atom_arg* arg15,
           atom_arg* arg16,  atom_arg* arg17,  atom_arg* arg18,
           atom_arg* arg19,  atom_arg* arg20,  atom_arg* arg21,
           atom_arg* arg22,  atom_arg* arg23,  atom_arg* arg24,
           atom_arg* arg25,  atom_arg* arg26)
{
    atom_arg* argv[26];

    long i;
    long j;
    long num_args;
    long num_undo_args;
    long num_redo_args;
    atom            *cur_atom;

    // determine size of argument string

    argv[0]  = arg1;    argv[1]  = arg2;    argv[2]  = arg3;    argv[3]  = arg4;
    argv[4]  = arg5;    argv[5]  = arg6;    argv[6]  = arg7;    argv[7]  = arg8;
    argv[8]  = arg9;    argv[9]  = arg10;   argv[10] = arg11;   argv[11] = arg12;
    argv[12] = arg13;   argv[13] = arg14;   argv[14] = arg15;   argv[15] = arg16;
    argv[16] = arg17;   argv[17] = arg18;   argv[18] = arg19;   argv[19] = arg20;
    argv[20] = arg21;   argv[21] = arg22;   argv[22] = arg23;   argv[23] = arg24;
    argv[24] = arg25;   argv[25] = arg26;

    for (i = 0; argv[i]; i++)
        ;
    num_undo_args = i;
    ++i;

    for (num_redo_args = 0; argv[i]; i++)
        ++num_redo_args;
        ;

    num_args = num_undo_args + num_redo_args;

    // first fit

    for (cur_atom = to_history->get_top_free_atom();
      cur_atom;
      cur_atom = cur_atom->get_next()) {
        if (num_args <= cur_atom->get_count()) {
            if (cur_atom->prev)
                cur_atom->prev->next = cur_atom->next;
            else
                to_history->set_top_free_atom(cur_atom->next);
            if (cur_atom->next)
                cur_atom->next->prev = cur_atom->prev;
            break;
            }
        }

    // fit found

    if (NULL != cur_atom)
        this = cur_atom;

    // no fit found

    else {
        this = (atom *) to_history->alloc_obj_bytes((long) sizeof(atom));
        args = (atom_arg**) to_history->alloc_obj_bytes(num_args * (long) sizeof(atom_arg*));
        count = num_args;
        }

    undo_func = (GEN_FUNCP) undo_funcp;
    redo_func = (GEN_FUNCP) redo_funcp;
    undo_args = num_undo_args;
    redo_args = num_redo_args;
    next = NULL;
    prev = NULL;

    for (i = 0; i < undo_args; i++)
        *(args + i) = argv[i];
    for (j = undo_args + 1; j <= num_args; j++)     // skip NULL arguments
        *(args + i++) = argv[j];
}

/* destroy an atom: NULL its pointers */
atom::~atom()
{
    undo_func = NULL;
    redo_func = NULL;
    args = NULL;
    next = NULL;
    prev = NULL;
    this = NULL;
}


void atom::free(history *from_history)
{
    long i, j;
    atom     *top_free_atom;
    atom_arg *cur_arg;

    if (args) {
        for (i = 0; i < undo_args + redo_args; i++) {
            cur_arg = *(args + i);
            for (j = 0; j < i; j++) {           // avoid freeing same arg twice
                if (cur_arg == *(args + j))
                    break;
                }
            if (j == i) {
                cur_arg->empty();
                cur_arg->set_top_free(from_history);
                }
            }
        }

    // this assignment made to prevent CC name confusion

    top_free_atom = from_history->get_top_free_atom();
    next = top_free_atom;
    prev = NULL;
    from_history->set_top_free_atom(this);
}

/*****************************************************************************

    Atom object and character heap list traits

*****************************************************************************/

ao_list::ao_list(unsigned long new_heap_size)
{
    this_heap = new atom_obj_heap(new_heap_size);
    next = NULL;
}

ao_list::~ao_list()
{
    delete this_heap;
    this_heap = NULL;
}

atom_obj_heap::atom_obj_heap(unsigned long new_heap_size)
{
    bytes = new char[new_heap_size];
    next_byte = 0;
    heap_size = new_heap_size;
}

atom_obj_heap::~atom_obj_heap()
{
    delete [heap_size] bytes;
    bytes = NULL;
}



/*****************************************************************************

    Undo and redo atom

*****************************************************************************/

/* undo atom */
void atom::undo()
{
    long i;

    void *arg1,  *arg2,  *arg3,  *arg4,  *arg5,  *arg6,  *arg7,  *arg8,  *arg9,  *arg10,
         *arg11, *arg12, *arg13, *arg14, *arg15, *arg16, *arg17, *arg18, *arg19, *arg20,
         *arg21;

    void **arg1p = &arg1, **arg2p = &arg2, **arg3p = &arg3, **arg4p = &arg4, **arg5p = &arg5,
         **arg6p = &arg6, **arg7p = &arg7, **arg8p = &arg8, **arg9p = &arg9, **arg10p = &arg10,
         **arg11p = &arg11, **arg12p = &arg12, **arg13p = &arg13, **arg14p = &arg14, **arg15p = &arg15,
         **arg16p = &arg16, **arg17p = &arg17, **arg18p = &arg18, **arg19p = &arg19, **arg20p = &arg20,
         **arg21p = &arg21;

    void* *argv[21];

    argv[0]  = arg1p;
    argv[1]  = arg2p;
    argv[2]  = arg3p;
    argv[3]  = arg4p;
    argv[4]  = arg5p;
    argv[5]  = arg6p;
    argv[6]  = arg7p;
    argv[7]  = arg8p;
    argv[8]  = arg9p;
    argv[9]  = arg10p;
    argv[10] = arg11p;
    argv[11] = arg12p;
    argv[12] = arg13p;
    argv[13] = arg14p;
    argv[14] = arg15p;
    argv[15] = arg16p;
    argv[16] = arg17p;
    argv[17] = arg18p;
    argv[18] = arg19p;
    argv[19] = arg20p;
    argv[20] = arg21p;

    for (i = 0; i < undo_args; i++)
        *argv[i] = (*(args + i))->origin;

    // contents of first argument indicates if undo function is a function call or a member function.
    // if contents of first argument is NULL, undo function is a function call, and the first argument isn't used
    // otherwise undo function is a member function, and the contents of the first argument is the object
    //   whose member function is to be called; in that case, the first argument is used

    if (NULL == arg1) {        // undo function is a regular function call
        switch (undo_args) {
        case 1:
            (*undo_func)();
            break;
        case 2:
            (*undo_func)(arg2);
            break;
        case 3:
            (*undo_func)(arg2,  arg3);
            break;
        case 4:
            (*undo_func)(arg2,  arg3,  arg4);
            break;
        case 5:
            (*undo_func)(arg2,  arg3,  arg4,  arg5);
            break;
        case 6:
            (*undo_func)(arg2,  arg3,  arg4,  arg5,  arg6);
            break;
        case 7:
            (*undo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7);
            break;
        case 8:
            (*undo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8);
            break;
        case 9:
            (*undo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9);
            break;
        case 10:
            (*undo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10);
            break;
        case 11:
            (*undo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11);
            break;
        case 12:
            (*undo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12);
            break;
        case 13:
            (*undo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13);
            break;
        case 14:
            (*undo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14);
            break;
        case 15:
            (*undo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15);
            break;
        case 16:
            (*undo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16);
            break;
        case 17:
            (*undo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16, arg17);
            break;
        case 18:
            (*undo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16, arg17, arg18);
            break;
        case 19:
            (*undo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16, arg17, arg18, arg19);
            break;
        case 20:
            (*undo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16, arg17, arg18, arg19, arg20);
            break;
        case 21:
            (*undo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16, arg17, arg18, arg19, arg20,  arg21);
            break;
            } /* end switch for function call case*/
        } /* end if */
    else {                                        // undo function is a member function call
        switch (undo_args) {
        case 1:
            if (NULL == undo_func)
                delete arg1;
            else
                (*undo_func)(arg1);
            break;
        case 2:
            (*undo_func)(arg1, arg2);
            break;
        case 3:
            (*undo_func)(arg1, arg2,  arg3);
            break;
        case 4:
            (*undo_func)(arg1, arg2,  arg3,  arg4);
            break;
        case 5:
            (*undo_func)(arg1, arg2,  arg3,  arg4,  arg5);
            break;
        case 6:
            (*undo_func)(arg1, arg2,  arg3,  arg4,  arg5,  arg6);
            break;
        case 7:
            (*undo_func)(arg1, arg2,  arg3,  arg4,  arg5,  arg6,  arg7);
            break;
        case 8:
            (*undo_func)(arg1, arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8);
            break;
        case 9:
            (*undo_func)(arg1, arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9);
            break;
        case 10:
            (*undo_func)(arg1, arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10);
            break;
        case 11:
            (*undo_func)(arg1, arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11);
            break;
        case 12:
            (*undo_func)(arg1, arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12);
            break;
        case 13:
            (*undo_func)(arg1,  arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13);
            break;
        case 14:
            (*undo_func)(arg1,  arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14);
            break;
        case 15:
            (*undo_func)(arg1,  arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15);
            break;
        case 16:
            (*undo_func)(arg1,  arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16);
            break;
        case 17:
            (*undo_func)(arg1,  arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16, arg17);
            break;
        case 18:
            (*undo_func)(arg1,  arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16, arg17, arg18);
            break;
        case 19:
            (*undo_func)(arg1,  arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16, arg17, arg18, arg19);
            break;
        case 20:
            (*undo_func)(arg1,  arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16, arg17, arg18, arg19, arg20);
            break;
        case 21:
            (*undo_func)(arg1,  arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16, arg17, arg18, arg19, arg20,  arg21);
            break;
            } /* end switch */
        } /* end else */
}

/* redo atom */
void atom::redo()
{
    long i;
    long j;

    void *arg1,  *arg2,  *arg3,  *arg4,  *arg5,  *arg6,  *arg7,  *arg8,  *arg9,  *arg10,
         *arg11, *arg12, *arg13, *arg14, *arg15, *arg16, *arg17, *arg18, *arg19, *arg20,
         *arg21;

    void **arg1p = &arg1, **arg2p = &arg2, **arg3p = &arg3, **arg4p = &arg4, **arg5p = &arg5,
         **arg6p = &arg6, **arg7p = &arg7, **arg8p = &arg8, **arg9p = &arg9, **arg10p = &arg10,
         **arg11p = &arg11, **arg12p = &arg12, **arg13p = &arg13, **arg14p = &arg14, **arg15p = &arg15,
         **arg16p = &arg16, **arg17p = &arg17, **arg18p = &arg18, **arg19p = &arg19, **arg20p = &arg20,
         **arg21p = &arg21;

    void* *argv[21];

    argv[0]  = arg1p;
    argv[1]  = arg2p;
    argv[2]  = arg3p;
    argv[3]  = arg4p;
    argv[4]  = arg5p;
    argv[5]  = arg6p;
    argv[6]  = arg7p;
    argv[7]  = arg8p;
    argv[8]  = arg9p;
    argv[9]  = arg10p;
    argv[10] = arg11p;
    argv[11] = arg12p;
    argv[12] = arg13p;
    argv[13] = arg14p;
    argv[14] = arg15p;
    argv[15] = arg16p;
    argv[16] = arg17p;
    argv[17] = arg18p;
    argv[18] = arg19p;
    argv[19] = arg20p;
    argv[20] = arg21p;

    j = 0;
    for (i = undo_args; i < undo_args + redo_args ; i++) // count >= undo_args + redo_args from
        *argv[j++] = (*(args + i))->origin;              //  first-first algorithm for atom reuse

    // contents of first argument indicates if redo function is a function call or a member function.
    // if contents of first argument is NULL, redo function is a function call, and the first argument isn't used
    // otherwise redo function is a member function, and the contents of the first argument is the object
    //   whose member function is to be called; in that case, the first argument is used

    if (NULL == arg1) {        // redo function is a regular function call
        switch (redo_args) {
        case 1:
            (*redo_func)();
            break;
        case 2:
            (*redo_func)(arg2);
            break;
        case 3:
            (*redo_func)(arg2,  arg3);
            break;
        case 4:
            (*redo_func)(arg2,  arg3,  arg4);
            break;
        case 5:
            (*redo_func)(arg2,  arg3,  arg4,  arg5);
            break;
        case 6:
            (*redo_func)(arg2,  arg3,  arg4,  arg5,  arg6);
            break;
        case 7:
            (*redo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7);
            break;
        case 8:
            (*redo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8);
            break;
        case 9:
            (*redo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9);
            break;
        case 10:
            (*redo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10);
            break;
        case 11:
            (*redo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11);
            break;
        case 12:
            (*redo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12);
            break;
        case 13:
            (*redo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13);
            break;
        case 14:
            (*redo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14);
            break;
        case 15:
            (*redo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15);
            break;
        case 16:
            (*redo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16);
            break;
        case 17:
            (*redo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16, arg17);
            break;
        case 18:
            (*redo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16, arg17, arg18);
            break;
        case 19:
            (*redo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16, arg17, arg18, arg19);
            break;
        case 20:
            (*redo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16, arg17, arg18, arg19, arg20);
            break;
        case 21:
            (*redo_func)(arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16, arg17, arg18, arg19, arg20,  arg21);
            break;
            } /* end switch for function call case*/
        } /* end if */
    else {                                        // redo function is a member function call
        switch (redo_args) {
        case 1:
            if (NULL == redo_func)
                delete arg1;
            else
                (*redo_func)(arg1);
            break;
        case 2:
            (*redo_func)(arg1, arg2);
            break;
        case 3:
            (*redo_func)(arg1, arg2,  arg3);
            break;
        case 4:
            (*redo_func)(arg1, arg2,  arg3,  arg4);
            break;
        case 5:
            (*redo_func)(arg1, arg2,  arg3,  arg4,  arg5);
            break;
        case 6:
            (*redo_func)(arg1, arg2,  arg3,  arg4,  arg5,  arg6);
            break;
        case 7:
            (*redo_func)(arg1, arg2,  arg3,  arg4,  arg5,  arg6,  arg7);
            break;
        case 8:
            (*redo_func)(arg1, arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8);
            break;
        case 9:
            (*redo_func)(arg1, arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9);
            break;
        case 10:
            (*redo_func)(arg1, arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10);
            break;
        case 11:
            (*redo_func)(arg1, arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11);
            break;
        case 12:
            (*redo_func)(arg1, arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12);
            break;
        case 13:
            (*redo_func)(arg1,  arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13);
            break;
        case 14:
            (*redo_func)(arg1,  arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14);
            break;
        case 15:
            (*redo_func)(arg1,  arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15);
            break;
        case 16:
            (*redo_func)(arg1,  arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16);
            break;
        case 17:
            (*redo_func)(arg1,  arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16, arg17);
            break;
        case 18:
            (*redo_func)(arg1,  arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16, arg17, arg18);
            break;
        case 19:
            (*redo_func)(arg1,  arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16, arg17, arg18, arg19);
            break;
        case 20:
            (*redo_func)(arg1,  arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16, arg17, arg18, arg19, arg20);
            break;
        case 21:
            (*redo_func)(arg1,  arg2,  arg3,  arg4,  arg5,  arg6,  arg7,  arg8,  arg9,  arg10,  arg11,
                         arg12, arg13, arg14, arg15, arg16, arg17, arg18, arg19, arg20,  arg21);
            break;
            } /* end switch */
        } /* end else */
}
