/**************************************************************************************
    kmp.cxx
    Utility functions implementing the Knuth-Morris-Pratt string search algorithm

    (c) Copyright Apollo Computer Inc., 1989 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/
               
#include <string.h>
#include <stdio.h>

#ifndef csm_kmp_hxx
#include <tml/csm_kmp.hxx>
#endif

#ifndef char_set_mgr_hxx
#include <tml/char_set_mgr.hxx>
#endif

#include <tml/tml_os_depends.hxx>

/**************************************************************************************
    member functions for class kmp_node
**************************************************************************************/

kmp_node::kmp_node()
{
    up = NULL;
    right = NULL;
    down = NULL;
    this_character = NULL;
    failure = NULL;
    matched = NULL;
}

// deletes entire tree: should therefore be called only on root
kmp_node::~kmp_node()
{
    up = NULL;
    right = NULL;
    down = NULL;
    delete this_character; // deletes vector for entire tree
    this_character = NULL;
    failure = NULL;
    matched = NULL;
}

/* breadth-first search */
kmp_node* kmp_node::breadth_first(kmp_dir& dir, unsigned long& levels)
{
    switch (dir) {                   // breadth first search
        case kmp_right:              //   requesting right sibling
            if (right)               //     have right sibling
                return get_right();  //       return it
            else if (up) {           //     no, but not at root node
                ++levels;            //       must go down one more than up
                dir = kmp_up;        //       signal going up
                return get_up();     //       return the parent
                }
            else                     //     at root node - not used in kmp_fsm
                return get_down();  //       return first child
            break;

        case kmp_up:                 //   going up
            ++levels;                //     count this move
            if (right) {             //     have right node?
                dir = kmp_down;      //       signal going down
                return get_right();  //       return right node
                }
            else if (up)             //     no, but haven't reached root
                return get_up();     //       continue going up
            else {                   //     have reached root node
                --levels;            //       going down one level
                dir = kmp_down;      //       signal going down
                return get_down();   //       return first child
                }
            break;

        case kmp_down:               //   going down
            if (down) {              //     have child?
                if (0 == --levels)   //       have reached goal level?
                    dir = kmp_right; //         signal success
                return get_down();   //       return first child
                }
            else if (right)          //     no, but have right sibling
                return get_right();  //       return it
            else {                   //     no sibling
                dir = kmp_up;        //       signal going up
                return this;         //       return self
                }
            break;
        }
}

/* depth-first search */
kmp_node* kmp_node::depth_first(kmp_dir& dir)
{
    switch (dir) {
        case kmp_down:
            if (down)
                return get_down();
            else if (right)
                return get_right();
            else {
                dir = kmp_up;
                return get_up();
                }
        case kmp_up:
            if (right) {
                dir = kmp_down;
                return get_right();
                }
            else
                return get_up();

        case kmp_right:
            printf("kmp_node::depth_first: bad direction\n");
            return this;
        }
}

kmp_node* kmp_node::check_lookup_backup(kmp_node* cur_node, unsigned char* s, unsigned long char_size)
{
    kmp_node* last_node;

    last_node = cur_node;
    cur_node = cur_node->get_right();                    //   try siblings
    if ( ! cur_node) {                                   //   nothing doing
        cur_node = last_node->get_failure();             //     go to last success
        if (cur_node && (this != cur_node))              //     still have a match
            cur_node->matched = s - char_size;           //       record character matched
        else                                             //     failure to root
            return NULL;
        }
    return cur_node;
}

/* recognize collating sequence beginning at s if any, return collation order */
/*   and advance s beyond collating sequence found, else return MAX_CHARS_IN_SET */
unsigned long kmp_node::lookup(unsigned char*& str)
{
    unsigned char* s           = str; 
    kmp_node*      cur_node    = down;
    kmp_node*      last_match  = NULL;
    char_set*      in_char_set;
    unsigned long  char_size;
    unsigned long  collation   = MAX_CHARS_IN_SET;

    in_char_set = this_character->in_char_set;
    if (NULL == in_char_set)
        in_char_set = get_dflt_union_char_set()->get_sets();
    char_size = in_char_set->get_char_size();
    while (cur_node) {
        if (cur_node->matches(s)) {                              // successful character match
            if (MAX_CHARS_IN_SET !=
              (collation = cur_node->this_character->collation)) //   terminal match; look for longer match
                last_match = cur_node;                           //     record terminal match
            cur_node->matched = s;                               //   record character matched
            in_char_set->advance(s);                             //   advance to next character
            cur_node = cur_node->get_down();                     //   try children if any
            }                                                    
        else if (NULL == (cur_node = 
          check_lookup_backup(cur_node, s, char_size)))          // failed character match; try backups
            break;                                               //   failure to root
        }
    if (last_match) {                                            // collating sequence match
        str = last_match->matched;                               //   advance to end of match
        in_char_set->advance(str);                               //   advance to next character
        return(last_match->this_character->collation);           //   return collation order
        }
    else                                                         // no collating sequence match
        return(MAX_CHARS_IN_SET);                                //   report failure
}

kmp_node* kmp_node::check_incr_lookup_backup(kmp_node* next_node, unsigned char* s, unsigned long char_size)
{
    kmp_node* last_node;

    last_node = next_node;
    next_node = next_node->get_right();
    if ( ! next_node) {
        next_node = last_node->get_failure();             //     go to last success
        if (( ! next_node) || (this == next_node))        //     match failure
            return NULL;
        }
    return next_node;
}

/* incremental lookup -- call this function repeatedly with a new character at s. Initial call takes */
/*   these argument values: "this" and "next_node" should be root of a KMP FSM, and "s", a pointer to a */
/*   (multibyte) character. On successive calls supply at "s" the next character to be considered. */
/*   The "num_chars" argument need not be initialized, but logically should be initialized to zero. */
/*   Function returns MAX_CHARS_IN_SET unless a successful string match is completed. On a successful */
/*   string match, the function returns the value associated with the successful match, sets "next_node" */
/*   to NULL if no longer match is possible, and returns in "num_chars" the number of characters */
/*   matched. If "next_node" is non-NULL, a match is still possible. The shortest possible match is */
/*   reported, but if a longer match is possible, "next_node" will be non-NULL. Note that successive */
/*   calls may yield a match, but that the length matched may be less than the number of characters */
/*   examined. For example, a KMP FSM that was designed to recognize "asdf" would yield a successful */
/*   match on "asdasdf", but only the last four characters would constitute the match. */
unsigned long kmp_node::incr_lookup(kmp_node*& next_node, unsigned long& num_chars, unsigned char* s)
{
    kmp_node*      found       = NULL;
    char_set*      in_char_set;
    unsigned long  char_size;
    unsigned long  collation   = MAX_CHARS_IN_SET;

    in_char_set = this_character->in_char_set;
    if (NULL == in_char_set)
        in_char_set = get_dflt_union_char_set()->get_sets();
    char_size = in_char_set->get_char_size();

    if (next_node == this)
        next_node = down;
    while (next_node) {
        if (next_node->matches(s)) {                              // successful character match
            ++num_chars;
            if (MAX_CHARS_IN_SET !=
              (collation = next_node->this_character->collation)) //   string match
                found = next_node;                                //     record it
            next_node = next_node->get_down();                    //   try children if any
            break;
            }
        else if (NULL == (next_node = 
          check_incr_lookup_backup(next_node, s, char_size)))     // failed character match
            break;
        }
    if (found)                                                    // string match
        return found->this_character->collation;                  // return value associated with string matched
    else if (next_node)                                           // successful character match or opportunity
        return MAX_CHARS_IN_SET;
    else {
        next_node = NULL;
        num_chars = 0;
        return MAX_CHARS_IN_SET;                                  //   report failure
        }
}

/**************************************************************************************
    utility functions for the Knuth-Morris-Pratt string-searching algorithm
**************************************************************************************/

long cmp_csm_str(csm_string* cs1, csm_string* cs2)
{
    long cmp;
    csm_character* cc1 = cs1->string;
    csm_character* cc2 = cs2->string;

    if (cc1->in_char_set && (cc1->in_char_set == cc2->in_char_set)) {
        while (0 == (cmp = cc1->in_char_set->cs_strcmp(cc1->value, cc2->value))) {
            ++cc1;
            ++cc2;
            }
        return(cmp);
        }
    else {
        while (0 == (cmp = strcmp((char*)(cc1->value),(char*)( cc2->value)))) {
            ++cc1;
            ++cc2;
            }
        return(cmp);
        }
}

class string_coll {
public:
    csm_string*   string;
    unsigned long collation;
};

long string_coll_cmp(string_coll* scp1, string_coll* scp2)
{
    return cmp_csm_str(scp1->string, scp2->string);
}

// Returns 1 when counting number of nodes to be allocated
//   if character j in string (i-1) has a distinct ancestor 
//   from character j in string i.
// Returns 0 otherwise.
// Allows proper count when pairs of strings differ only in
//   their middle portions.
// Private to this file.
long kmp_count_has_distinct_ancestor(csm_string** csm_strings, long i, long j)
{
    for (--j; j >= 0; --j) {
        // the ancestors differ in character value
        if (csm_strings[i]->string[j] != csm_strings[i-1]->string[j])
            return 1;
        }
    return 0;
}

// Returns 1 when making kmp_fsm if character j in string (i-1) 
//   has a distinct ancestor from character j in string i.
// Returns 0 otherwise.
// Allows proper tree construction when pairs of strings differ
//   only in their middle portions.
// Private to this file.
long kmp_make_has_distinct_ancestor(csm_string** csm_strings, long i, long j)
{
    for (--j; j >= 0; --j) {
        // current ancestor is valid
        //   AND differs from its left sibling
        if ((NULL != csm_strings[i]->string[j].get_value())
         && (csm_strings[i]->string[j].get_value() != csm_strings[i-1]->string[j].get_value()))
            return 1;
        }
    return 0;
}

// Return the root of a Knuth-Morris-Pratt finite state machine for the
//   set of strings in the first argument. One-byte characters and zero-byte
//   string terminator assumed if argument for_char_set is 0.
// See Sara Baase, "Computer Algorithms: Introduction to Design and Analysis"
//   (Addison Wesley, 1978) for a detailed explanation of the algorithm.
// Argument strings is an explicit set of strings. This function does not
//   support regular expression syntax.
kmp_node* kmp_fsm(unsigned char** strings, unsigned long num_strings,
                  unsigned long* collations, char_set* for_char_set)
{
    unsigned long  char_size = (for_char_set) ? for_char_set->get_char_size() : 1;
    unsigned long  maxlen;
    unsigned long  i;
    unsigned long  j;
    unsigned long  k;
    unsigned long  m;
    unsigned long  num_nodes;
    unsigned long  len;
    unsigned long* lens;
    unsigned char* string;
    csm_string**   in_csm_strings;
    csm_string**   csm_strings;
    string_coll*   string_colls;
    csm_character  null_char;
    csm_character  last_char;
    csm_character  this_char;
    csm_character* next_char;
    unsigned char* null_bytes;
    kmp_node*      root;
    kmp_node*      last_node;
    kmp_node*      this_node;
    kmp_node*      next_node;
    kmp_node*      parent;
    kmp_node*      failure;
    kmp_node*      failure_link;
    kmp_dir        dir;
    unsigned long  levels;

    // check arguments

    if ((NULL == strings) || (0 == num_strings) || (NULL == collations)) {
        printf("kmp_fsm: bad argument\n");
        return(NULL);
        }

    // allocate temporary storage

    lens = new unsigned long[num_strings];

    // set function pointers

    // get string lengths and maximum length

    maxlen = 0;
    for (i = 0; i < num_strings; i++) {
        if (for_char_set)
            len = for_char_set->cs_strlen(strings[i]);
        else
            len = strlen((char*)(strings[i]));
        if (len > maxlen)
            maxlen = len;
        lens[i] = len;
        }

    // copy the null character for this character set

    if (for_char_set) {
        null_bytes = (unsigned char*)(new char[char_size]);
        for_char_set->get_terminator(null_bytes);
        null_char.init(null_bytes, 1, for_char_set);
        }
    else
        null_char.init((unsigned char*)"", 1);

    // allocate and initialize working space

    in_csm_strings = new csm_string*[num_strings];
    for (i = 0; i < num_strings; i++)
        in_csm_strings[i] = new csm_string;
    for (i = 0; i < num_strings; i++) {
        in_csm_strings[i]->string = new csm_character[lens[i] + 1];
        in_csm_strings[i]->len = lens[i];
        }
    for (i = 0; i < num_strings; i++) {
        string = strings[i];
        for (j = 0; j < lens[i]; j++)
            in_csm_strings[i]->string[j].init(string + (j * char_size), 1, for_char_set);
        in_csm_strings[i]->string[j].init(null_bytes, 1, for_char_set);
        }
    string_colls = new string_coll[num_strings];
    for (i = 0; i < num_strings; i++) {
        string_colls[i].string = in_csm_strings[i];
        string_colls[i].collation = collations[i];
        }

    // sort the string/collation pairs

    tml_os_qsort((char*) string_colls, num_strings, (unsigned long) sizeof(string_coll), (SORT_FUNC) string_coll_cmp);

    // reinitialize the string, collation, and length arrays to sorted order

    csm_strings = new csm_string*[num_strings];
    for (i = 0; i < num_strings; i++) {
        csm_strings[i] = string_colls[i].string;
        collations[i] = string_colls[i].collation;
        lens[i] = string_colls[i].string->len;
        }

    // make the tree of character matches
    // Note: LINE-BY-LINE COMMENTS FOR THIS SECTION BEGIN IN COLUMN 88

    num_nodes = 1;                                                                     // count the root
    for (j = 0; j < maxlen; j++) {                                                     // for each string index
        last_char = null_char;                                                         // no left character for leftmost string
        for (i = 0; i < num_strings; i++) {                                            // for each string
            if (j >= csm_strings[i]->len) {                                            // strings[i] already exhausted
                last_char = null_char;                                                 //   next character at this level if any
                continue;                                                              //     will be new (bit copy)
                }
            this_char = csm_strings[i]->string[j];                                     // this character (bit copy)
            if ((last_char != this_char)                                               //   is a new character at this level
              || (kmp_count_has_distinct_ancestor(csm_strings, i, j))) {               //   OR has distinct ancestor from its left sib
                last_char = this_char;                                                 //      continue comparing with new character
                ++num_nodes;                                                           //      count its node
                }
            }
        }

    root = new kmp_node[num_nodes];
    root->up = NULL;
    root->failure = NULL;
    root->matched = NULL;
    this_node = root;
    next_char = new csm_character[num_nodes];
    root->this_character = next_char++;
    root->this_character->in_char_set = for_char_set;

    for (j = 0; j < maxlen; j++) {                                                     // for each string index
        last_char = null_char;                                                         // no left character for leftmost string (bit copy)
        last_node = NULL;                                                              // no previous node for first character at a given level
        for (i = 0; i < num_strings; i++) {                                            // for each string
            if (j >= csm_strings[i]->len) {                                            // strings[i] already exhausted
                last_char = null_char;                                                 //   next character at this level will be new
                last_node = NULL;                                                      //   next node at this level if any
                continue;                                                              //     will be a leftmost sibling
                }
            this_char = csm_strings[i]->string[j];                                     // this character - bit copy
            if ((last_char != this_char)                                               //   is a new character at this level
              || (kmp_make_has_distinct_ancestor(csm_strings, i,j))) {                 //   OR has distinct ancestor from its left sib
                last_char = this_char;                                                 //     copy it to last_char
                ++this_node;                                                           //     get node for it
                this_node->this_character = next_char++;                               //     get a csm_character for it
                this_node->this_character->init(this_char.value,
                                                this_char.num_chars,
                                                for_char_set);                         //     associate character with node
                if (j == lens[i] - 1)                                                  //     last character in string
                    this_node->this_character->collation = collations[i];              //       set its node's collation_order
                csm_strings[i]->string[j].value = (unsigned char*) this_node;          //     associate node with character (not bit copy)
                if (j == 0) {                                                          //     first character in string
                    if (i == 0)                                                        //       first string
                        root->down = this_node;                                        //         node is oldest child of root
                    else                                                               //       not first string so has left sib
                        last_node->right = this_node;                                  //         node is right sib of last_node
                    this_node->up = root;                                              //       becomes child of root
                    }
                else {                                                                 //     not first character in string
                    if (i == 0) {                                                      //       leftmost string
                        parent = (kmp_node*) csm_strings[0]->string[j-1].value;        //         parent is immediately above
                        parent->down = this_node;                                      //       associate child with parent
                        }
                    else {                                                             //       not leftmost string
                        m = j - 1;                                                     //         starting on level above
                        for (k = i; k >= 0; k--) {                                     //         potential parents, right to left
                            if (NULL != csm_strings[k]->string[m].value) {             //           parent found (NOTE: a parent will
                                parent = (kmp_node*) csm_strings[k]->string[m].value;  //           be found before an exhausted
                                break;                                                 //           string is encountered)
                                }
                            }
                        if ((k == i)                                                   //         leftmost sibling, parent above
                          || ((NULL == parent->down)                                   //         OR first "possible longer match" 
                          && (MAX_CHARS_IN_SET != parent->this_character->collation))) //         (e.g. "prefix" and "prefixsuffix")
                            parent->down = this_node;                                  //           associate child with parent
                        else if (last_node)                                            //         not leftmost, parent above and left
                            last_node->right = this_node;                              //           associate left node with right sib
                        }
                    this_node->up = parent;                                            //       associate parent with child
                    }
                last_node = this_node;                                                 //         new rightmost sibling
                }
            else                                                                       //   is same as previous character at this level
                csm_strings[i]->string[j].value = NULL;                                //     and shares its parent, so mark it as having
                                                                                       //     no node of its own (not bit copy)
            }
        }

    // set failure links

    next_node = root->down;                                                        // starting at root's first child
    dir = kmp_right;                                                               //   breadth first order
    levels = 0;
    while (maxlen >= levels) {                                                     //     while nodes remain
        if (root == next_node->up)                                                 //       failure to match a child of root
            next_node->failure = root;                                             //         is failure to match anything
        else {                                                                     //       not child of root
            failure = next_node->up->failure;                                      //         start with failure node of parent
            while ((failure) &&                                                    //         while failure non-NULL, failure
              (failure != root) &&                                                 //           not root and current character
              (*failure->this_character != *next_node->up->this_character))        //           doesn't match character at
                failure = failure->failure;                                        //           failure node, use failure's failure
            failure_link = next_node->up;
            while (failure_link->up != failure)                                    // find parent of failure node
                failure_link = failure_link->up;
            next_node->failure = failure_link;                                     // set failure link of this node
            }
        next_node = next_node->breadth_first(dir, levels);
        while ((kmp_right != dir) && (maxlen >= levels)) {                         // while next node in breadth first order not reached
            next_node = next_node->breadth_first(dir, levels);                          //   and rightmost node at deepest level not processed
            }
        }

    // delete temporary storage

    for (i = 0; i < num_strings; i++) {
        in_csm_strings[i] = NULL;
        string_colls[i].string = NULL;
        delete csm_strings[i]->string;
        csm_strings[i]->string = NULL;
        delete csm_strings[i];
        }
    delete in_csm_strings;
    delete string_colls;
    delete csm_strings;
    delete null_bytes;

    // return the finite state machine

    return(root);
}

