/**************************************************************************************
 
    mark_list_traits.cxx
    functions for the class containing the list of marks associated with a
      particular model in the text management library modelling component

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <tml/tml_objects.hxx>
#include <tml/tml_functions.hxx>
#include <tml/tml_os_depends.hxx>

/*constructor for a mark list item*/
mark_list::mark_list()
{
    next = prior = 0;
    mark_range = 0;
    is_visible = 0;
}

/*destructor for a mark list item*/
mark_list::~mark_list()
{
    next = prior = 0;
    delete mark_range;
    mark_range = 0;
}

/*set the model range of a mark to the range specified. Need to allocate
    our own space so the data being pointed to doesn't go away on us*/
model_range* mark_list::set_range(model_range new_range)
{
    model_range   hold_span;
    model_pos     range_start,
                        range_end;

    /*first, check for trash*/
    if (!this || !new_range.get_start().in_node || !new_range.get_end().in_node)
        return 0;

    /*make sure that the mark makes sense*/
    if (new_range.get_start().in_node->get_model() != new_range.get_end().in_node->get_model())
        return 0;

    /*allocate the space for the range in the mark_list*/
    mark_range = new model_range;
    if (!mark_range)
        return 0;
    /*make sure that the range points to terminals*/
    hold_span = new_range.get_start().in_node->get_span();
    if (hold_span.get_start().in_node != new_range.get_start().in_node)
        range_start = hold_span.get_start();
    else
        range_start = new_range.get_start();

    hold_span = new_range.get_end().in_node->get_span();
    if (hold_span.get_start().in_node != new_range.get_end().in_node)
        range_end = hold_span.get_end();
    else
        range_end = new_range.get_end();

    mark_range->set_start(range_start);
    mark_range->set_end(range_end);

    return mark_range;
}

/*extends a mark in the model. If the mark is currently drawn (in all viewports)
    the viewports will be updated*/
long mark_list::extend_mark(model_pos new_start, model_pos new_end)
{
    register long  return_val = 1;

    model_pos      old_start,
                   old_end,
                   lesser,
                   greater,
                   correct_start,
                   correct_end;
    mf_list       *in_formats;
    fv_list       *in_viewports;
    view_port     *cur_viewport;
    draw_mark     *viewport_mark;

    /*first, check for garbage in*/
    if ((!this) || (!new_start.in_node) || (!new_end.in_node))
        return 0;

    /*get old start and end in correct order*/
    old_start = mark_range->get_start();
    old_end = mark_range->get_end();
    if (old_end < old_start) {
        old_end = mark_range->get_start();
        old_start = mark_range->get_end();
        }

    /*get new_start and new_end in correct order*/
    if (new_end < new_start) {
        correct_start = new_end;
        greater = new_start;
        lesser = correct_start;
        }
    else {
        lesser = new_start;
        greater = new_end;
        }

    /*check to make sure that we aren't trying to span models*/
    if ((old_start.in_node->get_model() != new_start.in_node->get_model())
        || (old_start.in_node->get_model() != new_end.in_node->get_model()))
        return 0;

    /*if the mark isn't visible, just update the start and end*/
    if (!is_visible){
        mark_range->set_start(new_start);
        mark_range->set_end(new_end);
        return 1;
        }

    /*it is visible, so we need to (conceptually) erase the old mark in all
        viewports, update the range of the mark, and then draw the updated mark
        in all viewports but we minimize redrawing */

    correct_start = old_start;
    correct_end = old_end;

    for (in_formats = mark_range->get_start().in_node->get_model()->get_formats();
      in_formats;
      in_formats = in_formats->inq_next()){
        for (in_viewports = in_formats->inq_format()->get_fv_list();
          in_viewports;
          in_viewports = in_viewports->inq_next()){
            cur_viewport = in_viewports->inq_vp();
            /*see if the mark is already visible in the viewport*/
            for (viewport_mark = cur_viewport->get_marks();
              viewport_mark && (viewport_mark->get_model_mark() != this);
              viewport_mark = viewport_mark->get_next())
                ;

            // it's there
            if (viewport_mark) {

                // erase what must be erased, draw what must be drawn

                // new mark precedes old mark or old mark precedes new mark
                if (((lesser < old_start) && (greater < old_start))
                  || ((lesser > old_end) && (greater > old_end))) {
                    // erase old mark
                    if (0 == (*(viewport_mark->get_draw_func()))(cur_viewport, 
                                                                 viewport_mark->get_core_func(),
                                                                 old_start, old_end, -1))
                        return_val = 0;
                    // draw new mark
                    if (0 == (*(viewport_mark->get_draw_func()))(cur_viewport, 
                                                                 viewport_mark->get_core_func(),
                                                                 lesser, greater))
                        return_val = 0;
                    // remember correct range
                    correct_start = lesser;
                    correct_end = greater;
                    continue;
                    }

                // new mark shrinks old mark at beginning - erase and update beginning of range
                if (lesser > old_start) {
                    if (0 == (*(viewport_mark->get_draw_func()))(cur_viewport, 
                                                                 viewport_mark->get_core_func(),
                                                                 old_start, lesser.get_prior(), -1))
                        return_val = 0;
                    correct_start = lesser;
                    }

                // new mark shrinks old mark at end - erase and update end of range
                if (greater < old_end) {
                    if (0 == (*(viewport_mark->get_draw_func()))(cur_viewport,
                                                                 viewport_mark->get_core_func(),
                                                                 greater.get_next(), old_end, -1))
                        return_val = 0;
                    correct_end = greater;
                    }

                // new mark extends old mark at beginning - draw and update beginning of range
                if (lesser < old_start) {
                    if (0 == (*(viewport_mark->get_draw_func()))(cur_viewport, 
                                                                 viewport_mark->get_core_func(),
                                                                 lesser, old_start.get_prior()))
                        return_val = 0;
                    correct_start = lesser;
                    }

                // new mark extends old mark at end - draw and update end of range
                if (greater > old_end) {
                    if (0 == (*(viewport_mark->get_draw_func()))(cur_viewport,
                                                                 viewport_mark->get_core_func(),
                                                                 old_end.get_next(), greater))
                        return_val = 0;
                    correct_end = greater;
                    }
                }
            }
        }

    /*ensure the range cites correct values in same relative order as mark before extension*/
    /*except where new end and old end are on opposite sites of old start*/
    /*in that case, use opposite of old order*/
    /*these provisions make the code more robust because they make the order of*/
    /* the arguments relative to the order in the mark immaterial*/
    if (mark_range->get_start() <= mark_range->get_end()) {
        if (new_start <= new_end) {
            mark_range->set_start(correct_start);
            mark_range->set_end(correct_end);
            }
        else {
            mark_range->set_start(correct_end);
            mark_range->set_end(correct_start);
            }
        }
    else {
        if (new_start > new_end) {
            mark_range->set_start(correct_end);
            mark_range->set_end(correct_start);
            }
        else {
            mark_range->set_start(correct_start);
            mark_range->set_end(correct_end);
            }
        }

    /*the mark is drawn*/
    is_visible = return_val;

    return return_val;
}
                                                                                                                                                   
/*draw the mark in all viewports associated with all formats associated with the model*/
long mark_list::draw()
{
    register mf_list   *in_formats;
    register fv_list   *in_viewports;
    register view_port  *cur_viewport;
    register draw_mark *viewport_mark;
    register draw_mark *new_mark;

    /*first, check to protect from garbage*/
    if (!this)
        return 0;

    /*if the mark is already visible, just return success*/
    if (is_visible)
        return 1;
                                           
    /*find the formats associated with the model, and the viewports associated
        with those formats. See if the mark is already in those viewports; if not,
        add them*/
    for (in_formats = mark_range->get_start().in_node->get_model()->get_formats();
            in_formats;
            in_formats = in_formats->inq_next()){
        for (in_viewports = in_formats->inq_format()->get_fv_list();
                in_viewports;
                in_viewports = in_viewports->inq_next()){
            cur_viewport = in_viewports->inq_vp();
            /*see if the mark is already visible in the viewport*/
            for (viewport_mark = cur_viewport->get_marks();
                    viewport_mark && (viewport_mark->get_model_mark() != this);
                    viewport_mark = viewport_mark->get_next())
                    ;
                if (!viewport_mark){     //it's not already there
                    new_mark = new draw_mark;
                    new_mark->set_model_mark(this);
                    cur_viewport->add_mark(new_mark);
                    }
            }
        }

    is_visible = 1;
    return 1;
}
                       
/*erase the mark in all viewports assoicated with all formats associated with the model*/
long mark_list::erase()
{
    register mf_list   *in_formats;
    register fv_list   *in_viewports;
    register view_port  *cur_viewport;
    register draw_mark *viewport_mark;

    /*first, check to protect from garbage*/
    if (!this)
        return 0;

    /*if the mark isn't visible, just return success*/
    if (!is_visible)
        return 1;
                                           
    /*find the formats associated with the model, and the viewports associated
        with those formats. See if the mark is in those viewports; if so,
        delete them*/
    for (in_formats = mark_range->get_start().in_node->get_model()->get_formats();
            in_formats;
            in_formats = in_formats->inq_next()){
        for (in_viewports = in_formats->inq_format()->get_fv_list();
                in_viewports;
                in_viewports = in_viewports->inq_next()){
            cur_viewport = in_viewports->inq_vp();
            /*see if the mark is already visible in the viewport*/
            for (viewport_mark = cur_viewport->get_marks();
                    viewport_mark && (viewport_mark->get_model_mark() != this);
                    viewport_mark = viewport_mark->get_next())
                    ;
                if (viewport_mark)      //it's there
                    cur_viewport->detach_mark(viewport_mark);
            }
        }

    is_visible = 0;
    return 1;
}
                       

