/**************************************************************************************
    node_traits.cxx
    functions for the node classes in the text management library
      modelling component

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <tml/tml_objects.hxx>
#include <tml/tml_functions.hxx>
#include <tml/tml_os_depends.hxx>

const unsigned long def_tab_tbl_size = 20;
const unsigned long def_tab_pixels = 20;

/*constructor for the class node--just makes sure that the node is not associated
    with a model*/
node::node(long node_size, model *to_model)
{
    register node  *new_node;

    if (!this){
        if (to_model){
            new_node = to_model->get_current_oheap()->get_tf_node(node_size);
            if (new_node){
                if (new_node->up){
                    if (new_node->left)
                        new_node->left->right = new_node->up;
                    else
                        to_model->get_current_oheap()->tf_node = new_node->up;
                    new_node->up->right = new_node->right;
                    if (new_node->right)
                        new_node->right->left = new_node->up;
                    }
                else {
                    if (new_node->left)
                        new_node->left->right = new_node->right;
                    else
                        to_model->get_current_oheap()->tf_node = new_node->right;
                    if (new_node->right)
                        new_node->right->left = new_node->left;
                    }
                }
            else {
                new_node = (node *)(to_model->get_current_oheap()->alloc(node_size));
                if (!new_node){
                    return;
                    }
                }
            this = new_node;
            }
        else {
            new_node = (node *) (new char [(long) sizeof(node)]);
            if (!new_node)
                return;
            else
                this = new_node;
            }
        }
    else
        this = this;

    in_model = to_model;
    n_type = undefined_node;
    up = right = left = 0;
    properties = 0;
    sib_order = 1;
    node_plid_trace = 0;
}

/*destructor for the class node*/
node::~node()
{
    detach();
    prop_detach();
    n_type = is_free;
    if (in_model){
        sib_order = get_osize();
        in_model->get_current_oheap()->free(this);
        this = 0;
        }
}

/*Find the level of node 'this' in the tree by
counting number of nodes above 'this'.*/
long node::get_level()
{
    register node *temp;
    long level = 0;

    for(temp = this; temp->up; temp = temp->up)
        level++;

    return(level);
}
                               
/*update up and sib_order of right siblings */
/*  "this" gets sib_order (new_left->sib_order + 1) if new_left */
/*     is not NULL, or sib_order of 1 otherwise. Right siblings */
/*     get next higher sib_order in turn. */
/*  "this" and all right siblings get new_up as their new */
/*     parent pointer. */
/*in bnode::split(), allows a new node to adopt a vector of */
/*  children formerly belonging to its left node - cannot */
/*  use new_node->connect_node(first_child, n_down) there */
/*  because ::connect_node() would cause adoption of first */
/*  child only */
/*in node::abdicate(), reorders succession after abdication */
void node::update_sib_vec(node* new_up, node* new_left)
{
    register node* hold_node;
    register long  order = 0;

    history  *past;
    atom*     this_atom;
    atom_arg *arg1, *arg2, *arg3, *arg4, *arg5;
                                 
    if (NULL != get_model()) {
        past = get_model()->get_past();

        // if history is active, record this call

        if (past->has_hist() && past->is_active()) {
            arg1 = new atom_this_arg(past, this);
            arg2 = new atom_val_arg(past, up);
            arg3 = new atom_val_arg(past, left);
            arg4 = new atom_val_arg(past, new_up);
            arg5 = new atom_val_arg(past, new_left);
            this_atom = new atom(past, (WHOLE_PTR) &node::update_sib_vec,
                                       (WHOLE_PTR) &node::update_sib_vec,
                                       arg1, arg2, arg3, NULL,
                                       arg1, arg4, arg5, NULL);
            past->add_atom(this_atom);
            }
        }

    if (new_left) {
        order = new_left->sib_order;
        left = new_left;
        }
    for (hold_node = this; hold_node; hold_node = hold_node->right) {
        hold_node->up = new_up;
        hold_node->sib_order = ++order;
        }
}

/*return a model range which contains all and only the model positions
    dominated by this node*/
model_range node::get_span()
{
    model_range  return_range;
    model_pos    start_pos;
                                       
    if (!this) return return_range;

    start_pos.in_node = this;
    start_pos.char_index = -1;
    return_range.set_start(start_pos.get_next());
    if (return_range.get_start().in_node == this){
        start_pos.char_index = 0;
        return_range.set_start(start_pos);
        start_pos.char_index = get_max_index();
        return_range.set_end(start_pos);
        return return_range;
        }

    start_pos.in_node = get_last();
    start_pos.char_index = start_pos.in_node->get_max_index();
    return_range.set_end(start_pos);

    return return_range;
}
 
/*returns 1 if this is an ancestor of target; otherwise returns 0*/
long node::is_ancestor(node *target)
{
    register node  *hold_node;
                           
    /*guard against trash coming in*/
    if (!this || !target)
        return 0;
    /*see if we hit this in a traversal up the tree*/
    for (hold_node = target; hold_node; hold_node = hold_node->up)
        if (hold_node == this)
            return 1;

    /*didn't hit this; so this is not an ancestor of target*/
    return 0;
}

/*****************************************************************************************************
        
        Functions for placing and moving nodes in a tree

*****************************************************************************************************/

/*connect a node to a tree*/
node* node::connect_node(node *to_connect, node_dir dir)
{
    register node  *return_node,
                   *inc_node;

    history  *past;
    atom     *this_atom;
    atom_arg *arg1, *arg2, *arg3;
                                 
    // If to_connect not in same model as this, copy to_connect to the model
    //  "this" is in - NOTE: This block must precede block setting history

    if (in_model != to_connect->in_model)
        return_node = to_connect->copy(in_model);
    else
        return_node = to_connect;

    // NOTE: This block must follow block copying argument to_connect to
    //   same model as "this" - history is recorded for return_node

    if (NULL != get_model()) {
        past = get_model()->get_past();

        // if history is active, record this call

        if (past->has_hist() && past->is_active()) {
            arg1 = new atom_this_arg(past, return_node);
            arg2 = new atom_this_arg(past, this);
            arg3 = new atom_val_arg(past, (long)dir);
            if (n_up != dir)
                this_atom = new atom(past, (WHOLE_PTR) &node::detach, (WHOLE_PTR) &node::connect_node,
                                     arg1, NULL, arg2, arg1, arg3, NULL);
            else
                this_atom = new atom(past, (WHOLE_PTR) &node::abdicate, (WHOLE_PTR) &node::connect_node,
                                     arg1, arg2, NULL, arg2, arg1, arg3, NULL);
            past->add_atom(this_atom);
            }
        }

     switch (dir){
        case n_up:
        /*make the return_node the new parent of this*/
            if ((return_node->n_type == leaf) || (return_node->n_type == gobj)){
                printf("error in node::connect_node--attempt to connect leaf to a daughter\n");
                break;
                }

            /*if this is an ancestor of the current property position, set the property position
                to the parent of this*/
            if (get_model())
                if (is_ancestor(get_model()->get_prop_pos()))
                    get_model()->move_prop(up);

            return_node->up = up;
            if (up) {
                if (up->get_node(n_down) == this)
                    up->set_node(return_node, n_down);
                }
            return_node->right = right;
            if (right)
                right->left = return_node;
            right = 0;
            return_node->left = left;
            if (left)
                left->right = return_node;
            left = 0;
            return_node->set_node(this, n_down);
            up = return_node;
            return_node->sib_order = sib_order;
            sib_order = 1;
            break;

        case n_right:
        /*make the return_node the right sibling of this*/
            return_node->up = up;
            return_node->right = right;
            if (right)
                right->left = return_node;
            right = return_node;
            return_node->left = this;
            return_node->sib_order = sib_order + 1;
            for (inc_node = return_node->right; inc_node; inc_node = inc_node->right)
                inc_node->sib_order++;
            break;

        case n_left:
        /*make the return_node the left sibling of this*/
            return_node->up = up;
            return_node->left = left;
            if (left)
                left->right = return_node;
            else if (up)
                up->set_node(return_node, n_down);
            left = return_node;
            return_node->right = this;
            return_node->sib_order = sib_order;
            for (inc_node = this; inc_node; inc_node = inc_node->right)
                inc_node->sib_order++;
            break;

        case n_down:
        /*make the return_node the child of this*/
            if ((n_type == leaf) || (n_type == gobj)) {
                printf("error in node::connect_node, attempt to connect daughter to leaf\n");
                break;
                }

            /*if this is an ancestor of the current property position, set the property position
                to this*/
            if (get_model())
                if (is_ancestor(get_model()->get_prop_pos()))
                    get_model()->move_prop(this);

            return_node->right = get_node(n_down);
            if (return_node->right)
                return_node->right->left = return_node;
            set_node(return_node, n_down);
            return_node->up = this;
            return_node->sib_order = 1;
            for (inc_node = return_node->right; inc_node; inc_node = inc_node->right)
                inc_node->sib_order++;
            break;
        }
    return return_node;
}

/*take a node out of a tree*/
void node::detach()
{
    register mark_list   *m_list_item;
    register model_range *marked_range;
    model_range           this_span;
    model_pos             range_start,
                          range_end,
                          hold_pos,
                          this_span_start,
                          this_span_end;
    register mf_list     *form_list;
    register format      *this_format;
    register fv_list     *view_list;
    register view_port   *this_viewport;
    register v_cursor    *this_cursor;
    register node        *hold_node;
    register draw_mark   *d_list_item;
    register draw_mark   *d_list_next;

    history  *past;
    node_dir  dir;
    long      unconnected = 0;
    node     *connection;
    atom     *this_atom;
    atom_arg *arg1, *arg2, *arg3;

    /*if this is an ancestor of the current property position, set the property position
        to the parent of this*/
    if (get_model())
        if (is_ancestor(get_model()->get_prop_pos()))
            get_model()->move_prop(up);

    if (NULL != get_model()) {
        past = get_model()->get_past();

        // if history is active, record this call

        if (past->has_hist() && past->is_active()) {
            if (left) {
                connection = left;
                dir = n_right;
                }
            else if (right) {
                connection = right;
                dir = n_left;
                }
            else if (up) {
                connection = up;
                dir = n_down;
                }
            else if (get_node(n_down)) {
                connection = this->get_node(n_down);
                dir = n_up;
                }
            else
                unconnected = -1;

            if (!unconnected) {
                arg1 = new atom_this_arg(past, connection);
                arg2 = new atom_this_arg(past, this);
                arg3 = new atom_val_arg(past, (long)dir);
                this_atom = new atom(past, (WHOLE_PTR) &node::connect_node, (WHOLE_PTR) &node::detach,
                                     arg1, arg2, arg3, NULL, arg2, NULL);
                past->add_atom(this_atom);
                }
            }

        /*update any cursor positions which may have been changed by the detach*/
        /*update any viewport marks and associated model marks. Note that each draw_mark */
        /*  object on a viewport's list of draw_marks refers through its mark_list */
        /*  member model_mark to a mark_list object that is also queued on the model */
        this_span = get_span();
        this_span_start = this_span.get_start();
        this_span_end = this_span.get_end();
        for (form_list = in_model->get_formats(); form_list; form_list = form_list->inq_next()){
            this_format = form_list->inq_format();
            if (this_format){
                for (view_list = this_format->get_fv_list(); view_list; view_list = view_list->inq_next()){
                    this_viewport = view_list->inq_vp();
                    if (!this_viewport) continue;
                    /* update the list of marks in the viewports as necessary */
                    for (d_list_item = this_viewport->get_marks(); d_list_item; d_list_item = d_list_next) {
                        d_list_next = d_list_item->get_next();
                        m_list_item = d_list_item->get_model_mark();
                        marked_range = m_list_item->get_range();
                        range_start = marked_range->get_start();
                        range_end = marked_range->get_end();
                        /* if the model_mark is totally contained in the span of this node, delete it */
                        if ((range_start >= this_span_start) && (range_end <= this_span_end)) {
                            this_viewport->detach_mark(d_list_item);
                            in_model->delete_mark(m_list_item);
                            }
                        /*if the beginning of the range is in the span of this node, change the
                            beginning of the mark to the first character of the next terminal*/
                        else if (this_span.contains(range_start)){
                            hold_pos = this_span_end;
                            hold_pos = hold_pos.get_next();
                            marked_range->set_start(hold_pos);
                            }
                        /*if the end of the range is in the span of this node, change the
                            end of the mark to the last character of the previous terminal*/
                        else if (this_span.contains(range_end)){
                            hold_pos = this_span_start;
                            hold_pos = hold_pos.get_prior();
                            marked_range->set_end(hold_pos);
                            }
                        }
                    this_cursor = this_viewport->get_cursor();
                    if (!this_cursor) continue;
                    range_start = this_cursor->get_model_pos();
                    if (this_span.contains(range_start)){
                        range_start = this_span_start.get_prior();
                        if (!range_start.in_node)
                            range_start = this_span_end.get_next();
                        this_cursor->set_model_pos(range_start);
                        }
                    }
                }
            }
        }       

    /*decrement the sibling order for all nodes to the right of this one*/
    for (hold_node = right; hold_node; hold_node = hold_node->right)
        hold_node->sib_order--;

    if (right)
        right->left = left;
    if (left)
        left->right = right;
    if (up != NULL) {
        if (up->get_node(n_down) == this)
            up->set_node(right, n_down);
        }
    right = left = up = 0;
}

/* undo function for node::connect_node(..., n_up) */
/* replace a parent with a vector of its left siblings, heir, and right siblings */
void node::abdicate(node* heir)
{
    node* hold_node;

    if (this != heir->up) {
        printf("node::abdicate: heir is not my child\n");
        return;
        }
    else if (NULL == heir) {
        printf("node::abdicate: no heir\n");
        return;
        }

    /* abdicate in favor of heir */
    heir->right = right;
    if (right)
        right->left = heir;
    heir->left = left;
    if (left)
        left->right = heir;
    for (hold_node = heir; hold_node->left; hold_node = hold_node->left)
        ;
    up->set_node(hold_node, n_down);

    /* reorder the succession */
    hold_node->update_sib_vec(up, NULL);
}

/*move a node, perhaps from one model to another, and return a pointer to the node moved*/
node* node::move(node *move_to, node_dir dir)
{
    register node *moved_node;

    history *past;

    if (NULL != get_model())
        past = get_model()->get_past();
    detach();
    moved_node = move_to->connect_node(this, dir);
    if (moved_node != this) {                      // if connect_node() copied this ...
        if (NULL != get_model()) {
            if (0 == past->has_hist())             // .. and history not enabled,
                delete this;                       //   delete this as there will be no
            }                                      //   further reference to it
        }
    return moved_node;
}
   
/*copy a node--note--this should never be used, as the derived classes should have their
    own version of this function*/
node* node::copy(model *to_model)
{
    register node *return_node;

    if (!to_model)
        to_model = in_model;
    return_node = new node((long) sizeof(node), to_model);
    if (!return_node) return 0;
    if (properties)
        return_node->properties = properties->copy_all(to_model);
    else
        properties = 0;
    return_node->n_type = n_type;

    return return_node;
}

/*duplicate a node*/
node* node::dup()
{
    register node *return_node;

    return_node = new node((long) sizeof(node), get_model());
    if (!return_node) return 0;

    return_node->properties = properties;
    return_node->n_type = n_type;

    return return_node;
}

/*split a node -- base class implementation of virtual function   */
/*  examines the node being split for plid traces and copies them */
/*  to new node where appropriate                                 */
node* node::split(long split_point)
{
    return this;
}

/*****************************************************************************************************
        
        Navigation functions for the tree

*****************************************************************************************************/
                                                      
/*navigation of the tree--get the next node in traversal order. Incoming node_dir is the direction
you want to move, outgoing node_dir is the direction actually moved*/
node *node::get_next(node_dir& dir)
{
    switch (dir) {
        case n_up:
            if (right){
                dir = n_down;
                return(right);
                }
            else return(up);
                     
        case n_right:
        case n_left:
            printf("bad direction in get_next\n");
            return(this);

        case n_down:
            if (right)
                return(right);
            else {
                dir = n_up;
                return(up);
                }
        }
}
                                                                                                  
/*navigation of the tree--get the previous node in traversal order. Incoming node_dir is the direction
    you want to move, outgoing node_dir is the direction actually moved*/
node *node::get_prev(node_dir& dir)
{

    switch (dir){
        case n_up:
            if (left) {
                dir = n_down;
                return(left);
                }
            else return(up);
            
        case n_right:
        case n_left:
            printf("error in node::get_prev--bad direction\n");
            return(this);

        case n_down:
            if (left)
                return(left);
            else {
                dir = n_up;
                return(up);
                }
        }
}
                                                                                                      
/*navigation of the tree--get the node in the node_dir direction from this*/
node* node::get_node(node_dir dir)
{
    switch(dir){
        case n_up:
            return(up);

        case n_right:
            return(right);

        case n_left:
            return(left);

        case n_down:
                return(0);
        }
}
                   
/*set the node field indicated to the supplied node*/
/*  NOTE: no history should be added for this function. It is */
/*    most often called by functions that write their own history */
/*    in circumstances where recording of its history would disturb */
/*    the undo and redo functions for the function that called it. */
/*    Where necessary (e.g. bnode::split() and model::make_root()) */
/*    record an atom of history for a call to node::set_node() that */
/*    must be recorded. */
void node::set_node(node *to_node, node_dir dir)
{
    switch(dir){
      case n_up:
        up = to_node;
        break;

      case n_right:
        right = to_node;
        break;

      case n_left:
        left = to_node;
        break;

      case n_down:
        break;
    }
}

/*****************************************************************************************************
        
        functions dealing with properties of nodes

*****************************************************************************************************/
/*attach a property vector to a node*/
void node::prop_attach(prop_vec *add_pvec)
{
    register prop_vec *new_props;

    history  *past;
    long      has_hist = 0;
    long      is_active = 0;
    atom     *this_atom;
    atom_arg *arg1, *arg2;

    /*if this is an ancestor of the current property position, set the property position
        to the parent of this*/
    if (get_model())
        if (is_ancestor(get_model()->get_prop_pos()))
            get_model()->move_prop(up);

    // if add_pvec is not in the same model "this" is in, copy add_pvec to
    //   same model "this" is in  - NOTE: This block must precede block
    //   setting history

    if (0 == add_pvec)
        new_props = 0;
    else {
        if (in_model == add_pvec->in_model)
            new_props = add_pvec;
        else
            new_props = add_pvec->copy_all(in_model);
        }

    if (properties)
        properties->append(new_props);
    else
        properties = new_props;

    // NOTE: This block must follow copy of add_pvec to same model as "this"
    //   so history will be set properly

    if (NULL != get_model()) {
        past = get_model()->get_past();

        // if history is active, record this call

        if ((has_hist = past->has_hist()) && (is_active = past->is_active())) {
            arg1 = new atom_this_arg(past, this);
            arg2 = new atom_ptr_arg(past, new_props);
            this_atom = new atom(past, (WHOLE_PTR) &node::prop_detach, (WHOLE_PTR) &node::prop_attach,
                                 arg1, NULL, arg1, arg2, NULL);
            past->add_atom(this_atom);
            }
        }

    if (is_active || (0 == has_hist)) {
        for( ; new_props != 0; new_props = new_props->next)
            new_props->ref_count++;
        }
}

/*detach a property vector from a node*/
void node::prop_detach()
{
    history  *past;
    long      has_hist = 0;
    atom     *this_atom;
    atom_arg *arg1, *arg2;

    register prop_vec   *hold_vec;

    /*if this is an ancestor of the current property position, set the property position
        to the parent of this*/
    if (get_model())
        if (is_ancestor(get_model()->get_prop_pos()))
            get_model()->move_prop(up);

    if (NULL != get_model()) {

        // if history is active, record this call

        past = get_model()->get_past();
        if ((has_hist = past->has_hist()) && past->is_active()) {
            arg1 = new atom_this_arg(past, this);
            arg2 = new atom_ptr_arg(past, properties);
            this_atom = new atom(past, (WHOLE_PTR) &node::prop_attach, (WHOLE_PTR) &node::prop_detach,
                                 arg1, arg2, NULL, arg1, NULL);
            past->add_atom(this_atom);
            }
        }

    if (NULL != properties) {
        if (0 == has_hist) {
            for (hold_vec = properties; hold_vec; hold_vec = hold_vec->next)
                hold_vec->ref_count--;
            if (!properties->ref_count)
                delete properties;
            }
        properties = 0;
        }
}

/*activate all of the properties for this node*/
void node::enter_props()
{
    if (properties)
        properties->activate();
}

/*deactivate all of the properties for this node*/
void node::exit_props()
{
    if (properties)
        properties->deactivate();
}

/*****************************************************************************************************
        
        functions which alter the data dominated by a node--used for input
        Note--all of these return an error if the node being altered is not a terminal node

*****************************************************************************************************/
/*add data*/
long    node::attach_data(void *data_start, long data_count)
{
    return 0;
}

/*insert*/
model_pos node::insert(long after_index, char *add_data, long add_count, char_set* ins_char_set)
{
    model_pos return_pos;

    return_pos.in_node = 0;
    return_pos.char_index = -1;
    return return_pos;
}

/*delete*/
model_pos node::del_cont(long delete_index, long delete_num)
{
    model_pos return_pos;

    return_pos.in_node = 0;
    return_pos.char_index = -1;
    return return_pos;
}

/*overstrike*/
model_pos node::overstrike(long begin_index, char *rep_with, long over_num)
{
    model_pos return_pos;

    return_pos.in_node = 0;
    return_pos.char_index = -1;
    return return_pos;
}

/* do-nothing virtual function in base class; made out-of-line so compiler */
/*   won't complain that argument with_node is not used */
long node::combine_data(node* with_node)
{
    return(0);
}

/*****************************************************************************************************

        Formatting functions

*****************************************************************************************************/

/*get the height of the formatted node--in this case, it is always 0*/
long node::get_format_hi(format *in_format)
{
    return 0;
}

/*format a node, returning the cline. In this case, always return null (base class nodes can
    not be formatted)*/
cline* node::format_line(format *in_format, long start_index, long space_hi, long space_wide, long first_concep_pixel)
{
    return 0;
}
           
cline* node::tamrof_line(format *in_format, long start_index, long space_hi, long space_wide, long first_concep_pixel)
{
    return 0;
}

/*****************************************************************************************************

        Member functions for derived class bnode

*****************************************************************************************************/

/*constructor for objects of type bnode*/
bnode::bnode(model *to_model, long size) : (size ? size : (long) sizeof(bnode), to_model)
{
    register bnode    *new_node;

    size = size ? size : sizeof(bnode);

    if (!this){
        if (to_model){
            new_node = (bnode *)(to_model->get_current_oheap()->get_tf_node(size));
            if (new_node){
                if (new_node->up){
                    if (new_node->left)
                        new_node->left->set_right(new_node->up);
                    else
                        to_model->get_current_oheap()->tf_node = new_node->up;
                    new_node->up->set_right( new_node->get_right());
                    if (new_node->right)
                        new_node->right->set_left(new_node->up);
                    }
                else {
                    if (new_node->left)
                        new_node->left->set_right(new_node->right);
                    else
                        to_model->get_current_oheap()->tf_node = new_node->right;
                    if (new_node->right)
                        new_node->right->set_left(new_node->left);
                    }
                }
            else {
                new_node = (bnode *)(to_model->get_current_oheap()->alloc(size));
                if (!new_node){
                    return;
                    }
                }
            this = new_node;
            }
        else {
            new_node = (bnode *) (new char [size]);
            if (!new_node)
                return;
            else
                this = new_node;
            }
        }
    else
        this = this;

    n_type = bobj;
    down_ptr = 0;
}

/*split a branch node--i.e., create a new branch as the parent for all nodes under this
    from split_point on*/
node* bnode::split(long split_point)
{
    register bnode *new_node;
    register node  *first_node;

    history  *past;
    atom*     this_atom;
    atom_arg *arg1, *arg2, *arg3, *arg4;
                                 
    // first, find the boundary node

    first_node = down_ptr;   // need this assignment and check since
    if (first_node) {        //   loop test examines member
        for(;
            first_node && (first_node->get_sib_order()) != split_point;
            first_node = first_node->get_right())
            ;
        }
    if (!first_node) return 0;

    //make a duplicate of the node being split (will share properties)

    new_node = (bnode*) dup();

    if (!new_node) return 0;

    // make the new node the right sibling of the node being split

    new_node = (bnode*) connect_node(new_node, n_right); // connect_node records its own history

    // make the split point node the first child of the new node,
    //   and disconnect it from the children of the node being split

    if (NULL != get_model()) {
        past = get_model()->get_past();
        if (past->has_hist() && past->is_active()) {

            // if history is active, record this call

            arg1 = new atom_this_arg(past, new_node);
            arg2 = new atom_val_arg(past, new_node->down_ptr);
            arg3 = new atom_val_arg(past, first_node);
            arg4 = new atom_val_arg(past, (long)n_down);
            this_atom = new atom(past, (WHOLE_PTR) &node::set_node,
                                       (WHOLE_PTR) &node::set_node,
                                       arg1, arg2, arg4, NULL, arg1, arg3, arg4, NULL);
            past->add_atom(this_atom);

            if (first_node->get_left()) {
                arg1 = new atom_this_arg(past, first_node->get_left());
                arg2 = new atom_val_arg(past, first_node->get_left()->get_right());
                arg3 = new atom_val_arg(past, (node*) NULL);
                arg4 = new atom_val_arg(past,(long)n_right);
                this_atom = new atom(past, (WHOLE_PTR) &node::set_node,
                                           (WHOLE_PTR) &node::set_node,
                                           arg1, arg2, arg4, NULL, arg1, arg3, arg4, NULL);
                past->add_atom(this_atom);
                }
            }
        }

    new_node->set_node(first_node, n_down);
    if (first_node->get_left())
        first_node->get_left()->set_node(NULL, n_right);

    // update the parent and sibling-order members of new_node's
    //   new children - new node adopts the vector of children
    //   beginning with first_node - cannot use
    //       new_node->connect_node(first_child, n_down)
    //   because ::connect_node() would cause adoption of first
    //   child only
   
    first_node->update_sib_vec(new_node, NULL); // update_sib_vec records its own history

    return new_node;
}

/*****************************************************************************************************

        Navigation functions--special in that they allow moving into a subtree

*****************************************************************************************************/

/*implementation of get_next--special only for going down into the subtree*/
node* bnode::get_next(node_dir& dir)
{
    if ((dir == n_down) && down_ptr)
        return down_ptr;
    else
        return (node::get_next(dir));
}

/*implementation of get_prev--special only for going down into the subtree*/
node* bnode::get_prev(node_dir& dir)
{
    register node *temp,
                       *temp_next;

    if ((dir == n_down) && down_ptr){
        for (temp = down_ptr, temp_next = temp->get_right();
                temp_next; temp_next= temp_next->get_right())
            temp = temp_next;
        return temp;
        }
    else
        return(node::get_prev(dir));
}

/*implementation of get_node--special only for allowing getting the down_ptr*/
node* bnode::get_node(node_dir dir)
{

    if (dir == n_down)
        return(down_ptr);
    else
        return(node::get_node(dir));
}


/*implementation of set_node--special only for allowing setting of the down_ptr*/
/*  NOTE: no history should be added for this function. It is */
/*    most often called by functions that write their own history */
/*    in circumstances where recording of its history would disturb */
/*    the undo and redo functions for the function that called it. */
/*    Where necessary (e.g. bnode::split() and model::make_root()) */
/*    record an atom of history for a call to node::set_node() that */
/*    must be recorded. */
void bnode::set_node(node *to_node, node_dir dir)
{
    if (dir == n_down)
        down_ptr = to_node;
    else
        node::set_node(to_node, dir);
}

/*implementation of get_last--special only for allowing descent into a subtree*/
node* bnode::get_last()
{
    register node *temp,
                  *temp_next;
                             
    if (down_ptr){
        for (temp = down_ptr; temp_next = temp->get_right(); temp = temp_next)
            ;
        temp = temp->get_last();
        }
    else
        temp = this;
    
    return temp;
}

/*****************************************************************************************************

        Copy functions for the subclass bnode

*****************************************************************************************************/

/*copy a node, perhaps to a different model*/
node* bnode::copy(model *to_model)
{
    register node* new_node;

    if (!to_model)
        to_model = get_model();
    new_node = new bnode(to_model);
    if (properties)
        new_node->prop_attach(properties->copy_all(to_model));

    return(new_node);
}

/*copy the tree with root this, perhaps to another model*/
node* bnode::copy_all(model *to_model )
{
    register node *new_node,
                  *first_child,
                  *first_copy,
                  *next_copy;

    new_node = copy(to_model);        //make a new subtree root

    if (down_ptr){                                       //there is something in the subtree
        first_copy = down_ptr->copy_all(to_model);       //recursively copy the tree dominated by the first child
        new_node->connect_node(first_copy, n_down);      //connect the new first child to the new root
        for(first_child = down_ptr->get_right();         //get the next child
                first_child;                             //make sure there is something to copy
                first_child = first_child->get_right()){ //move to the next sibling
            next_copy = first_child->copy_all(to_model); //copy this subtree
            first_copy->connect_node(next_copy,n_right); //attach it to the right of the last subtree
            first_copy = next_copy;                      //make the new subtree the last subtree
            }
        }
    return new_node;
}
                                                                                 
/*duplicate a node--create a new copy in the same model, pointing to the same property vector*/
node* bnode::dup()
{
    register node  *new_node;

    new_node = new bnode(get_model());
    if (properties)
        new_node->prop_attach(properties);

    return new_node;
}

/*duplicate a tree--create a new copy in the same model, pointing to all of the same properties and contents*/
node* bnode::dup_all()
{
    register node  *new_node,
                   *first_child,
                   *first_dup,
                   *next_dup;

    new_node = dup();
    
    if (down_ptr) {
        first_dup = down_ptr->dup_all();
        new_node->connect_node(first_dup, n_down);
        for (first_child = down_ptr->get_right();
                first_child; first_child = first_child->get_right()){
            next_dup = first_child->dup_all();
            first_dup->connect_node(next_dup, n_right);
            first_dup = next_dup;
            }
        }
    return new_node;
}

/*****************************************************************************************************

        Member functions for derived class gnode

*****************************************************************************************************/

/*constructor for objects of type gnode*/
gnode::gnode(model *to_model, long size) : (size ? size : (long) sizeof(gnode), to_model)
{
    register gnode *new_node;

    size = size ? size : sizeof(gnode);

    if (!this){
        if (to_model){
            new_node = (gnode *)(to_model->get_current_oheap()->get_tf_node(size));
            if (new_node){
                if (new_node->up){
                    if (new_node->left)
                        new_node->get_left()->set_right(new_node->up);
                    else
                        to_model->get_current_oheap()->tf_node = new_node->up;
                    new_node->up->set_right(new_node->get_right());
                    if (new_node->right)
                        new_node->right->set_left(new_node->up);
                    }
                else {
                    if (new_node->get_left())
                        new_node->get_left()->set_right(new_node->get_right());
                    else
                        to_model->get_current_oheap()->tf_node = new_node->right;
                    if (new_node->right)
                        new_node->right->set_left(new_node->get_left());
                    }
                }
            else {
                new_node = (gnode *)(to_model->get_current_oheap()->alloc(size));
                if (!new_node){
                    return;
                    }
                }
            this = new_node;
            }
        else {
            new_node = (gnode *) (new char [size]);
            if (!new_node)
                return;
            else
                this = new_node;
            }
        }
    else
        this = this;

    n_type = gobj;
    goo_ptr = 0;
    count = 0;
}

/*overload of = operator, allowing direct assignment of strings to nodes*/
gnode& gnode::operator= (char *s)
{
    history  *past;
    atom     *this_atom;
    atom_arg *arg1, *arg2, *arg3, *arg4, *arg5;
    char_set *cur_char_set;

    if (NULL != get_model()) {
        past = get_model()->get_past();

        // if history is active, record this call

        if (past->has_hist() && past->is_active()) {
            arg1 = new atom_this_arg(past, this);
            arg2 = new atom_val_arg(past, goo_ptr);
            arg3 = new atom_val_arg(past, count);
            }
        cur_char_set = (char_set*) get_model()->inq_state(this, tml_char_set_prop);
        count = cur_char_set->cs_strlen((unsigned char*) s) * cur_char_set->get_char_size();
        goo_ptr = get_model()->get_current_gheap()->char_alloc(s,count);
        if (past->has_hist() && past->is_active()) {
            arg4 = new atom_val_arg(past, goo_ptr);
            arg5 = new atom_val_arg(past, count);
            this_atom = new atom(past, (WHOLE_PTR) &gnode::set_goo_ptr,
                                 (WHOLE_PTR) &gnode::set_goo_ptr,
                                 arg1, arg2, NULL, arg1, arg4, NULL);
            past->add_atom(this_atom);
            this_atom = new atom(past, (WHOLE_PTR) &gnode::set_char_count, (WHOLE_PTR) &gnode::set_char_count,
                                 arg1, arg3, NULL, arg1, arg5, NULL);
            past->add_atom(this_atom);
            }
        }
    else {
        delete [count] goo_ptr;
        cur_char_set = (char_set*) get_model()->inq_state(this, tml_char_set_prop);
        count = cur_char_set->cs_strlen((unsigned char *) s) * cur_char_set->get_char_size();
        goo_ptr = new char [count];
        goo_ptr = tml_os_byte_copy(s,goo_ptr,count);
        }
    return(*this);
}

/*get the null terminated string dominated by the node*/
char *gnode::get_0t_contents()
{
    register char      *buf;
             char_set  *in_char_set = NULL;
    unsigned long       char_size;

    if (get_model())
        in_char_set = (char_set*) get_model()->inq_state(this, tml_char_set_prop);
    if (NULL == in_char_set)
        in_char_set = get_dflt_union_char_set()->get_sets();
    char_size = in_char_set->get_char_size();
    buf = new char[count + ((in_char_set) ? in_char_set->get_char_size() : 1)];
    buf = tml_os_byte_copy(goo_ptr, buf, count);
    in_char_set->get_terminator(((unsigned char *) buf) + count);
    return (buf);
}
                                                  
/*return the last in a list of sibling gnodes*/
node* gnode::get_last()
{
    return this;
}

/* split a leaf node into two nodes */
/* The new node will be the right sibling of this and will dominate the characters */
/*   from new_count on. This will dominate only new_count characters. */
/* If the node being split has properties, create a new parent for it */
/*   and move the properties to that node */
node* gnode::split(long new_count)
{
    register gnode       *new_node;
    register bnode       *new_branch;
    register mark_list   *mark_item;
    register model_range *mark_item_range;
    register mf_list     *form_list;
    register format      *this_format;
    register fv_list     *view_list;
    register view_port   *this_viewport;
    register v_cursor    *this_cursor;
    model_pos             range_start,
                          range_end;

    history  *past;
    atom     *this_atom;
    atom_arg *arg1, *arg2, *arg3;

    // some history recorded by functions called

    if (new_count > count) {
        printf("error in gnode::split--attempt to split beyond length of string\n");
        return 0;
        }

    if (properties) {
        new_branch = new bnode(get_model());
        if (!new_branch) {
            delete new_branch;
            return 0;
            }
        new_branch = (bnode*)(connect_node(new_branch, n_up));
        new_branch->prop_attach(properties);
        prop_detach();
        }
    
    if ((count == new_count) || (new_count < 1))
        return (this);

    new_node = new gnode(get_model());
    if (!new_node) return(new_node);
    
    new_node = (gnode*)(connect_node(new_node, n_right));
    new_node->goo_ptr = goo_ptr + new_count;
    new_node->count = count - new_count;

    if (NULL != get_model()) {
        past = get_model()->get_past();

        // if history is active, record reset of count

        if (past->has_hist() && past->is_active()) {
            arg1 = new atom_this_arg(past, this);
            arg2 = new atom_val_arg(past, count);
            arg3 = new atom_val_arg(past, new_count);
            this_atom = new atom(past, (WHOLE_PTR) &gnode::set_char_count,
                                 (WHOLE_PTR) &gnode::set_char_count,
                                 arg1, arg2, NULL,
                                 arg1, arg3, NULL);
            past->add_atom(this_atom);
            }
        }

    count = new_count;

    /*now, check any of the marks associated with the model and update if necessary*/
    for (mark_item = get_model()->get_marks(); mark_item; mark_item = mark_item->get_next()){
        mark_item_range = mark_item->get_range();
        range_start = mark_item_range->get_start();
        range_end = mark_item_range->get_end();
        if ((this == range_start.in_node) && (range_start.char_index >= count)){
            range_start.in_node = new_node;
            range_start.char_index -= count;
            mark_item_range->set_start(range_start);
            }
        if ((this == range_end.in_node) && (range_end.char_index >= count)){
            range_end.in_node = new_node;
            range_end.char_index -= count;
            mark_item_range->set_end(range_end);
            }
        }

    /*now, update any cursor positions which may have been changed by the split*/
    if (get_model()){
        for (form_list = get_model()->get_formats(); form_list; form_list = form_list->inq_next()){
            this_format = form_list->inq_format();
            if (this_format){
                for(view_list = this_format->get_fv_list(); view_list; view_list = view_list->inq_next()){
                    this_viewport = view_list->inq_vp();
                    if (!this_viewport) continue;
                    this_cursor = this_viewport->get_cursor();
                    if (!this_cursor) continue;
                    range_start = this_cursor->get_model_pos();
                    if ((this == range_start.in_node) && (range_start.char_index >= count)){
                        range_start.in_node = new_node;
                        range_start.char_index -= count;
                        this_cursor->set_model_pos(range_start);
                        }
                    }
                }
            }
        }
    return(new_node);
}


/*copy node 'this' to a different model.
Note: 'copy' does not connect the new node into the target model;
you must call connect_node to do this.
Note: default parameter in_chars, when set, enables copy of goo and string pointers only,
disabling copy of string and goo characters */
node* gnode::copy(model* to_model)
{
    register gnode* new_node;

    if (!to_model) to_model = get_model();
    new_node = new gnode(to_model);
    if (!new_node) return 0;

    new_node->n_type = gobj;
    if (properties)
        new_node->prop_attach(properties->copy_all(to_model));

    new_node->count = count;
    if (to_model)
        new_node->goo_ptr = to_model->get_current_gheap()->char_alloc(goo_ptr,count);
    else {
        new_node->goo_ptr = new char[count];
        tml_os_byte_copy(goo_ptr, new_node->goo_ptr, count);
        }

    return(new_node);
}

/*duplicate a node--create a duplicate of the node in the model, sharing properties but
  duplicating characters*/
node* gnode::dup()
{
    register gnode *return_node;

    return_node = new gnode(get_model());
    if (!return_node) return 0;
    if (properties)
        return_node->prop_attach(properties);

    return_node->count = count;
    return_node->goo_ptr = goo_ptr;

    return(return_node);
}

/*****************************************************************************************************
        
        functions which alter the data dominated by a node--used for input
        Note--all of these assume that the data altered is characters

*****************************************************************************************************/

/*add data to the heap and set the goo pointer*/
long    gnode::attach_data(void *data_start, long data_count)
{
    history  *past;
    atom     *this_atom;
    atom_arg *arg1 = 0, *arg2 = 0, *arg3, *arg4, *arg5;

    if (NULL != get_model()) {
        past = get_model()->get_past();

        // if history is active, record this call

        if (past->has_hist() && past->is_active()) {
            arg1 = new atom_this_arg(past, this);
            arg2 = new atom_val_arg(past, goo_ptr);
            arg3 = new atom_val_arg(past, count);
            }
        goo_ptr = get_model()->get_current_gheap()->char_alloc((char *)(data_start),data_count);
        if (!goo_ptr) {
            delete arg1;
            delete arg2;
            return 0;
            }
        else
            count = data_count;
        if (past->has_hist() && past->is_active()) {
            arg4 = new atom_val_arg(past, goo_ptr);
            arg5 = new atom_val_arg(past, count);
            this_atom = new atom(past, (WHOLE_PTR) &gnode::set_goo_ptr,
                                 (WHOLE_PTR) &gnode::set_goo_ptr,
                                 arg1, arg2, NULL, arg1, arg4, NULL);
            past->add_atom(this_atom);
            this_atom = new atom(past, (WHOLE_PTR) &gnode::set_char_count, (WHOLE_PTR) &gnode::set_char_count,
                                 arg1, arg3, NULL, arg1, arg5, NULL);
            past->add_atom(this_atom);
            }
        }
    else {  /*node is not in a model; delete any string it already dominates*/
        delete [count] goo_ptr;
        goo_ptr = new char [data_count];
        if (!goo_ptr)
            return 0;
        goo_ptr = tml_os_byte_copy((char *)(data_start),goo_ptr,data_count);
        count = data_count;
        }
    return(count);
}


/*insert a character or string--returns the model position of the last character inserted*/
model_pos gnode::insert(long after_index, char *add_data, long add_count, char_set* ins_char_set)
{
    register goo_heap   *in_goo;
    register gnode      *new_node;

    model_pos return_pos;
    node_dir  connect_dir;
    history  *past;
    long      new_count;
    long      partial_count;
    char_set *in_char_set;
    atom     *this_atom;
    atom_arg *arg1, *arg2, *arg3;

    /* get the appropriate character sets */
    in_char_set = (char_set*) get_model()->inq_state(this, tml_char_set_prop);
    if (NULL == ins_char_set)
        ins_char_set = in_char_set;

    /*get the goo heap into which we are inserting*/
    in_goo = get_model()->get_current_gheap();
    connect_dir = n_right;
    new_node = NULL;

    if ((in_char_set == ins_char_set) && (after_index == (count - in_char_set->get_char_size()))) {
        /* insertion at end of goo dominated by node */
        if (((goo_ptr + after_index)) == (in_goo->get_start() + (in_goo->get_byte() - in_char_set->get_char_size()))
          && ((in_goo->get_byte() + add_count) < in_goo->get_size())) {
            /* insertion at end of goo heap; increment count and return */

            in_goo->string_alloc(add_count, add_data);

            if (NULL != get_model()) {
                past = get_model()->get_past();                 //split takes care of history in
                                                                //  other cases
                // if history is active, record this call

                if (past->has_hist() && past->is_active()) {
                    arg1 = new atom_this_arg(past, this);
                    arg2 = new atom_val_arg(past, count);
                    new_count = count + add_count;
                    arg3 = new atom_val_arg(past, new_count);
                    this_atom = new atom(past, (WHOLE_PTR) &gnode::set_char_count, (WHOLE_PTR) &gnode::set_char_count,
                                     arg1, arg2, NULL, arg1, arg3, NULL);
                    past->add_atom(this_atom);
                    }
                }

            count += add_count;
            return_pos.in_node = this;
            return_pos.char_index = count - in_char_set->get_char_size();
            return return_pos;
            }
        }
    else if (after_index < 0)
        /* insertion before first character dominated by node */
        connect_dir = n_left;
    else
        /* insertion between two characters dominated by node */
        (gnode*) split(after_index + 1);
         
    new_node = (gnode*) new gnode(get_model());
    if (ins_char_set != in_char_set)
        connect_prop(make_char_set_prop(ins_char_set, get_model()), new_node);

    // if amount to be added exceeds goo heap capacity, add capacity and recurse
    partial_count = (add_count > MAX_HEAP_INCR) ? MAX_HEAP_INCR : add_count;
    (gnode*) connect_node(new_node, connect_dir);
    new_node->goo_ptr = in_goo->string_alloc(partial_count, add_data);
    new_node->count = partial_count;
    add_count -= partial_count;
    add_data += partial_count;
    // recurse if data remains to be inserted
    if (add_count)
        return_pos = new_node->insert(partial_count, add_data, add_count);
    // else return model_pos of last character inserted
    else {
        return_pos.in_node = new_node;
        return_pos.char_index = ((gnode*) return_pos.in_node)->get_char_count();
        return_pos.char_index -= in_char_set->get_char_size();
        }
    return return_pos;
}

/*delete a character--returns the model position of the character prior to the one deleted*/
model_pos gnode::del_cont(long delete_index, long delete_num)
{
    register mark_list   *mark_item, *hold_item;
    register model_range *mark_item_range;
    register mf_list     *form_list;
    register format      *this_format;
    register fv_list     *view_list;
    register view_port   *this_viewport;
    register v_cursor    *this_cursor;
    long                  updated;

    model_pos return_pos,
              delete_start,
              delete_end,
              prior_pos,
              next_pos,
              next_after,
              range_start,
              range_end,
              cursor_pos;

    history  *past;
    long      hist_active = 0;
    atom     *this_atom;
    atom_arg *arg1, *arg2, *arg3, *arg4, *arg5;

    /* request to delete zero characters */
    /* obtain model_pos suggested by "this" and "delete_index" */
    /* correct and return it */
    if (0 == delete_num) {
        return_pos.in_node = this;
        return_pos.char_index = 0;
        return_pos += delete_index;
        return_pos.normalize();
        return return_pos;
        }

    /* are we deleting something in the middle of a node? if so, split and call again */
    if (delete_index && ((delete_index + delete_num) < count)) {
        return_pos.in_node = split(delete_index);
        if (return_pos.in_node)
            return_pos = return_pos.in_node->del_cont(0, delete_num);
        else
            return_pos = del_cont(delete_index, count - delete_index);
        return return_pos;
        }

    /*make sure that we are in the correct node--if not, get the node and delete there*/
    if (delete_index > (count - 1)) {
        delete_start.in_node = this;
        delete_start.char_index = delete_index;
        delete_start.normalize();
        if (delete_start.in_node != this)
            return (delete_start.in_node->del_cont((delete_index - (count - 1)), delete_num));
    }

    /*if we are deleting more than is in this node, make calls to delete the end of this
        node and the beginning of the next one, if there is one*/
    if (delete_num > (count - delete_index)) {
        delete_start.in_node = this;
        delete_start.char_index = 0;
        delete_start += (count + 1);
        if (delete_start.in_node != this){
            delete_end = delete_start.in_node->del_cont(0, delete_num - (count - delete_index));
            return (del_cont(delete_index, count - delete_index));
            }
        else
            delete_num = count - delete_index;
        }

    /*if we've gotten here, we are really going to do something. First, get the character before
        and after the delete, and what the character after the delete will be*/
    delete_start.in_node = this;
    delete_start.char_index = delete_index;
    prior_pos = delete_start.get_prior();
    delete_end = delete_start + (delete_num - 1);
    next_pos = delete_end.get_next();
    if ((next_pos.in_node == this) && (!delete_index)){
        next_pos.char_index = 0;
        prior_pos.char_index -= delete_num;
        }

    /*update any marked positions if necessary*/
    for (mark_item = get_model()->get_marks(); mark_item; mark_item = hold_item) {
        hold_item = mark_item->get_next();
        mark_item_range = mark_item->get_range();
        range_start = mark_item_range->get_start();
        range_start.normalize();
        range_end = mark_item_range->get_end();
        range_end.normalize();
        updated = 0;
        if (range_start.in_node == this) {
            if ((0 == delete_index) && (delete_index < range_start.char_index)) {
                range_start.char_index -= delete_num;
                if (range_start.char_index < 0)
                    range_start.char_index = 0;
                mark_item_range->set_start(range_start);
                }
            else if ((range_start >= delete_start) && (range_start <= delete_end)) {
                updated++;
                range_start = next_pos.in_node ? next_pos : prior_pos;
                mark_item_range->set_start(range_start);
                }
            }
        if (range_end.in_node == this) {
            if ((0 == delete_index) && (range_end.char_index > delete_index)) {
                range_end.char_index -= delete_num;
                if (range_end.char_index < 0)
                    range_end.char_index = 0;
                mark_item_range->set_end(range_end);
                }
            else if ((range_end >= delete_start) && (range_end <= delete_end)) {
                updated++;
                range_end = prior_pos.in_node ? prior_pos : next_pos;
                mark_item_range->set_end(range_end);
                }
            }
        if ((updated == 2) && (range_start == range_end))
            get_model()->delete_mark(mark_item);
        }

    /*now, update the cursor position if necessary*/
    if (get_model()) {
        for (form_list = get_model()->get_formats(); form_list; form_list = form_list->inq_next()) {
            this_format = form_list->inq_format();
            if (this_format) {
                for(view_list = this_format->get_fv_list(); view_list; view_list = view_list->inq_next()) {
                    this_viewport = view_list->inq_vp();
                    if (!this_viewport) continue;
                    this_cursor = this_viewport->get_cursor();
                    if (!this_cursor) continue;
                    cursor_pos = this_cursor->get_model_pos();
                    if ((this == cursor_pos.in_node) &&
                      (cursor_pos.char_index >= delete_start.char_index) &&
                      (cursor_pos.char_index <= delete_end.char_index)){
                        cursor_pos = next_pos.in_node ? next_pos : prior_pos;
                        this_cursor->set_model_pos(cursor_pos);
                        }
                    }
                }
            }
        }

    if (NULL != get_model()) {
        past = get_model()->get_past();

        // if history is active, record this call

        hist_active = (past->has_hist() && past->is_active());
        }

    /*are we deleting the first delete_num characters of the gnode?*/
    if (!delete_index) {
        /*if we are only deleting part of the goo, update the goo_ptr and the count*/
        if (delete_num < count) {
            if (hist_active) {
                arg1 = new atom_this_arg(past, this);
                arg2 = new atom_val_arg(past, goo_ptr);
                arg3 = new atom_val_arg(past, count);
                }

            goo_ptr += delete_num;
            count -= delete_num;

            if (hist_active) {
                arg4 = new atom_val_arg(past, goo_ptr);
                arg5 = new atom_val_arg(past, count);
                this_atom = new atom(past, (WHOLE_PTR) &gnode::set_goo_ptr, (WHOLE_PTR) &gnode::set_goo_ptr,
                                     arg1, arg2, NULL, arg1, arg4, NULL);
                past->add_atom(this_atom);
                this_atom = new atom(past, (WHOLE_PTR) &gnode::set_char_count, (WHOLE_PTR) &gnode::set_char_count,
                                     arg1, arg3, NULL, arg1, arg5, NULL);
                past->add_atom(this_atom);
                }
            }
        else {
        /*if we are deleting all of the goo, get rid of the node*/
            delete this;            // delete calls detach; history taken care of there
            }                       // delete puts node on free list since this call
        }                           //   occurs only when model member is set

    /*if we are deleting the last of the goo, just update the count and return*/
    else if ((delete_index + delete_num) >= count) {

        if (hist_active) {
            arg1 = new atom_this_arg(past, this);
            arg2 = new atom_val_arg(past, count);
            arg3 = new atom_val_arg(past, delete_index);
            this_atom = new atom(past, (WHOLE_PTR) &gnode::set_char_count, (WHOLE_PTR) &gnode::set_char_count,
                                 arg1, arg2, NULL, arg1, arg3, NULL);
            past->add_atom(this_atom);
            }

        count = delete_index;
        }

    return ((next_pos.in_node) ? next_pos : prior_pos);
        
}

/*overstrike*/
model_pos gnode::overstrike(long begin_index, char *rep_with, long over_num)
{
    register long   i;
    register gnode *new_node;

    model_pos  return_pos, old_pos;
    char      *new_goo_ptr;
    history   *past;
    atom      *this_atom;
    atom_arg  *arg1, *arg2, *arg3;

    /*the first ugly--does overstrike span more than one node? if so, break it up and call again*/
    /*NOTE--this may not be the correct way to do this, as if the next node has different properties,
        the characters used in the overstrike will have those properties*/
    if (over_num > (count - begin_index)){
        i = count - begin_index;

        // recursive call NOT surrounded by history->deactivate/reactivate
        //   sequence as this function not used for either undo and redo 
        //   function pointer in history atom

        return_pos = overstrike(begin_index, rep_with, i);

        rep_with += i;
        over_num -= i;
        old_pos = return_pos;
        return_pos = return_pos.get_next();
        /*if the next formatting node is another gnode, overstrike it; otherwise insert*/
        if (return_pos.in_node) {
            if ( (return_pos.in_node->get_type()) == gobj )

                // recursive call NOT surrounded by history->deactivate/reactivate
                //   sequence as this function not used for either undo and redo 
                //   function pointer in history atom

                return_pos = return_pos.in_node->overstrike(0, rep_with, over_num);
            else
                return_pos = return_pos.in_node->insert(return_pos.char_index, rep_with, over_num);   //insert handles history
            }
        else
            return_pos = old_pos.in_node->insert(old_pos.char_index, rep_with, over_num);
        return return_pos;
        }
                                                           
    /*are we replacing the entire contents of the node?*/
    if ((begin_index == 0) && (over_num == count)){
        new_goo_ptr = get_model()->get_current_gheap()->string_alloc(over_num, rep_with);

        if (NULL != get_model()) {
            past = get_model()->get_past();

            // if history is active, record this call

            if (past->has_hist() && past->is_active()) {
                arg1 = new atom_this_arg(past, this);
                arg2 = new atom_val_arg(past, goo_ptr);
                arg3 = new atom_val_arg(past, new_goo_ptr);
                this_atom = new atom(past, (WHOLE_PTR) &gnode::set_goo_ptr,
                                           (WHOLE_PTR) &gnode::set_goo_ptr,
                                           arg1, arg2, NULL, arg1, arg3, NULL);
                past->add_atom(this_atom);
                // don't bother recording count -- it doesn't change
                }
            }

        goo_ptr = new_goo_ptr;

        return_pos.in_node = this;
        return_pos.char_index = count - 1;
        return return_pos;
        }

    /*are we replacing at the beginning of the node?*/
    if (begin_index == 0){

        return_pos.in_node = split(over_num); // split records its own history
        if (!return_pos.in_node){             //   including change to count
            return_pos.char_index = -1;
            return return_pos;
            }
        new_goo_ptr = get_model()->get_current_gheap()->string_alloc(over_num, rep_with);

        if (NULL != get_model()) {
            past = get_model()->get_past();

            // if history is active, record this call

            if (past->has_hist() && past->is_active()) {
                arg1 = new atom_this_arg(past, this);
                arg2 = new atom_val_arg(past, goo_ptr);
                arg3 = new atom_val_arg(past, new_goo_ptr);
                this_atom = new atom(past, (WHOLE_PTR) &gnode::set_goo_ptr,
                                           (WHOLE_PTR) &gnode::set_goo_ptr,
                                           arg1, arg2, NULL, arg1, arg3, NULL);
                past->add_atom(this_atom);
                }
            }

        goo_ptr = new_goo_ptr;

        return_pos.in_node = this;
        return_pos.char_index = count - 1;
        return return_pos;
        }
    
    /*are we replacing at the end of the node?*/
    if ((begin_index + over_num) == count) {
        new_node = (gnode *)split(begin_index);  // split records its own history
        new_goo_ptr = get_model()->get_current_gheap()->string_alloc(over_num, rep_with);
        if (NULL != get_model()) {
            past = get_model()->get_past();
        
            // if history is active, record this call
        
            if (past->has_hist() && past->is_active()) {
                arg1 = new atom_this_arg(past, new_node);
                arg2 = new atom_val_arg(past, goo_ptr);
                arg3 = new atom_val_arg(past, new_goo_ptr);
                this_atom = new atom(past, (WHOLE_PTR) &gnode::set_goo_ptr,
                                           (WHOLE_PTR) &gnode::set_goo_ptr,
                                           arg1, arg2, NULL, arg1, arg3, NULL);
                past->add_atom(this_atom);
                }
            }
        new_node->goo_ptr = new_goo_ptr;
        return_pos.in_node = new_node;
        return_pos.char_index = over_num - 1;
        return return_pos;
        }

    /*final case--we are replacing in the middle of a node--split it so that we are
        extending at the end of a node, and call this routine again*/
    new_node = (gnode *)split(begin_index + over_num);  // split records its own history

    // recursive call NOT surrounded by history->deactivate/reactivate
    //   sequence as undo and redo are not performed by calls to this function

    return_pos = overstrike(begin_index, rep_with, over_num);
    return return_pos;
}

/*****************************************************************************************************

        Formatting functions

*****************************************************************************************************/

/*return the height of the formatted text dominated by the node*/
long gnode::get_format_hi(format *in_format)
{
    register font_tbl     *use_font;

    if (!in_format || (in_format->get_model() != get_model()))
        return 0;

    use_font = (font_tbl*)in_format->inq_state(this, tml_font_prop);
    if (!use_font)
        use_font = (font_tbl*)(get_model()->get_state()->get_default(tml_font_prop));
    if (use_font)
        return use_font->get_hi();
    else
        return 0;
}

/***********************************************************************************************
    Formatting utility routines
***********************************************************************************************/

/* returns index of last character formatted */
long    format_no_wrap(char *string_in, long string_len, font_tbl *use_font, short& open_space, short& space_used, char_set* in_char_set = 0)
{
    register long       i;
    register long       icspace;
    register char      *use_string;
    register long       fixed_width;
    long                num_chars;
    unsigned long       char_size;
    lrud_dir            start_direc;

    if (NULL == in_char_set)
        in_char_set = get_dflt_union_char_set()->get_sets();
    char_size = in_char_set->get_char_size();

    space_used = 0;
    icspace = use_font->get_icspace();
    use_string = string_in;
    if (!use_font->get_width_table((unsigned long*)(&num_chars)))
        fixed_width = use_font->get_width((unsigned long)(in_char_set->as_ulong((unsigned char*) use_string)));
    else
        fixed_width = 0;
    start_direc = in_char_set->get_dir_to_char((unsigned char *) use_string);
    if (fixed_width) {
        for (i = 0; i < string_len; i += char_size, use_string += char_size) {
            if (in_char_set->is_newline((unsigned char *) use_string)) {
                open_space = 0;
                break;
                }
            if ((in_char_set->is_htab((unsigned char *) use_string))
              || (start_direc != in_char_set->get_dir_to_char((unsigned char *) use_string))) {
                i -= char_size;
                break;
                }
            space_used += (short) (fixed_width + icspace);
            }
        }
    else {
        for (i = 0; i < string_len; i += char_size, use_string += char_size) {
            if (in_char_set->is_newline((unsigned char *)use_string)) {
                open_space = 0;
                break;
                }
            if ((in_char_set->is_htab((unsigned char *) use_string))
              || (start_direc != in_char_set->get_dir_to_char((unsigned char *) use_string))) {
                i -= char_size;
                break;
                }
            space_used += (short) (use_font->get_width(in_char_set->as_ulong((unsigned char*) use_string)) + icspace);
            }
        }
    if (i == string_len) i -= char_size;
    return i;
}

long    tamrof_no_wrap(char *string_in, long string_len, font_tbl *use_font, short& open_space, short& space_used, char_set* in_char_set = 0)
{
    register          long      i;
    register          long      icspace;
    register          long      this_space;
    register          char     *use_string;
    register          long      fixed_width;
                      long      num_chars;
             unsigned long      char_size;
                      lrud_dir  start_direc;

    if (NULL == in_char_set)
        in_char_set = get_dflt_union_char_set()->get_sets();
    char_size = in_char_set->get_char_size();

    space_used = 0;
    icspace = use_font->get_icspace();
    use_string = string_in + string_len - char_size;
    if (!use_font->get_width_table((unsigned long*) &num_chars))
        fixed_width = use_font->get_width(in_char_set->as_ulong((unsigned char*) use_string));
    else
        fixed_width = 0;
    start_direc = in_char_set->get_dir_to_char((unsigned char*) use_string);
    if (fixed_width){
        this_space = fixed_width + icspace;
        for (i = string_len - char_size; i >= 0; i -= char_size, use_string -= char_size) {
            if (in_char_set->is_newline((unsigned char*) use_string)) {
                i += char_size;
                open_space = 0;
                break;
                }
            if ((in_char_set->is_htab((unsigned char*) use_string))
              || (start_direc != in_char_set->get_dir_to_char((unsigned char*) use_string))) {
                i += char_size;
                break;
                }
            space_used += (short) this_space;
            }
        }
    else {
        for (i = string_len - char_size; i >= 0; i -= char_size, use_string -= char_size) {
            if (in_char_set->is_newline((unsigned char *) use_string)) {
                i += char_size;
                open_space = 0;
                break;
                }
            if ((in_char_set->is_htab((unsigned char *) use_string))
              || (start_direc != in_char_set->get_dir_to_char((unsigned char*) use_string))) {
                i += char_size;
                break;
                }
            space_used += (short) (use_font->get_width(in_char_set->as_ulong((unsigned char*) use_string)) + icspace);
            }
        }
    return (i < 0) ? 0 : i;
}

/* returns index of last character formatted, or -1 for error */
long    format_char_wrap(char *string_in, long string_len, font_tbl *use_font, short& open_space, short& space_used, char_set* in_char_set = 0)
{
    register long   i,
                    fixed_width,
                    this_space;
             long   icspace;
    register char  *use_string;
    long            num_chars;
    unsigned long   char_size;
    lrud_dir        start_direc;

    if (NULL == in_char_set)
        in_char_set = get_dflt_union_char_set()->get_sets();
    char_size = in_char_set->get_char_size();

    use_string = string_in;
    icspace = use_font->get_icspace();
    space_used = 0;
    if (!use_font->get_width_table((unsigned long*) &num_chars))
        fixed_width = use_font->get_width(in_char_set->as_ulong((unsigned char*) use_string));
    else
        fixed_width = 0;
                            
    start_direc = in_char_set->get_dir_to_char((unsigned char*) use_string);

    /* add character widths until open space is exceeded */
    /* end line on newline or before horizontal tab */
    /*   or if direction changes */
    /* assumes caller skipped initial horizontal tabs */
    if (fixed_width){
        this_space = fixed_width + icspace;
        for (i = 0;
          (i < string_len) && (open_space > (space_used + this_space));
          i += char_size, use_string += char_size) {
            if (in_char_set->is_newline((unsigned char*) use_string)) {
                open_space = 0;
                return i;
                }
            if ((in_char_set->is_htab((unsigned char*) use_string))
              || (start_direc != in_char_set->get_dir_to_char((unsigned char*) use_string))) {
                open_space -= space_used;
                return i - char_size;
                }
            space_used += (short) this_space;
            }
        }
    else {
        for (i = 0;
          (i < string_len) &&
          (open_space > (space_used + (this_space = use_font->get_width(in_char_set->as_ulong((unsigned char *) use_string)) + icspace)));
          i += char_size, use_string += char_size) {
            if (in_char_set->is_newline((unsigned char*) use_string)) {
                open_space = 0;
                return i;
                }
            if ((in_char_set->is_htab((unsigned char*) use_string))
              || (start_direc != in_char_set->get_dir_to_char((unsigned char*) use_string))) {
                open_space -= space_used;
                return i - char_size;
                }
            space_used += (short) this_space;
            }
        }

    if (i == 0) {
        /* if i is zero, space ran out before first character  */
        /*   could be accommodated, or string had zero length. */
        /*   Do not consider failure to format a zero-length   */
        /*   string as an error.                               */
        if (string_len) {
            open_space = 0;
            return -1;
            }
        else
            return 0;
        }

    if (i == string_len) {
        /* all characters were formatted */
        i -= char_size;
        open_space -= space_used;
        return i;
        }

    /* at least one character was formatted, but not all */
    /* skip spaces or spaces-and-newline following formatted text */

    i -= char_size;
    use_string -= char_size;

    while (i < string_len) {
        if (in_char_set->is_newline(((unsigned char *) use_string) + char_size)) {
            open_space = 0;
            return i + char_size;
            }
        else if (in_char_set->cs_isspace(((unsigned char *) use_string) + char_size) 
          && ( ! 
            (in_char_set->is_htab(((unsigned char *)use_string) + char_size)
          || in_char_set->is_cr(((unsigned char *)use_string) + char_size)
          || in_char_set->is_vtab(((unsigned char *)use_string) + char_size)
          || in_char_set->is_formfeed(((unsigned char *)use_string) + char_size)))) {
            /* space character found */
            /* assumes: */
            /*  (a) no-break-space character if any is not */
            /*      included in space ranges of char_set */
            /*  (b) caller checks for and handles horizontal tab, */
            /*      carriage return, vertical tab, and formfeed */
            /*      if they mean something in caller's context */
            i += char_size;
            use_string += char_size;
            }
        else
            break;
        }

    if (i == string_len)
        /* all characters formatted */
        open_space -= space_used;
    else
        /* no more room or EOL */
        open_space = 0;

    return i;
}

long    tamrof_char_wrap(char *string_in, long string_len, font_tbl *use_font, short& open_space, short& space_used, char_set* in_char_set = 0)
{
    register long   i,
                    fixed_width,
                    this_space,
                    icspace;
    register char  *use_string;
    long            num_chars;
    unsigned long   char_size;
    lrud_dir        start_direc;
                       
    if (NULL == in_char_set)
        in_char_set = get_dflt_union_char_set()->get_sets();
    char_size = in_char_set->get_char_size();

    use_string = string_in + string_len - char_size;
    icspace = use_font->get_icspace();
    space_used = 0;

    if (!use_font->get_width_table((unsigned long*) &num_chars))
        fixed_width = use_font->get_width(in_char_set->as_ulong((unsigned char*)(use_string - char_size)));
    else
        fixed_width = 0;
                            
    /* add character widths until open space is exceeded */
    /* end line on newline or after horizontal tab */
    /* assumes caller backed up beyond final horizontal tabs */
    start_direc = in_char_set->get_dir_to_char((unsigned char*) use_string);
    if (fixed_width) {
        this_space = fixed_width + icspace;
        for (i = string_len - char_size;
          (i >= 0) && (open_space > (space_used + this_space));
          i -= char_size, use_string -= char_size) {
            if (in_char_set->is_newline((unsigned char*) use_string)) {
                open_space = 0;
                return i + char_size;
                }
            if ((in_char_set->is_htab((unsigned char*) use_string))
              || (start_direc != in_char_set->get_dir_to_char((unsigned char*) use_string))) {
                open_space -= space_used;
                return i + char_size;
                }
            space_used += (short) this_space;
            }
        }
    else {
        for (i = string_len - char_size;
          (i >= 0) 
          && (open_space > (space_used + (this_space = use_font->get_width(in_char_set->as_ulong((unsigned char*) use_string))) + icspace));
          i -= char_size, use_string -= char_size) {
            if (in_char_set->is_newline((unsigned char*) use_string)) {
                open_space = 0;
                return i + char_size;
                }
            if ((in_char_set->is_htab((unsigned char*) use_string))
              || (start_direc != in_char_set->get_dir_to_char((unsigned char*) use_string))) {
                open_space -= space_used;
                return i + char_size;
                }
            space_used += (short) this_space;
            }
        }

    if (i == (string_len - char_size)) {
        /* if i remained at its initial value, space ran out  */
        /*   before first character could be accommodated, or */
        /*   string had zero length. Do not consider failure  */
        /*   to format a zero-length string as an error.      */
        if (string_len) {
            open_space = 0;
            return -1;
            }
        return 0;
        }

    if (i < 0) {
        /* all characters were formatted */
        open_space -= space_used;
        return 0;
        }

    /* at least one character was formatted, but not all */
    /* skip spaces preceding formatted text */

    while (i >= 0) {
        if (in_char_set->is_newline((unsigned char*) use_string)) {
            open_space = 0;
            return i + char_size;
            }
        else if (in_char_set->cs_isspace((unsigned char*) use_string)
          && ( ! 
            (in_char_set->is_htab((unsigned char*) use_string)
          || in_char_set->is_cr((unsigned char*) use_string)
          || in_char_set->is_vtab((unsigned char*) use_string)
          || in_char_set->is_formfeed((unsigned char*) use_string)))) {
            /* space character found */
            /* assumes: */
            /*  (a) no-break-space character if any is not */
            /*      included in space ranges of char_set */
            /*  (b) caller checks for and handles horizontal tab, */
            /*      carriage return, vertical tab, and formfeed */
            /*      if they mean something in caller's context */
            i -= char_size;
            use_string -= char_size;
            }
        else
            break;
        }

    if (i < 0)
        /* all characters formatted */
        open_space -= space_used;
    else
        /* no more room or BOL */
        open_space = 0;

    return i + char_size;
}

// Per-process global variable indicating whether line break found.
// Global used, rather than reference parameter in recursive function
//   gnode::format_line, to avoid interface change.
// Necessary for incremental reformat under word_wrap of line with 
//   initial word longer than conceptual page width when initial
//   word spans more than one tline. Set to nonzero by format_word_wrap 
//   when break found in a line. Reset to zero by gnode::format_line 
//   when, under word_wrap, it starts a line at conceptual x of zero.
//   When clear this variable signals to gnode::format_line() that
//   no break exists, allowing return of an effectively character-
//   wrapped line. When set, signals that break exists, allowing 
//   rejection of unbroken line segments at EOL otherwise.
// See further notes in gnode::format_line.
// To make this global per-process, must be declared with attribute specifier
//   for Apollo builds to allow use of global and inlib'd libraries
#ifdef apollo
long tml_word_wrap_break_found __attribute((section(tml_word_wrap_break_found))) = 0;
#else
long tml_word_wrap_break_found = 0;
#endif

/* returns index of last character formatted, or -1 for error */
long    format_word_wrap(char *string_in, long string_len, font_tbl *use_font, short& open_space, short& space_used, char_set* in_char_set = 0)
{
    register long  i,
                   this_space,
                   icspace,
                   fixed_width,
                   last_space_width,
                   last_space;
    register char *use_string;
    long           num_chars;
    unsigned long  char_size;
    lrud_dir       start_direc;

    if (NULL == in_char_set)
        in_char_set = get_dflt_union_char_set()->get_sets();
    char_size = in_char_set->get_char_size();

    use_string = string_in;
    icspace = use_font->get_icspace();
    last_space = last_space_width = space_used = 0;
    if (!use_font->get_width_table((unsigned long*) &num_chars))
        fixed_width = use_font->get_width(in_char_set->as_ulong((unsigned char*) use_string));
    else
        fixed_width = 0;

    /* add character widths until open space is exceeded */
    /* end line on newline or or before horizontal tab */
    /*   or when direction changes */
    /* assumes caller skipped initial horizontal tabs */
    start_direc = in_char_set->get_dir_to_char((unsigned char*) use_string);
    if (fixed_width){
        this_space = fixed_width + icspace;
        for (i = 0;
          (i < string_len) && (open_space > (space_used + this_space));
          i += char_size, use_string += char_size) {
            if (in_char_set->is_newline((unsigned char*) use_string)){
                tml_word_wrap_break_found = -1;
                open_space = 0;
                return i;
                }
            if (in_char_set->is_htab((unsigned char*) use_string)){
                tml_word_wrap_break_found = -1;
                open_space -= space_used;
                i -= char_size;
                return i;
                }
            if (start_direc != in_char_set->get_dir_to_char((unsigned char*) use_string)) {
                open_space -= space_used;
                i -= char_size;
                return i;
                }
            space_used += (short) this_space;
            if (in_char_set->cs_isspace((unsigned char*) use_string)) {
                /* any kind of space will do as a break */
                tml_word_wrap_break_found = -1;
                last_space_width = space_used;
                last_space = i;
                }
            }
        }
    else {
        for (i = 0;
          (i < string_len) &&
          (open_space > (space_used + (this_space = use_font->get_width(in_char_set->as_ulong((unsigned char*) use_string)) + icspace)));
          i += char_size, use_string += char_size) {
            if (in_char_set->is_newline((unsigned char*) use_string)) {
                tml_word_wrap_break_found = -1;
                open_space = 0;
                return i;
                }
            if (in_char_set->is_htab((unsigned char*) use_string)){
                tml_word_wrap_break_found = -1;
                open_space -= space_used;
                i -= char_size;
                return i;
                }
            if (start_direc != in_char_set->get_dir_to_char((unsigned char*) use_string)) {
                open_space -= space_used;
                i -= char_size;
                return i;
                }
            space_used += (short) this_space;
            if (in_char_set->cs_isspace((unsigned char*) use_string)) {
                /* any kind of space will do as a break */
                tml_word_wrap_break_found = -1;
                last_space_width = space_used;
                last_space = i;
                }
            }
        }

    if (i == 0) {
        /* if i is zero, space ran out before first character  */
        /*   could be accommodated, or string had zero length. */
        /*   Do not consider failure to format a zero-length   */
        /*   string as an error.                               */
        if (string_len) {
            open_space = 0;
            return -1;
            }
        return 0;
        }

    /* at least one character formatted                       */
    /* count as formatted any spaces, newline, or             */
    /*   spaces-and-newline following:                        */
    /*      last break (if found)                             */
    /*      formatted material (otherwise)                    */
    /* when formatted material ends in nonbreak character and */
    /*   unformatted remainder begins with space, the latter  */
    /*   condition allows identification of a possible EOL    */
    /*   break even when break character itself does not fit. */
    if (last_space)
        /* break found */
        i = last_space;
    else
        i -= char_size;
    use_string = string_in + i;
    while (i < string_len) {
        if (in_char_set->is_newline((unsigned char*) use_string)) {
            tml_word_wrap_break_found = -1;
            open_space = 0;
            return i + char_size;
            }
        else if (in_char_set->cs_isspace(((unsigned char*) use_string) + char_size) 
          && ( ! 
            (in_char_set->is_htab((unsigned char*) use_string)
          || in_char_set->is_cr((unsigned char*) use_string)
          || in_char_set->is_vtab((unsigned char*) use_string + char_size)
          || in_char_set->is_formfeed(((unsigned char *) use_string) + char_size)))) {
            /* space character found */
            /* assumes: */
            /*  (a) no-break-space character if any is not */
            /*      included in space ranges of char_set */
            /*  (b) caller checks for and handles horizontal tab, */
            /*      carriage return, vertical tab, and formfeed */
            /*      if they mean something in caller's context */
            tml_word_wrap_break_found = -1;
            i += char_size;
            use_string += char_size;
            }
        else
            break;
        }

    if (i == string_len)
        /* correct loop index */
        i -= char_size;

    /* count as used only space through last break if there is one */
    /*   or through end of formatted material if not */
    open_space -= (last_space) ? (space_used = (short) last_space_width) : space_used;

    /* return offset to last character formatted */
    /*   up to and including last break character */
    return i;
}

long    tamrof_word_wrap(char *string_in, long string_len, font_tbl *use_font, short& open_space, short& space_used, char_set* in_char_set = 0)
{
    register long  i,
                   this_space,
                   icspace,
                   fixed_width,
                   last_space_width,
                   last_space;
    register char *use_string;
    long           num_chars;
    unsigned long  char_size;
    lrud_dir       start_direc;

    if (NULL == in_char_set)
        in_char_set = get_dflt_union_char_set()->get_sets();
    char_size = in_char_set->get_char_size();

    use_string = string_in + string_len - char_size;
    icspace = use_font->get_icspace();
    space_used = 0;
    last_space = string_len;
    last_space_width = open_space;
    
    if (!use_font->get_width_table((unsigned long*) &num_chars))
        fixed_width = use_font->get_width(in_char_set->as_ulong((unsigned char*) (use_string - char_size)));
    else
        fixed_width = 0;

    /* add character widths until open space is exceeded */
    /* end line on newline or after horizontal tab */
    /*   or when direction changes */
    /* assumes caller skipped initial horizontal tabs */
    start_direc = in_char_set->get_dir_to_char((unsigned char*) (use_string - char_size));
    if (fixed_width){
        this_space = fixed_width + icspace;
        for (i = string_len - char_size;
          (i >= 0) && (open_space > (space_used + this_space));
          i -= char_size, use_string -= char_size) {
            if (in_char_set->is_newline((unsigned char*) use_string)) {
                tml_word_wrap_break_found = -1;
                open_space = 0;
                return i + char_size;
                }
            if (in_char_set->is_htab((unsigned char*) use_string)) {
                tml_word_wrap_break_found = -1;
                open_space -= space_used;
                return i + char_size;
                }
            if (start_direc != in_char_set->get_dir_to_char((unsigned char*) use_string)) {
                open_space -= space_used;
                return i + char_size;
                }
            space_used += (short) this_space;
            if (in_char_set->cs_isspace((unsigned char*) use_string)) {
                /* any kind of space will do as a break */
                tml_word_wrap_break_found = -1;
                last_space_width = space_used;
                last_space = i;
                }
            }
        }
    else {
        for (i = string_len - char_size;
          (i >= 0)
          && (open_space > (space_used + (this_space = use_font->get_width(in_char_set->as_ulong((unsigned char*) use_string)) + icspace)));
          i -= char_size, use_string -= char_size) {
            if (in_char_set->is_newline((unsigned char*) use_string)) {
                tml_word_wrap_break_found = -1;
                open_space = 0;
                return i + char_size;
                }
            if (in_char_set->is_htab((unsigned char*) use_string)) {
                tml_word_wrap_break_found = -1;
                open_space -= space_used;
                return i + char_size;
                }
            if (start_direc != in_char_set->get_dir_to_char((unsigned char*) use_string)) {
                open_space -= space_used;
                return i + char_size;
                }
            space_used += (short) this_space;
            if (in_char_set->cs_isspace((unsigned char*) use_string)) {
                /* any kind of space will do as a break */
                tml_word_wrap_break_found = -1;
                last_space_width = space_used;
                last_space = i;
                }
            }
        }


    if (i == (string_len - char_size)) {
        /* if i equals string_len, space ran out before first  */
        /*   character could be accommodated, or string had    */
        /*   zero length. Do not consider failure to format    */
        /*   a zero-length string as an error.                 */
        if (string_len) {
            open_space = 0;
            return -1;
            }
        return 0;
        }

    /* at least one character formatted                        */
    /* count as formatted any spaces, newline, or              */
    /*   spaces-and-newline preceding:                         */
    /*      last break (if found)                              */
    /*      formatted material (otherwise)                     */
    /* when formatted material begins with nonbreak character  */
    /*   and unformatted remainder ends with space, the latter */
    /*   condition allows identification of a possible BOL     */
    /*   break even when break character itself does not fit.  */

    if (last_space != string_len)
        /* break found */
        i = last_space - char_size;
    use_string = string_in + i;
    while (i >= 0) {
        if (in_char_set->is_newline((unsigned char *) use_string)) {
            tml_word_wrap_break_found = -1;
            open_space = 0;
            return i + char_size;
            }
        else if (in_char_set->cs_isspace((unsigned char *) use_string)
          && ( ! 
            (in_char_set->is_htab((unsigned char *) use_string)
          || in_char_set->is_cr((unsigned char *) use_string)
          || in_char_set->is_vtab((unsigned char *) use_string)
          || in_char_set->is_formfeed((unsigned char *) use_string)))) {
            /* space character found */
            /* assumes: */
            /*  (a) no-break-space character if any is not */
            /*      included in space ranges of char_set */
            /*  (b) caller checks for and handles horizontal tab, */
            /*      carriage return, vertical tab, and formfeed */
            /*      if they mean something in caller's context */
            tml_word_wrap_break_found = -1;
            i -= char_size;
            use_string -= char_size;
            }
        else
            break;
        }

    /* count as used only space through first break if there is one */
    /*   or through beginning of formatted material if not */
    open_space -= (last_space != string_len) ? (space_used = (short) last_space_width) : space_used;

    /* return offset to last character formatted */
    /*   up to and including first break character */
    return i + char_size;
}
                  
long calc_tab_pixel_off (node *at_node, format *in_format, long tab_cnt, long current_pos)
{
    register long     def_tab_val,
                      tab_tbl_size;
    argument         *tab_vals;
             long     def_tab_tbl[def_tab_tbl_size];
             long    *tab_tbl;
             long     return_offset,
                      i;

    /* get value of tab property */
    /* initialize tab offset table */
    tab_vals = (argument*) in_format->inq_state(at_node, tml_tab_prop);
    if (!tab_vals){
        def_tab_val = def_tab_pixels;
        tab_tbl_size = 0;
        }
    else {
        def_tab_val = *((long *) tab_vals->get_val());
        tab_vals = tab_vals->get_next();
        tab_tbl_size = *((long *) tab_vals->get_val());
        if (tab_tbl_size <= def_tab_tbl_size)
            tab_tbl = def_tab_tbl;
        else
            tab_tbl = (long *) tml_os_alloc(tab_tbl_size, sizeof(long));
        for (i = 0; i < tab_tbl_size; i++){
            tab_vals = tab_vals->get_next();
            tab_tbl[i] = *((long *) tab_vals->get_val());
            }
        }

    if ((0 == tab_tbl_size) || (tab_tbl[tab_tbl_size - 1] < current_pos)) {
        if (def_tab_val) {
            tab_cnt += (current_pos / def_tab_val);
            if (0 == tab_tbl_size) {
                return_offset = tab_cnt * def_tab_val;
                return_offset -= current_pos;
                }
            else {
                tab_cnt -= tab_tbl_size;
                return_offset = current_pos + (def_tab_val - (current_pos % def_tab_val));
                if (current_pos % def_tab_val)
                    --tab_cnt;
                return_offset += (tab_cnt * def_tab_val);
                }
            }
        else
            /* tab encountered after last explicit tab stop */
            /*   and value of implicit tab offset is 0 */
            /* indicate clear to EOL */
            return_offset = -1;
        }
    else {
        for (i = 0; (i < tab_tbl_size) && (tab_tbl[i] < current_pos); i++)
            ;
        tab_cnt += i;
        if (tab_cnt <= tab_tbl_size)
            return_offset = tab_tbl[tab_cnt-1] - current_pos;
        else {
            i = tab_tbl_size - 1;
            return_offset = tab_tbl[i];
            return_offset += (def_tab_val * i);
            return_offset -= current_pos;
            }
        }

    if (tab_tbl_size > def_tab_tbl_size)
        tml_os_free((char *) tab_tbl);

    return return_offset;
}

long calc_tab_pixel_off_backward (node *at_node, format *in_format, long tab_cnt, long current_pos)
{
    register long     def_tab_val,
                      tab_tbl_size;
    argument         *tab_vals;
             long     def_tab_tbl[def_tab_tbl_size];
             long    *tab_tbl;
             long     return_offset,
                      i;

    /* get value of tab property */
    /* initialize tab offset table */
    tab_vals = (argument*) in_format->inq_state(at_node, tml_tab_prop);
    if (!tab_vals){
        def_tab_val = def_tab_pixels;
        tab_tbl_size = 0;
        }
    else {
        def_tab_val = *((long *) tab_vals->get_val());
        tab_vals = tab_vals->get_next();
        tab_tbl_size = *((long *) tab_vals->get_val());
        if (tab_tbl_size <= def_tab_tbl_size)
            tab_tbl = def_tab_tbl;
        else
            tab_tbl = (long *) tml_os_alloc(tab_tbl_size, sizeof(long));
        for (i = 0; i < tab_tbl_size; i++){
            tab_vals = tab_vals->get_next();
            tab_tbl[i] = *((long *) tab_vals->get_val());
            }
        }

    if (0 == tab_tbl_size) {
        /* no tab property or tab table */
        return_offset = 0;
        if (current_pos % def_tab_val) {
            return_offset += (current_pos % def_tab_val);
            --tab_cnt;
            }
        return_offset += (def_tab_val * tab_cnt);
        }
    else if (current_pos < tab_tbl[tab_tbl_size - 1]) {
        /* current position is before maximum offset specified in tab table */
        return_offset = current_pos;
        for (i = tab_tbl_size; i && (current_pos < tab_tbl[i-1]); i--)
            ;
        while (i && tab_cnt) {
            --i;
            --tab_cnt;
            }
        if (i)
            return_offset -= tab_tbl[i-1];
        }
    else {
        /* current position is beyond maximum offset specified in tab table */
        return_offset = current_pos;
        if (def_tab_val) {
            tab_cnt += ((current_pos - tab_tbl[tab_tbl_size - 1]) / def_tab_val);
            for (i = (tab_cnt <= tab_tbl_size) ? tab_cnt : tab_tbl_size;
              i && tab_tbl[i-1];
              i--)
                ;
            if (i)
                return_offset -= tab_tbl[i-1];
            }
        else
            /* tab encountered before first explicit tab stop */
            /*   and value of implicit tab offset is 0 */
            /* indicate clear to BOL */
            return_offset = -1;
        }

    if (tab_tbl_size > def_tab_tbl_size)
        tml_os_free((char *) tab_tbl);

    if (return_offset > current_pos)
        return current_pos;

    return return_offset;
}

/* format a line of text until there is no more space */
/* the type of formatting performed is determined by the value of the wrap property*/
cline* gnode::format_line(format* in_format, long start_index, long space_hi, long space_wide, long first_concep_pixel)
{
    register tline     *return_line,
                       *attach_line;
    register font_tbl  *use_font;
    register long       old_up,
                        old_down,
                        new_up,
                        new_down;
             long      *format_value,
                        format_length,
                       *ff_state,
                        end_index,
                        end_or_tab;
             short      space_left,
                        space_used;
             model_pos  hold_pos;
             char      *ff_found = NULL, *ff_start;
             char_set  *in_char_set = NULL;
             char_set  *hold_pos_char_set;
             char_set  *attach_char_set;
             char_set  *attach_end_next_char_set;
             node      *attach_node;
             model_pos  attach_end_next;
             char      *attach_end_char;
             char      *attach_end_next_char;
             long       char_size,
                        i,
                        tab_pixel_off;
             char      *format_chars;
    unsigned long       char_off;
             lrud_dir   attach_dir;

    if (start_index >= count)
        /* request to format data beyond end of node */
        return 0;

    /* get appropriate character set and bytes per character */
    if (get_model())
        in_char_set = (char_set *) get_model()->inq_state(this, tml_char_set_prop);
    if (NULL == in_char_set)
        in_char_set = get_dflt_union_char_set()->get_sets();
    char_size = in_char_set->get_char_size();

    /* get characters to format and how many */
    format_chars = ((char*) goo_ptr) + start_index;                
    format_length = count - start_index;

    /* get value of formfeed property */
    ff_state = (long*) in_format->inq_state(this, tml_form_feed_prop);
            
    if (ff_state && *ff_state) {
        /* formfeed property enabled */
        /* seek formfeed in string to format */
        for (ff_start = ff_found = format_chars;
          ff_found < (ff_start + format_length);
          ff_found += char_size) {
            if (in_char_set->is_formfeed((unsigned char *) ff_found)) {
                /* formfeed found */
                /* truncate string at formfeed */
                format_length = ff_found - ff_start;
                break;
                }
            }
        if (ff_found == (ff_start + format_length))
            ff_found = NULL;
        }

    /*get value of formatting property*/
    format_value = (long*) in_format->inq_state(this, tml_wrap_prop);

    /* get appropriate font */
    if (NULL == (use_font = (font_tbl*) in_format->inq_state(this, tml_font_prop))) {
        if (NULL == (use_font = (font_tbl*) get_model()->get_state()->get_default(tml_font_prop)))
            /* no font: error */
            return 0;
        }

    /* set tab offset value for leading horizontal tab characters */
    /* obtain goo offset to skip leading tab characters */
    for (i = 0, char_off = 0; 
      (i < format_length) && in_char_set->is_htab(((unsigned char *) format_chars) + char_off); 
      i++, char_off += char_size)
        ;
    format_length -= char_off;
    if (i) {
        if (-1 == (tab_pixel_off = calc_tab_pixel_off(this, in_format, i, first_concep_pixel)))
            /* tab beyond last special tab when default tab value is zero */
            /*   clears line to EOL */
            tab_pixel_off = space_wide;
        }
    else
        tab_pixel_off = 0;

    /* set space offset value for leading space characters when */
    /*   at beginning of untabbed line with word wrap enabled */
    /* prevents wrap of space characters to beginning of line */
    if ((0 == first_concep_pixel) 
      && (0 == char_off)
      && format_value 
      && (word_wrap == *format_value)) {
        for (i = 0;
          (i < format_length) && in_char_set->cs_isspace(((unsigned char *) format_chars) + char_off);
          i += char_size, char_off += char_size)
            if (in_char_set->is_newline(((unsigned char *) format_chars) + char_off))
                break;
            ;
        format_length -= char_off;
        }

    /* get space limits for formatting */
    space_left = (short) (space_wide - tab_pixel_off);
    space_used = 0;

    if (0 == format_length)
        /* string empty or composed entirely of formfeed and/or tab character(s): */
        /*   nothing to format */
        end_index = 0;
    else {
        if (NULL == format_value)
            end_index = format_no_wrap(format_chars + char_off, format_length, use_font, space_left, space_used, in_char_set);
        else {
            switch (*format_value) {
                case no_wrap:
                    end_index = format_no_wrap(format_chars + char_off, format_length, use_font, space_left, space_used, in_char_set);
                    break;
                case char_wrap:
                    end_index = format_char_wrap(format_chars + char_off, format_length, use_font, space_left, space_used, in_char_set);
                    break;
                case word_wrap:
                    if (0 == first_concep_pixel)
                        /* indicate no break found when new line begun */
                        /*   at conceptual page x of zero */
                        tml_word_wrap_break_found = 0;
                    end_index = format_word_wrap(format_chars + char_off, format_length, use_font, space_left, space_used, in_char_set);
                    if ((0 == first_concep_pixel) && (0 > end_index)) {
                        /* format_length characters are more than enough to fill line */
                        /*   and first word on line is too long for line width: */
                        /*   use character wrap */
                        space_left = (short) (space_wide - tab_pixel_off);
                        space_used = 0;
                        end_index = format_char_wrap(format_chars + char_off, format_length, use_font, space_left, space_used, in_char_set);
                        /* indicate best fit achieved */
                        space_left = 0;
                        }
                    break;
                default:
                    end_index = format_no_wrap(format_chars + char_off, format_length, use_font, space_left, space_used, in_char_set);
                    break;
                }
            }
        }
                     
    /* if end_index < 0, error */
    if (end_index < 0)
        return 0;

    return_line = new tline(in_format);
    hold_pos.in_node = this;
    hold_pos.char_index = start_index;
    return_line->set_start(hold_pos);
    return_line->set_font(use_font);
    return_line->set_baseline(use_font->get_baseline());
    return_line->set_start_x_off(tab_pixel_off);

    if (ff_found && (0 == format_length)) {
        /* initial formfeed: make rest of page blank */
        return_line->set_type(cl_gutter);
        return_line->set_wide((short) space_wide);
        return_line->set_text_width(space_wide);
        return_line->set_hi((short) space_hi);
        return_line->set_count(char_size);
        space_left = 0;
        }
    else {
        /* normal formatting */
        return_line->set_type(cl_text);
        if (end_index || format_length)
            /* segment containing more than one nontab character */
            /* OR segment consisting of tabs and one character */
            return_line->set_count(char_off + end_index + char_size);
        else
            /* segment consisting of tabs only */
            return_line->set_count(char_off);
        return_line->set_hi((short) use_font->get_hi());
        return_line->set_wide((short) space_wide - space_left);
        return_line->set_text_width(space_used);
        }
                           
    /* no space left */
    if (0 == space_left)
        return return_line;

    /* NOTE: when                                             */
    /*   (1) no recursion has taken place because             */
    /*   (2) format_length contains more than enough material */
    /*       to fill line,                                    */
    /*   (3) first word is too long for line, and             */
    /*   (4) word_wrap is enabled,                            */
    /*  the character-wrapped initial portion of the first    */
    /*  word is returned here, because space_left is set to   */
    /*  zero in switch statement above. See below after       */
    /*  recursive call for code that handles case in which    */
    /*  (3) and (4), but not (1) and (2), obtain.             */

    /* get next model position after formatted material */
    hold_pos += char_off + end_index;
    if (end_index || 
      (0 == char_off) || 
      ((end_index + char_off) < return_line->get_count())) {
        /* segment contains at least one character other than tab and newline or*/
        /*   segment is empty or consists of newline alone or */
        /*   segment consists solely of tabs and newline. */
        /* do not advance hold_pos for segment consisting of tabs alone: */
        /*   char_off is a (one-based) count of tabs; adding char_off */
        /*   when segment consists of tabs alone advances model position */
        /*   appropriately without need for call to model_pos::get_next() */
        /* note that end_index is the (zero-based) index */
        /*   of the character on which formatting ended */
        /* it will be zero for segment consisting of newline alone */
        /*   and zero when segment consists of tabs followed by one */
        /*   character other than tab or newline */
        hold_pos = hold_pos.get_next();
        }

    /* no more goo */
    if (!hold_pos.in_node || !hold_pos.in_node->can_format() || ((hold_pos.in_node->get_type()) != gobj))
        return return_line;

    /* more goo but font too high for line */
    if (hold_pos.in_node->get_format_hi(in_format) > space_hi){
        delete return_line;
        return 0;
        }

    /* !!! RECURSIVE CALL !!! RECURSIVE CALL !!! RECURSIVE CALL !!! */
    /* space remains, goo is available, font fits in vertical space: try to format more             */
    /* the last argument gives the conceptual line pixel at which additional formatting will start: */
    /*     conceptual line pixel at which this formatting attempt began       (first_concep_pixel)  */
    /*   + number of conceptual line pixels used through last formatting call (space_used)          */
    /*   + tab offset in conceptual pixels                                    (tab_pixel off)       */
    attach_char_set = (char_set*) hold_pos.in_node->get_model()->inq_state(hold_pos.in_node, tml_char_set_prop);
    if (::right == (attach_dir = attach_char_set->get_dir_to_char(((unsigned char*) hold_pos.in_node->get_contents()) + hold_pos.char_index)))
        attach_line = (tline*) hold_pos.in_node->format_line(in_format, hold_pos.char_index, space_hi, space_left, first_concep_pixel + space_used + tab_pixel_off);
    else if (::left == attach_dir)
        attach_line = (tline*) hold_pos.in_node->tamrof_line(in_format, hold_pos.char_index, space_hi, space_left, first_concep_pixel + space_used + tab_pixel_off);

    if (format_value && (word_wrap == *format_value)) {
        /* word wrap is enabled ... */

        if ( ! attach_line) {
            /* no more space: no additional material formatted */

            if (this == hold_pos.in_node)
                hold_pos_char_set = in_char_set;
            else
                hold_pos_char_set = (char_set*) hold_pos.in_node->get_model()->inq_state(hold_pos.in_node, tml_char_set_prop);

            if (tml_word_wrap_break_found
              /* but a break exists in previously formatted material */

              && ( ! 
                (hold_pos_char_set->is_htab(((unsigned char*) hold_pos.in_node->get_contents()) + hold_pos.char_index)
              || in_char_set->is_newline(((unsigned char *) format_chars) + char_off + end_index)
              || in_char_set->cs_isspace(((unsigned char *) format_chars) + char_off + end_index)))) {
                /* and the current segment does not end with a break: */

                /* reject unbroken segment at EOL */
                delete return_line;
                return 0;
                }

            /* else: */
                /*     no word break found: allow return of final segment     */
                /*       of a first word too long for line. Relevant when     */
                /*       first word spans more than one node and therefore    */
                /*       more than one tline.                                 */

            /* indicate that this line is complete */
            return_line->set_wide(return_line->get_wide() + space_left);
            /* return last segment that fits */
            return return_line;
            }

        else if (tml_word_wrap_break_found) {
            /* additional material fit and break was found */

            /* get information about that material */
            attach_node = attach_line->get_start().in_node;
            attach_end_char = ((char*) attach_node->get_contents()) + attach_line->get_count();
            attach_end_char -= attach_char_set->get_char_size();
            attach_end_next = attach_line->get_end().get_next();
            /* check if there is no next formatting position after material that fit */
            /*   OR if next formatting position after material that fit begins with tab */
            end_or_tab = 0;
            if ( ( ! attach_end_next.in_node)
              || ( ! attach_end_next.in_node->can_format())
              || ((attach_end_next.in_node->get_type()) != gobj))
                end_or_tab = -1;
            else {
                attach_end_next_char = ((char*) attach_end_next.in_node->get_contents()) + attach_end_next.char_index;
                if (this == attach_end_next.in_node)
                    attach_end_next_char_set = in_char_set;
                else
                    attach_end_next_char_set = (char_set*) attach_end_next.in_node->get_model()->inq_state(attach_end_next.in_node, tml_char_set_prop);
                if (attach_end_next_char_set->is_htab((unsigned char*) attach_end_next_char))
                    end_or_tab = -1;
                }

            if (( ! attach_line->get_right() )
              /* additional material is not followed by anything */

              && (!
                (end_or_tab
              || attach_char_set->is_newline((unsigned char*) attach_end_char)
              || attach_char_set->cs_isspace((unsigned char*) attach_end_char)))) {
              /* and does not end with break */
            
                /* reject unbroken segment at EOL */
                delete attach_line;
                /* indicate that this line is complete */
                return_line->set_wide(return_line->get_wide() + space_left);
                return return_line;
                }
            }
        }
    else if ( ! attach_line)
        /* word wrap not enabled and no additional material formatted */
        return return_line;

    /* adjust baseline offset to be larger of baseline offsets */
    /*   of previously formatted and newly formatted lines */
    old_up = return_line->get_baseline();
    old_down = return_line->get_hi() - old_up;
    new_up = attach_line->get_baseline();
    new_down = attach_line->get_hi() - new_up;
    if (new_up > old_up){
        return_line->adjust_baseline(new_up);
        old_up = new_up;
        }
    else if (old_up > new_up){
        attach_line->adjust_baseline(old_up);
        new_up = old_up;
        }
                                                     
    /*  set line heights to sum of greater baseline offset */
    /*    and greater descender depth of the two lines*/
    if (new_down > old_down){
        return_line->set_hi((short) (new_up + new_down));
        attach_line->set_hi((short) (new_up + new_down));
        old_down = new_down;
        }
    else if (old_down > new_down){
        return_line->set_hi((short) (old_up + old_down));
        attach_line->set_hi((short) (old_up + old_down));
        }

    return_line->set_right(attach_line);
    return_line->set_wide(return_line->get_wide() + attach_line->get_wide());
    return return_line;
}


/* reverse format a line of text until there is no more space */
/* the type of formatting performed is determined by the value of the wrap property*/
cline* gnode::tamrof_line(format* in_format, long start_index, long space_hi, long space_wide, long first_concep_pixel)
{
    register tline     *return_line,
                       *attach_line;
    register font_tbl  *use_font;
    register long       old_up,
                        old_down,
                        new_up,
                        new_down;
             long      *format_value,
                        format_length,
                       *ff_state,
                        end_index,
                        end_or_tab,
                        ends_line;
             short      space_left,
                        space_used;
             model_pos  hold_pos;
             char      *ff_found = NULL, *ff_start;
             char      *prior_contents;
             char      *attach_char;
             char_set  *in_char_set = NULL;
             char_set  *hold_pos_char_set;
             char_set  *attach_char_set;
             char_set  *attach_start_prior_char_set;
             node      *attach_node;
             model_pos  attach_start_prior;
             char      *attach_start_char;
             char      *attach_start_prior_char;
             long       char_size,
                        i,
                        tab_pixel_off;
    unsigned long       char_off;
             lrud_dir   attach_dir;
             lrud_dir   text_dir;
             tline     *hold_line;

    if (start_index >= count)
        /* request to format data beyond end of node */
        return 0;

    /* get appropriate character set and bytes per character */
    if (get_model())
        in_char_set = (char_set *) get_model()->inq_state(this, tml_char_set_prop);
    if (NULL == in_char_set)
        in_char_set = get_dflt_union_char_set()->get_sets();
    char_size = in_char_set->get_char_size();

    /* get number of characters to format */
    format_length = start_index + char_size;

    /* get value of formfeed property */
    ff_state = (long*) in_format->inq_state(this, tml_form_feed_prop);
            
    /* seek formfeed in string to format */
    /* if found, truncate string at formfeed */
    if (ff_state && *ff_state) {
        for (ff_start = ff_found = ((char*) goo_ptr) + start_index;
          ff_found > (ff_start - format_length);
          ff_found -= char_size) {
            if (in_char_set->is_formfeed((unsigned char *)ff_found)) {
                format_length = ff_start - ff_found;
                break;
                }
            }
        if (ff_found == (ff_start - format_length))
            ff_found = NULL;
        }

    /* get value of formatting property */
    format_value = (long*) in_format->inq_state(this, tml_wrap_prop);

    /* get appropriate font */
    if (NULL == (use_font = (font_tbl*) in_format->inq_state(this, tml_font_prop))) {
        if (NULL == (use_font = (font_tbl*) (get_model()->get_state()->get_default(tml_font_prop))))
            /* no font: error */
            return 0;
        }
                                 
    /* set tab offset value for trailing horizontal tab characters */
    /* obtain goo offset to back up over trailing tab characters */
    for (i = format_length, char_off = 0; 
      (i && in_char_set->is_htab(((unsigned char*) goo_ptr) + i - char_size));
      i -= char_size, char_off += char_size)
        ;
    format_length = i;
    if (char_off) {
        /* tab beyond first special tab when default tab value is zero */
        /*   clears line to BOL */
        if (-1 == (tab_pixel_off = calc_tab_pixel_off_backward(this, 
                                                               in_format, 
                                                               char_off / char_size, 
                                                               (first_concep_pixel) ? first_concep_pixel : space_wide)))
            tab_pixel_off = space_wide;
        }
    else
        tab_pixel_off = 0;

    /* get space limits for formatting */
    space_left = (short) (space_wide - tab_pixel_off);
    space_used = 0;

    /* skip trailing newline when starting at a line end */
    if ((0 == first_concep_pixel) && in_char_set->is_newline(((unsigned char*) goo_ptr) + format_length - char_size)) {
        format_length -= char_size;
        ends_line = -1;
        }
    else
        ends_line = 0;

    if (ff_found && (0 == format_length))
        /* first character is formfeed */
        end_index = 1;
    else if (0 == space_left)
        /* tabs used up remaining space */
        end_index = format_length;
    else {
        if (NULL == format_value)
            end_index = tamrof_no_wrap((char*) goo_ptr, format_length, use_font, space_left, space_used, in_char_set);
        else {
            switch (*format_value) {
                case no_wrap:
                    end_index = tamrof_no_wrap((char*) goo_ptr, format_length, use_font, space_left, space_used, in_char_set);
                    break;
                case char_wrap:
                    end_index = tamrof_char_wrap((char*) goo_ptr, format_length, use_font, space_left, space_used, in_char_set);
                    break;
                case word_wrap:
                    if (0 == first_concep_pixel)
                        /* indicate no break found when new line begun */
                        /*   at largest conceptual page x */
                        tml_word_wrap_break_found = 0;
                    end_index = tamrof_word_wrap((char*) goo_ptr, format_length, use_font, space_left, space_used, in_char_set);
                    if ((0 == first_concep_pixel) && (0 > end_index)) {
                        /* format length characters are more than enough to fill line */
                        /*   and last word on line is too long for line width: */
                        /*   use character wrap */
                        space_left = (short) (space_wide - tab_pixel_off);
                        space_used = 0;
                        end_index = tamrof_char_wrap((char*) goo_ptr, format_length, use_font, space_left, space_used, in_char_set);
                        /* indicate best fit achieved */
                        space_left = 0;
                        }
                    break;
                default:
                    end_index = tamrof_no_wrap((char*) goo_ptr, format_length, use_font, space_left, space_used, in_char_set);
                    break;
                }
            }
        }
                     
    /* if end_index < 0, error */
    if (end_index < 0)
        return 0;

    return_line = new tline(in_format);
    hold_pos.in_node = this;
    hold_pos.char_index = end_index;
    return_line->set_start(hold_pos);
    return_line->set_font(use_font);
    return_line->set_baseline(use_font->get_baseline());
    return_line->set_start_x_off(0);
                     
    if (ff_found && (0 == format_length)) {
        /* formfeed encountered: make rest of page blank */
        return_line->set_type(cl_gutter);
        return_line->set_wide((short) space_wide);
        return_line->set_text_width(space_wide);
        return_line->set_hi((short) space_hi);
        return_line->set_count(char_size);
        space_left = 0;
        }
    else {
        /* normal formatting */
        return_line->set_type(cl_text);
        if (end_index || format_length) {
            /* segment containing more than one nontab character */
            /* OR segment consisting of tabs and one character */
            return_line->set_count(start_index - end_index + char_size);
            if (ends_line && ((NULL == format_value) || (no_wrap == *format_value)))
                /* line end in no wrap; set to full width available */
                return_line->set_wide((short) space_wide);
            else
                return_line->set_wide((short) (space_wide - space_left));
            }
        else if ((0 == format_length) && (in_char_set->is_newline(((unsigned char*) goo_ptr) + end_index))) {
            /* segment consisting of newline only */
            return_line->set_count(char_size);
            return_line->set_wide(0);
            }
        else {
            /* segment consisting of tabs only */
            return_line->set_count(char_off);
            return_line->set_wide((short) (space_wide - space_left));
            }
        return_line->set_hi((short) use_font->get_hi());
        return_line->set_text_width(space_used);
        }
                           
    /* no space left */
    if (0 == space_left)
        return return_line;

    /* NOTE: when                                             */
    /*   (1) no recursion has taken place because             */
    /*   (2) format_length contains more than enough material */
    /*       to fill line,                                    */
    /*   (3) last word is too long for line, and              */
    /*   (4) word_wrap is enabled,                            */
    /*  the character-wrapped final portion of the last       */
    /*  word is returned here, because space_left is set to   */
    /*  zero in switch statement above. See below after       */
    /*  recursive call for code that handles case in which    */
    /*  (3) and (4), but not (1) and (2), obtain.             */

    /* get prior model position before formatted material */
    hold_pos.char_index = end_index;
    hold_pos = hold_pos.get_prior();

    /* no more goo */
    if (!hold_pos.in_node || !hold_pos.in_node->can_format() || ((hold_pos.in_node->get_type()) != gobj))
        return return_line;
    if (!(prior_contents = (char *)(hold_pos.in_node->get_contents())))
        return return_line;

    /* more goo but font too high for line */
    if (hold_pos.in_node->get_format_hi(in_format) > space_hi) {
        delete return_line;
        return 0;
        }

    /* next character to be formatted is a newline */
    attach_char_set = (char_set*) hold_pos.in_node->get_model()->inq_state(hold_pos.in_node, tml_char_set_prop);
    attach_char = ((char*) hold_pos.in_node->get_contents() + hold_pos.char_index);
    if (attach_char_set->is_newline((unsigned char*)attach_char)) {
        return_line->set_wide((short) space_wide);
        return return_line;
        }

    /* in backward formatting, first_concep_pixel is the last available pixel */
    /* in recursive calls it cannot be zero since when no space remains, this function is not called */
    if (0 == first_concep_pixel)
        first_concep_pixel = space_wide;

    /* !!! RECURSIVE CALL !!! RECURSIVE CALL !!! RECURSIVE CALL !!! */
    /* space remains, goo is available, line end not encountered, font fits in vertical space: try to format more */
    /* the last argument gives the conceptual line pixel at which additional formatting will                      */
    /*     conceptual line pixel at which this formatting attempt began       (first_concep_pixel                 */
    /*   - number of conceptual line pixels used through last formatting call (space_used)                        */
    /*   - tab offset in conceptual pixels                                    (tab_pixel_off)                     */
    if (::right == (attach_dir = attach_char_set->get_dir_to_char((unsigned char*) attach_char)))
        attach_line = (tline*) hold_pos.in_node->tamrof_line(in_format, hold_pos.char_index, space_hi, space_left, first_concep_pixel - space_used - tab_pixel_off);
    else if (::left == attach_dir)
        attach_line = (tline*) hold_pos.in_node->format_line(in_format, hold_pos.char_index, space_hi, space_left, first_concep_pixel - space_used - tab_pixel_off);
    if (format_value && (word_wrap == *format_value)) {
        /* word wrap is enabled ... */

        if ( ! attach_line) {
            /* no more space: no additional material formatted */

            if (this == hold_pos.in_node)
                hold_pos_char_set = in_char_set;
            else
                hold_pos_char_set = (char_set*) hold_pos.in_node->get_model()->inq_state(hold_pos.in_node, tml_char_set_prop);

            if (tml_word_wrap_break_found
              /* but a break exists in previously formatted material */

              && ( ! 
                (hold_pos_char_set->is_htab(((unsigned char*) hold_pos.in_node->get_contents()) + hold_pos.char_index)
              || in_char_set->is_newline(((unsigned char*) goo_ptr) + end_index)
              || in_char_set->cs_isspace(((unsigned char*) goo_ptr) + end_index)))) {
                /* and the current segment does not begin with a break: */

                /* reject unbroken segment at BOL */
                delete return_line;
                return 0;
                }

            /* else: */
                /*     no word break found: allow return of initial segment   */
                /*       of a last word too long for line. Relevant when      */
                /*       last word spans more than one node and therefore     */
                /*       more than one tline.                                 */

            /* indicate that this line is complete */
            return_line->set_wide(return_line->get_wide() + space_left);
            /* return last segment that fits */
            return return_line;
            }

        else if (tml_word_wrap_break_found) {
            /* additional material fit and break was found */

            /* get information about that material */
            attach_node = attach_line->get_end().in_node;
            attach_start_char = ((char*) attach_node->get_contents()) + attach_line->get_start().char_index;
            attach_start_prior = attach_line->get_start().get_prior();
            /* check if there is no prior formatting position before material that fit */
            /*   OR if prior formatting position before material that fit starts with tab */
            end_or_tab = 0;
            if ( ( ! attach_start_prior.in_node)
              || ( ! attach_start_prior.in_node->can_format())
              || ((attach_start_prior.in_node->get_type()) != gobj))
                end_or_tab = -1;
            else {
                attach_start_prior_char = ((char*) attach_start_prior.in_node->get_contents()) + attach_start_prior.char_index;
                if (this == attach_start_prior.in_node)
                    attach_start_prior_char_set = in_char_set;
                else
                    attach_start_prior_char_set = (char_set*) attach_start_prior.in_node->get_model()->inq_state(attach_start_prior.in_node, tml_char_set_prop);
                if (attach_start_prior_char_set->is_htab((unsigned char *) attach_start_prior_char))
                    end_or_tab = -1;
                }

            if (( ! attach_line->get_left())
              /* additional material is not preceded by anything */

              && (!
                (end_or_tab
              || attach_char_set->is_newline((unsigned char *) attach_start_char)
              || attach_char_set->cs_isspace((unsigned char *) attach_start_char)))) {
              /* and does not begin with break */
            
                /* reject unbroken segment at BOL */
                delete attach_line;
                /* indicate that this line is complete */
                return_line->set_wide(return_line->get_wide() + space_left);
                return return_line;
                }
            }
        }
    else if ( ! attach_line)
        /* word wrap not enabled and no additional material formatted */
        return return_line;
                                    
    /* adjust baseline offset to be larger of baseline offsets */
    /*   of previously formatted and newly formatted lines */
    old_up = return_line->get_baseline();
    old_down = return_line->get_hi() - old_up;
    new_up = attach_line->get_baseline();
    new_down = attach_line->get_hi() - new_up;
    if (new_up > old_up){
        return_line->adjust_baseline(new_up);
        old_up = new_up;
        }
    else if (old_up > new_up){
        attach_line->adjust_baseline(old_up);
        new_up = old_up;
        }
                                                     
    /* set line heights to sum of greater baseline offset */
    /*   and greater descender depth of the two lines */
    if (new_down > old_down){
        return_line->set_hi((short) (new_up + new_down));
        attach_line->set_hi((short) (new_up + new_down));
        old_down = new_down;
        }
    else if (old_down > new_down){
        return_line->set_hi((short) (old_up + old_down));
        attach_line->set_hi((short) (old_up + old_down));
        }

    text_dir = in_char_set->get_dir_to_char(((unsigned char*) goo_ptr) + end_index);
    if ((::left == text_dir) && (::right == attach_dir)) {
        /* leftward text is tamrof'd when page formatting direction is forward */
        /* in that case, an attached rightward text segment must be placed AFTER */
        /*   the leftward segment to which it is to be attached */
        for (hold_line = return_line; hold_line->get_right(); hold_line = (tline*) hold_line->get_right())
            ;
        hold_line->set_right(attach_line);
        return_line->set_wide(return_line->get_wide() + attach_line->get_wide());
        return return_line;
        }
    else {
        attach_line->set_right(return_line);
        attach_line->set_wide(return_line->get_wide() + attach_line->get_wide());
        return attach_line;
        }
}

/*****************************************************************************************************

        Member functions for derived class pgnode

*****************************************************************************************************/

/*constructor for objects of type pgnode*/
pgnode::pgnode(model *to_model, long size) : (to_model, size ? size : (long) sizeof(pgnode))
{
    register pgnode *new_node;

    size = size ? size : sizeof(pgnode);

    if (!this){
        if (to_model){
            new_node = (pgnode *)(to_model->get_current_oheap()->get_tf_node(size));
            if (new_node){
                if (new_node->up){
                    if (new_node->left)
                        new_node->left->set_right(new_node->up);
                    else
                        to_model->get_current_oheap()->tf_node = new_node->up;
                    new_node->up->set_right(new_node->get_right());
                    if (new_node->right)
                        new_node->right->set_left(new_node->up);
                    }
                else {
                    if (new_node->left)
                        new_node->left->set_right(new_node->get_right());
                    else
                        to_model->get_current_oheap()->tf_node = new_node->right;
                    if (new_node->right)
                        new_node->right->set_left(new_node->get_left());
                    }
                }
            else {
                new_node = (pgnode *)(to_model->get_current_oheap()->alloc(size));
                if (!new_node){
                    return;
                    }
                }
            this = new_node;
            }
        else {
            new_node = (pgnode *) (new char [size]);
            if (!new_node)
                return;
            else
                this = new_node;
            }
        }
    else
        this = this;

    n_type = pgobj;
}

/*copy a pgnode to another model, or to same model if model argument is NULL*/
node* pgnode::copy(model* to_model)
{
    register pgnode* new_node;

    if (!to_model) to_model = get_model();
    new_node = new pgnode(to_model);
    if (!new_node) return 0;

    new_node->n_type = pgobj;
    if (properties)
        new_node->prop_attach(properties->copy_all(to_model));

    new_node->set_char_count(get_char_count());
    if (to_model)
        new_node->set_goo_ptr(to_model->get_current_gheap()->char_alloc((char*) get_contents(), get_char_count()));
    else {
        new_node->set_goo_ptr(new char[get_char_count()]);
        tml_os_byte_copy((char*) get_contents(), (char*) new_node->get_contents(), get_char_count());
        }

    return(new_node);
}

/*format page number*/
/* if a character set is in force at the node, use the digit_ranges of that character set */
/*   to determine the single- or multibyte characters that represent the Hindu-Arabic */
/*   digits in the page number. Assumes that digit_ranges for that character set specify */
/*   characters representing the ten Hindu-Arabic digits in increasing order of the digits' */
/*   value (i.e. the character representing 'zero' through the character representing 'nine'.) */
cline* pgnode::format_line(format* in_format, long start_index, long space_hi, long space_wide, long first_concep_pixel)
{
    unsigned long   i;
    char_set*       in_char_set = NULL;
    unsigned long   char_size;
    unsigned long   temp_num;
    unsigned long   num_digits;
    unsigned long   num_bytes;
    unsigned long   buf_index;
    unsigned long   dividend;
    unsigned long   remainder;
    char_range_set* range_set;
    char_range*     range;
    char            pg_num_buf[100]; // NOTE: limit of 100 bytes (minimum 25 digits,
                                     //   at maximum of four bytes per character
                                     //   in character set manager) for a page number

    // get applicable character set
    if (get_model())
        in_char_set = (char_set*) get_model()->inq_state(this, tml_char_set_prop);
    if (NULL == in_char_set)
        in_char_set = get_dflt_union_char_set()->get_sets();
    char_size = in_char_set->get_char_size();
    range_set = in_char_set->get_digit_ranges();
    range = range_set->get_ranges();

    // get number of digits in page number
    temp_num = start_index;
    num_digits = 0;
    while (temp_num) {
        temp_num = temp_num / 10;
        ++num_digits;
        }

    // write null terminator in page number buffer
    num_bytes = (num_digits + 1) * char_size;
    if (num_bytes > 100) {
        printf("pgnode::format_line: more than maximum 100 bytes needed for page number\n");
        return NULL;
        }
    buf_index = num_bytes;
    in_char_set->get_terminator((unsigned char*)(pg_num_buf + (buf_index -= char_size)));

    // copy the digits into the buffer, starting with last digit in number
    temp_num = start_index;
    remainder = temp_num;
    for (i = 0; i < num_digits; i++) {
        dividend = temp_num / 10;
        remainder = temp_num - (dividend * 10);
        temp_num = dividend;
        remainder += range->get_low_index();
        range_set->lookup(remainder, ((unsigned char*) (pg_num_buf + (buf_index -= char_size))));
        }

    // there is already sufficient space allocated for the digits
    if (num_digits <= get_char_count()) {
        in_char_set->cs_strncpy((unsigned char*) get_contents(), (unsigned char*) pg_num_buf, num_digits);
        set_char_count(num_bytes - char_size);
        }
    // no space yet allocated for digits, or not enough
    else
        gnode::attach_data ((void*) pg_num_buf, num_bytes - char_size);

    return gnode::format_line(in_format, 0, space_hi, space_wide, first_concep_pixel);
}

/* format a page number expressed in Hindu-Arabic digits in right-to-left font */
cline* pgnode::tamrof_line(format* in_format, long start_index, long space_hi, long space_wide, long first_concep_pixel)
{
    unsigned long   i;
    char_set*       in_char_set = NULL;
    unsigned long   char_size;
    unsigned long   temp_num;
    unsigned long   num_digits;
    unsigned long   num_bytes;
    unsigned long   buf_index;
    unsigned long   dividend;
    unsigned long   remainder;
    char_range_set* range_set;
    char_range*     range;
    char            pg_num_buf[100]; // NOTE: limit of 100 bytes (minimum 25 digits,
                                     //   at maximum of four bytes per character
                                     //   in character set manager) for a page number

    // get applicable character set
    if (get_model())
        in_char_set = (char_set*) get_model()->inq_state(this, tml_char_set_prop);
    if (NULL == in_char_set)
        in_char_set = get_dflt_union_char_set()->get_sets();
    char_size = in_char_set->get_char_size();
    range_set = in_char_set->get_digit_ranges();
    range = range_set->get_ranges();

    // get number of digits in page number
    temp_num = start_index;
    num_digits = 0;
    while (temp_num) {
        temp_num = temp_num / 10;
        ++num_digits;
        }

    // write null terminator in page number buffer
    num_bytes = (num_digits + 1) * char_size;
    if (num_bytes > 100) {
        printf("pgnode::tamrof_line: more than maximum 100 bytes needed for page number\n");
        return NULL;
        }
    buf_index = num_bytes;
    in_char_set->get_terminator((unsigned char*) (pg_num_buf + (buf_index -= char_size)));

    // copy the digits into the buffer, starting with last digit in number
    temp_num = start_index;
    remainder = temp_num;
    for (i = 0; i < num_digits; i++) {
        dividend = temp_num / 10;
        remainder = temp_num - (dividend * 10);
        temp_num = dividend;
        remainder += range->get_low_index();
        range_set->lookup(remainder, ((unsigned char*) (pg_num_buf + (buf_index -= char_size))));
        }

    // there is already sufficient space allocated for the digits
    if (num_digits <= get_char_count()) {
        in_char_set->cs_strncpy((unsigned char*) get_contents(),(unsigned char*) pg_num_buf, num_digits);
        set_char_count(num_bytes - char_size);
        }
    // no space yet allocated for digits, or not enough
    else
        gnode::attach_data ((void*) pg_num_buf, num_bytes - char_size);

    return gnode::tamrof_line(in_format, num_bytes - char_size, space_hi, space_wide, first_concep_pixel);
}

