/**************************************************************************************
    tml_dp_x_graph.c
    Implementation of the device-dependent service routines for
      XWindows-based systems

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <stdio.h>
#include <string.h>

#ifndef apollo
#ifdef bsd
#include <strings.h>
#endif
#endif

#include <X11/Xlib.h>
#include <tml/tml_dp_x_graph.h>
#include <tml/tml_c_types.h>
#include <tml/tml_os_depends.h>

#define MAXNAMLEN 1024

#define TML_MAX_X_RECT 32      /* number plucked from "Introduction to the X Window System" */
#define TML_WHOLE_WINDOW -1    /* number of available XRects for clipping exceeded; repaint whole window */

/* service types for calls to x_serve_default_font_name() */

#define get_default_name      1
#define set_default_name      2

/* service types for calls to x_serve_font_handle_list() */

#define get_font_handle            1
#define release_font_handle        2
#define set_font_display           3
#define unload_font                4


/******************************************************************************
    Globals for font management
******************************************************************************/

/* To make these globals per-process, must be declared with attribute specifier */
/*   for Apollo builds to allow use in global and inlib'd libraries */

/* tml_default_font_name: default font name */
/* tml_display:           the display that this process is addressing */
/* tml_fhandle_list_head: list of active font handles */

#ifdef apollo
char* tml_default_font_name __attribute((section(tml_default_font_name))) = "fixed";
Display* tml_display __attribute((section(tml_display))) = NULL;
struct fhandle *tml_fhandle_list_head __attribute((section(tml_fhandle_list_head))) = NULL;
#else
char* tml_default_font_name = "fixed";
Display* tml_display = NULL;
struct fhandle *tml_fhandle_list_head = NULL;
#endif

/*****************************************************************************
    Function to get and set name of default font
*****************************************************************************/

#ifdef not_vanilla_C
    char *get_x_default_font_name(void*, long*);
#else
    char *get_x_default_font_name();
#endif

char *x_serve_default_font_name(service, display, length, font_name)
    long     service;
    Display *display;
    long    *length;
    char    *font_name;
{

    char* *name_table;
    long actual_count_return;

    switch(service) {

    case get_default_name:
                              
        if (!display)
            display = tml_display;
        if (NULL == display) {
            *length = 0;
            printf("x_serve_default_font_name: failure to call tml_dp_set_font_display()\n");
            return NULL;
            }
        if (NULL == tml_default_font_name) {
            if (NULL == display) {
                *length = 0;
                return(NULL);
                }
            name_table = XListFonts(display, "*", 1, &actual_count_return);
            if (1 != actual_count_return)
                return(NULL);
            *length = strlen(*(name_table + 0));
            tml_default_font_name = tml_c_tml_os_alloc(*length + 1, (long) sizeof(char));
            tml_default_font_name = tml_c_tml_os_byte_copy(*(name_table + 0), tml_default_font_name, *length);
            *(tml_default_font_name + *length) = '\0';
            XFreeFontNames(name_table);
            }
        else
            *length = strlen(tml_default_font_name);
        return(tml_default_font_name);

    case set_default_name:

        if (NULL == font_name) {
            if (NULL != tml_default_font_name)
                tml_c_tml_os_free(tml_default_font_name);
            tml_default_font_name = get_x_default_font_name(display, length);
            }
        else
            tml_default_font_name = font_name;

        return(tml_default_font_name);
    }
}


/* gets default font name IF it has already been set. Otherwise, returns NULL */
char* tml_dp_x_get_default_font(length)
    long *length;
{
    return(x_serve_default_font_name(get_default_name, NULL, length, NULL));
}


/* get the default font name. If default is NULL, call XListFonts() to get first */
/*   font returned and set default to that font. */
char *get_x_default_font_name(display, length)
    void    *display;
    long    *length;
{
    return(x_serve_default_font_name(get_default_name, display, length, NULL));
}


/* set the default font name. If the font_name argument is NULL, call */
/* XListFonts() to get first font returned and set default to that font. */
/* Otherwise set to named font, which is assumed to exist. */
char *set_x_default_font_name(display, length, font_name)
    Display *display;
    long    *length;
    char    *font_name;
{
    return(x_serve_default_font_name(set_default_name, display, length, font_name));
}

/*****************************************************************************
    Structure for font management
*****************************************************************************/

struct fhandle {
    struct fhandle *next;
    struct fhandle *prev;
    char *font_name;
    short name_len;
    XFontStruct *font_info;
    short is16;
    short interline_space;
    short icspace;
    short is_fixed;
    short ref_count;
    };

/*****************************************************************************
    Structures for window management
*****************************************************************************/

struct bm_info {
    struct window_info     *bitmap_id;
    char                   *viewport_id;
    long                    size_x;         /* viewport and pixmap size in points */
    long                    size_y;
    long                    new_x;          /* same info from resize event. If differs */
    long                    new_y;          /* from size_x/size_y, must reformat viewport */
    long                    depth;
    struct fhandle         *font_id;
    long                    last_direction; /* direction of last text write */
    long                    last_ascent;    /* ascent of last text write */
    long                    last_descent;   /* descent of last text write */
    XCharStruct             last_extent;    /* all possible info on last text write */
    long                    exp_nrect;
    XRectangle              exp_rect[TML_MAX_X_RECT];
    long                    exp_ulx;
    long                    exp_uly;
    long                    exp_lrx;
    long                    exp_lry;
    };

/******************************************************************************
    Functions for manipulating font handles
******************************************************************************/


/* allocate and release font handles. */
/* return NULL on failure, struct fhandle * otherwise */
struct fhandle *x_serve_font_handle_list(service, font_name, name_len, font_handle)
    long            service;
    char           *font_name;
    short           name_len;
    struct fhandle *font_handle;
{
    struct fhandle *cur_fhandle;
    char *cur_font_name;
    char full_path[MAXNAMLEN];

    switch (service) {

        case set_font_display:

            tml_display = (Display *) font_name;
            return(NULL);

        case get_font_handle:

            /* note: depends on tml_c_tml_os_alloc()'s nulling of space it allocates */
                                                   
            cur_fhandle = tml_fhandle_list_head;
        
            /* see if font is already catalogued, and if so, is it still loaded */
            /*   if not loaded, load and query it */

            while (NULL != cur_fhandle) {
                cur_font_name = cur_fhandle->font_name;
                if ((name_len == cur_fhandle->name_len) && (0 == strncmp(font_name, cur_font_name, name_len))) {
                    if (NULL == cur_fhandle->font_info)
                        cur_fhandle->font_info = XLoadQueryFont(tml_display, cur_fhandle->font_name); 
                    cur_fhandle->ref_count++;
                    return(cur_fhandle);
                    }
                cur_fhandle = cur_fhandle->next;
                }
        
            /* font not already loaded */
        
            cur_fhandle = (struct fhandle *) tml_c_tml_os_alloc(1, (long) sizeof(struct fhandle));
            cur_fhandle->font_name = (char *) tml_c_tml_os_alloc(name_len + 1, (long) sizeof(char));
            strncpy(cur_fhandle->font_name, font_name, name_len);
            cur_fhandle->name_len = name_len;
            cur_fhandle->font_info = XLoadQueryFont(tml_display, cur_fhandle->font_name);

            /* Check to see if font is valid before updating list */

            if (!cur_fhandle->font_info)
            {
                tml_c_tml_os_free(cur_fhandle->font_name);
                tml_c_tml_os_free((char*)cur_fhandle);
                return 0;
            }

            cur_fhandle->prev = NULL;
            cur_fhandle->next = tml_fhandle_list_head;
            if (tml_fhandle_list_head)
                tml_fhandle_list_head->prev = cur_fhandle;
            tml_fhandle_list_head = cur_fhandle;
            cur_fhandle->is16 = ((0 == cur_fhandle->font_info->min_byte1)
                && (0 == cur_fhandle->font_info->max_byte1)) ? 0 : 1;
            cur_fhandle->interline_space = cur_fhandle->font_info->ascent
                                           + cur_fhandle->font_info->descent;
            cur_fhandle->icspace = 0;
            cur_fhandle->ref_count = 1;
            return(cur_fhandle);
            
        case unload_font:

            cur_fhandle = (struct fhandle *) font_name;
            XFreeFont(tml_display, cur_fhandle->font_info);
            cur_fhandle->font_info = NULL;

        case release_font_handle:

            font_handle->ref_count--;
            if (!font_handle->ref_count) {
                if (NULL != font_handle->prev)
                    font_handle->prev->next = font_handle->next;
                else
                    tml_fhandle_list_head = font_handle->next;
                if (NULL != font_handle->next)
                    font_handle->next->prev = font_handle->prev;
                tml_c_tml_os_free(font_handle->font_name);
                tml_c_tml_os_free((char *)font_handle);
                }
            return(NULL);
        }
}

#ifdef not_vanilla_C
void tml_dp_set_font_display(display)
#else
tml_dp_set_font_display(display)
#endif
    Display *display;
{
    x_serve_font_handle_list(set_font_display, (char *) display, 0, NULL);
}

/* Get a font handle from a font name. Initialize the font_info member with the XFontStruct */
/*    returned from a call to XLoadQueryFont(). Seeks font names in static lists
/*    information direct from the font file. Look for the font under the name given, */
/*    and if it is not found, look for it under "/sys/dm/fonts/<name_given>". If still */
/*    not found, return error. If found, map in the first (long) sizeof(smd_$font_table_t) bytes, */
/*    copy them, and unmap the font file */
struct fhandle *tml_dp_x_get_fhandle(font_name, name_len)
    char  *font_name;
    short  name_len;
{
    if (tml_display)
        return(x_serve_font_handle_list(get_font_handle, font_name, name_len, NULL));
    else
        return 0;
}


/* load a font - a noop in X - included only to keep compiler from complaining */
long tml_dp_x_load_font(font_handle)
    struct fhandle *font_handle;
{
    return(1);
}


/* return bytes-per-character for the font */
long tml_dp_x_get_fcharsize(font_handle)
    struct fhandle *font_handle;
{
    XFontStruct *font_info;

    font_info = font_handle->font_info;
    return((font_info->min_byte1 || font_info->max_byte1) ? 2 : 1);
}


/* return number of characters in font from font handle */
long tml_dp_x_get_fcharnum(font_handle)
    struct fhandle *font_handle;
{
             long         num_chars;
             XFontStruct *font_info;
    unsigned long         lo_index;
    unsigned long         hi_index;

    font_info = font_handle->font_info;
    if (font_info->min_byte1 || font_info->max_byte1) {
        lo_index = ((font_info->min_byte1 << 8) + font_info->min_char_or_byte2);
        hi_index = ((font_info->max_byte1 << 8) + font_info->max_char_or_byte2);
        }
    else {
        lo_index = font_info->min_char_or_byte2;
        hi_index = font_info->max_char_or_byte2;
        }
    if (font_info->default_char < lo_index)
        lo_index = font_info->default_char;
    if (hi_index < font_info->default_char)
        hi_index = font_info->default_char;

    return 1 + hi_index - lo_index;
}


/* return index of first character in font */
long tml_dp_x_get_fchar1(font_handle)
    struct fhandle *font_handle;
{
    long low_code;

    if (font_handle->font_info->min_byte1) {
        low_code = (font_handle->font_info->min_byte1 << 8) + font_handle->font_info->min_char_or_byte2;
        if (low_code < font_handle->font_info->default_char)
            return low_code;
        return font_handle->font_info->default_char;
        }
    low_code = font_handle->font_info->min_char_or_byte2;
    if (low_code < font_handle->font_info->default_char)
        return low_code;
    return font_handle->font_info->default_char;
}


/* return font height from XFontStruct - sum of "logical" ascent and descent */
/*   - equals baseline-to-baseline distance, or "leading" (pronouced as in the metal) */
long tml_dp_x_get_fhi(font_handle)
    struct fhandle *font_handle;
{
    return(font_handle->font_info->ascent + font_handle->font_info->descent);
}


/* return pseudo-point size for font handle by returning tml_dp_x_get_fhi() value */
long tml_dp_x_get_fpoint(font_handle)
    struct fhandle *font_handle;
{
    return(tml_dp_x_get_fhi(font_handle));
}


/* return offset from top of font matrix to baseline */
long tml_dp_x_get_fbline(font_handle)
struct fhandle *font_handle;
{
    return(font_handle->font_info->ascent);
}


/* return character width in rasters for fixed fonts, zero for proportional-spaced fonts */
long tml_dp_x_is_fixed(font_handle)
    struct fhandle *font_handle;
{
    if (NULL == font_handle->font_info->per_char)
        return(font_handle->font_info->max_bounds.width);
    else
        return(0);
}


/* return pointer to array of character widths for proportional fonts, NULL otherwise */
short** tml_dp_x_get_fwidths(font_handle)
    struct fhandle *font_handle;
{
    register long          i, j;
    register short        *widths, 
                          *lbearings, 
                          *rbearings;
             short       **tbl_arr;
             long          num_slots, num_entries, lo_code, hi_code;
             XCharStruct  *char_info;

    if (NULL == font_handle->font_info->per_char)
        /* monospace font; no width table needed */
        return 0;

    /* number of entries in the per_char array for the font */
    num_entries = tml_dp_x_get_fcharnum(font_handle);

    if (0 == num_entries) {
        fprintf(stderr,"tml_dp_x_get_fwidths: font has no characters, hence no widths\n");
        return NULL;
        }
    if (tml_dp_x_is_fixed(font_handle))
        return NULL;

    /* array of arrays to be returned */
    tbl_arr = (short**) tml_c_tml_os_alloc(3, (long) sizeof(short*));

    /* array of widths to be returned for each indexed character */
    num_slots = num_entries;
    if (num_entries & 1)
        ++num_slots;
    tbl_arr[0] = widths    = (short *) tml_c_tml_os_alloc(num_slots, (long) sizeof(short));
    /* array of left bearings to be returned for each indexed character */
    tbl_arr[1] = lbearings = (short *) tml_c_tml_os_alloc(num_slots, (long) sizeof(short));
    /* array of right bearings to be returned for each indexed character */
    tbl_arr[2] = rbearings = (short *) tml_c_tml_os_alloc(num_slots, (long) sizeof(short));

    if ((font_handle->font_info->min_byte1) || font_handle->font_info->max_byte1) {
        lo_code = (font_handle->font_info->min_byte1 << 8) + font_handle->font_info->min_char_or_byte2;
        hi_code = (font_handle->font_info->max_byte1 << 8) + font_handle->font_info->max_char_or_byte2;
        }
    else {
        lo_code = font_handle->font_info->min_char_or_byte2;
        hi_code = font_handle->font_info->max_char_or_byte2;
        }

    char_info = font_handle->font_info->per_char;

    /* correct faulty font information */
    if (0 < font_handle->font_info->min_bounds.lbearing)
        font_handle->font_info->min_bounds.lbearing *= -1;
    if (0 > font_handle->font_info->min_bounds.rbearing)
        font_handle->font_info->min_bounds.rbearing *= -1;
    if (0 < font_handle->font_info->max_bounds.lbearing)
        font_handle->font_info->max_bounds.lbearing *= -1;
    if (0 > font_handle->font_info->max_bounds.rbearing)
        font_handle->font_info->max_bounds.rbearing *= -1;

    for (i = font_handle->font_info->default_char; i < lo_code; i++) {
        *(widths + i)    = font_handle->font_info->min_bounds.width;
        *(lbearings + i) = font_handle->font_info->min_bounds.lbearing;
        *(rbearings + i) = font_handle->font_info->min_bounds.rbearing;
        }
    for (i = lo_code; i <= hi_code; i++, char_info++) {
        *(widths + i)    = char_info->width;
        if (0 < char_info->lbearing)
            char_info->lbearing *= -1;
        else if ((0 > char_info->width)
          && (char_info->lbearing > char_info->width))
            char_info->lbearing = char_info->width;
        *(lbearings + i) = char_info->lbearing;
        if (0 > char_info->rbearing)
            char_info->rbearing *= -1;
        else if ((0 < char_info->width)
          && (char_info->rbearing < char_info->width))
            char_info->rbearing = char_info->width;
        *(rbearings + i) = char_info->rbearing;
        }
    for (i = hi_code + 1; i < num_entries; i++) {
        *(widths + i)    = font_handle->font_info->min_bounds.width;
        *(lbearings + i) = font_handle->font_info->min_bounds.lbearing;
        *(rbearings + i) = font_handle->font_info->min_bounds.rbearing;
        }

    return tbl_arr;
}


/* return intercharacter space */
/*   - zero in X until set by user (not an X concept - only in our font_handle) */
long tml_dp_x_get_ficspace(font_handle)
    struct fhandle *font_handle;
{
    return((long) font_handle->icspace);
}


/* set fixed intercharacter space for fixed fonts, nominal letterspacing for proportional fonts */
/* Implements track kerning. */
/* For left-to-right fonts, a positive value increases the distance between characters. */
/* For right-to-left fonts, a positive value decreases the distance between characters. */
/* Negative values have the opposite effects. */
long tml_dp_x_set_ficspace(font_handle, new_icspace)
struct fhandle *font_handle;
       long     new_icspace;
{
    font_handle->icspace = new_icspace;
    return(new_icspace);
}

/* get the default character, if any, for the font */
long tml_dp_x_get_fspace(font_handle)
    struct fhandle *font_handle;
{
    return font_handle->font_info->default_char;
}

#ifdef not_vanilla_C
void tml_dp_x_set_fspace(font_handle, new_space)
#else
tml_dp_x_set_fspace(font_handle, new_space)
#endif
    struct fhandle *font_handle;
{
    font_handle->font_info->default_char = new_space;
}

/* unload a font */
long tml_dp_x_funload(font_handle)
    struct fhandle *font_handle;
{
    if (tml_display) {
        x_serve_font_handle_list(unload_font, (char *) font_handle, 0, NULL);
        return 1;
        }
    else
        return -1;
}


tml_dp_x_frelease(font_handle)
    struct fhandle *font_handle;
{
    if (tml_display)
        x_serve_font_handle_list(release_font_handle, NULL, 0, font_handle);
}

/********************************************************************************
    Functions to manage bitmap information
********************************************************************************/

/* make a bitmap-viewport coordination structure - struct bm_info */
tml_c_bitmap_id tml_dp_x_make_graphic_handle(bitmap_id, viewport_id)
    struct window_info *bitmap_id;
    tml_c_view_port viewport_id;
{
    struct bm_info         *bitmap_info;
    Status                  status;
    tml_c_font_tbl          cur_font_tbl;
    char                   *font_name;
    short                   font_name_len;
    short                   current_font;

    if (!(bitmap_info = (struct bm_info *) tml_c_tml_os_alloc(1, (long) sizeof(struct bm_info))))
        return(0);
    
    bitmap_info->size_x = bitmap_id->creation_x_size;
    bitmap_info->size_y = bitmap_id->creation_y_size;
    bitmap_info->new_x = bitmap_id->creation_x_size;
    bitmap_info->new_y = bitmap_id->creation_y_size;
    bitmap_info->depth = bitmap_id->depth;

    bitmap_info->viewport_id = viewport_id;
    if (NULL == (cur_font_tbl = (tml_c_font_tbl) view_port_inq_state(viewport_id, tml_font_prop)))
        bitmap_info->font_id = NULL;
    else {
        font_name = font_tbl_get_name(cur_font_tbl);
        font_name_len = font_tbl_get_name_count(cur_font_tbl);
        bitmap_info->font_id = (struct fhandle *) tml_dp_x_get_fhandle(font_name, font_name_len);
        }
    bitmap_info->bitmap_id = (struct window_info *) tml_c_tml_os_alloc(1, (long) sizeof(struct window_info));
    if (!bitmap_info->bitmap_id){
        tml_c_tml_os_free((char *)bitmap_info);
        return 0;
        }
    *(bitmap_info->bitmap_id) = *bitmap_id;
    bitmap_info->viewport_id = viewport_id;
    bitmap_info->exp_nrect = 0;
    bitmap_info->exp_ulx = bitmap_info->size_x;
    bitmap_info->exp_uly = bitmap_info->size_y;
    bitmap_info->exp_lrx = 0;
    bitmap_info->exp_lry = 0;
    return((tml_c_bitmap_id) bitmap_info);

}

/* delete a bitmap/viewport coordination structure (struct bm_info) */
void tml_dp_x_delete_graphic_handle(bitmap_info)
    struct bm_info *bitmap_info;
{
    if (!bitmap_info)
        return;
    tml_c_tml_os_free((char *)bitmap_info->bitmap_id);
    tml_c_tml_os_free((char *)bitmap_info);

}

/********************************************************************************
    Functions to get information about a bitmap.
        The tml_dp_x_get_* functions access the bitmap_info
        associated with an argument bitmap_handle.
********************************************************************************/

long tml_dp_get_creation_x_size(bitmap_handle)
#ifdef not_vanilla_C
    void* bitmap_handle;
#else
    char* bitmap_handle;
#endif
{
    return ((struct window_info*) (((struct bm_info*) bitmap_handle)->bitmap_id))->creation_x_size;
}

long tml_dp_get_creation_y_size(bitmap_handle)
#ifdef not_vanilla_C
    void* bitmap_handle;
#else
    char* bitmap_handle;
#endif
{
    return ((struct window_info*) (((struct bm_info*) bitmap_handle)->bitmap_id))->creation_y_size;
}

#ifdef not_vanilla_C
void tml_dp_record_new_size(bitmap_handle, width, height)
    void* bitmap_handle;
#else
tml_dp_record_new_size(bitmap_handle, width, height)
    char* bitmap_handle;
#endif
    long width;
    long height;
{
    ((struct bm_info*) bitmap_handle)->new_x = width;
    ((struct bm_info*) bitmap_handle)->new_y = height;
}

void tml_dp_x_reset_dimen(bitmap_handle)
    struct bm_info *bitmap_handle;
{
    if ((bitmap_handle->new_x != bitmap_handle->size_x) || (bitmap_handle->new_y != bitmap_handle->size_y)) {
        /* initial condition */
        if ((bitmap_handle->exp_ulx == bitmap_handle->size_x) 
          && (bitmap_handle->exp_lrx == 0)
          && (bitmap_handle->exp_uly == bitmap_handle->size_y)
          && (bitmap_handle->exp_lry == 0)) {
            bitmap_handle->exp_ulx = bitmap_handle->new_x;
            bitmap_handle->exp_uly = bitmap_handle->new_y;
            }
        /* whole window is clip region */
        else if ((bitmap_handle->exp_ulx == 0)
          && (bitmap_handle->exp_lrx == bitmap_handle->size_x)
          && (bitmap_handle->exp_uly == 0)
          && (bitmap_handle->exp_lry == bitmap_handle->size_y)) {
            bitmap_handle->exp_lrx = bitmap_handle->new_x;
            bitmap_handle->exp_lry = bitmap_handle->new_y;
            }
        bitmap_handle->size_x = bitmap_handle->new_x;
        bitmap_handle->size_y = bitmap_handle->new_y;
        }
}

/*get the height of a bitmap from its bitmap_info if window size hasn't changed */
/*  if window size has changed, create new pixmap and return new height */
/*  or -1 if new pixmap couldn't be made */
long tml_dp_x_get_hi(bitmap_handle)
    struct bm_info *bitmap_handle;
{
    Status                  status;

    tml_dp_x_reset_dimen(bitmap_handle);
    return bitmap_handle->size_y;
}

/*get the width of a bitmap from its bitmap_info if window size hasn't changed */

long tml_dp_x_get_wide(bitmap_handle)
    struct bm_info *bitmap_handle;
{
    tml_dp_x_reset_dimen(bitmap_handle);
    return bitmap_handle->size_x;
}

/*get the foreground (text drawing) color of a bitmap from its bitmap_info*/
long tml_dp_x_get_foreg_id(bitmap_handle)
    struct bm_info *bitmap_handle;
{
    return(bitmap_handle->bitmap_id->gc->values.foreground);
}

/*get the (text) background color of a bitmap from its bitmap_info*/
long tml_dp_x_get_backg_id(bitmap_handle)
    struct bm_info *bitmap_handle;
{
    return(bitmap_handle->bitmap_id->gc->values.background);
}

/* get the background color for a viewport from its bitmap's bitmap_info */
/* - a noop in X */
long tml_dp_x_get_fill_backg_id(bitmap_handle)
    tml_c_bitmap_id bitmap_handle;
{
    return(0);
}

/* get the background color for a viewport from its bitmap's bitmap_info */
/* - a noop in X */
long tml_dp_x_get_fill_obj_id(bitmap_handle)
    tml_c_bitmap_id bitmap_handle;
{
    return(0);
}

/********************************************************************************
    Functions to set bitmap information.
********************************************************************************/

/*set the foreground (text drawing) color for a bitmap*/
void tml_dp_x_set_foreg_id(bitmap_handle, new_fg_id)
    struct bm_info *bitmap_handle;
    unsigned long   new_fg_id;
{
    if (new_fg_id != bitmap_handle->bitmap_id->gc->values.foreground)
        XSetForeground(bitmap_handle->bitmap_id->display, bitmap_handle->bitmap_id->gc, new_fg_id);
}

/*set the background value*/
void tml_dp_x_set_backg_id(bitmap_handle, new_bg_id)
    struct bm_info *bitmap_handle;
    unsigned long   new_bg_id;
{
    if (new_bg_id != bitmap_handle->bitmap_id->gc->values.background)
        XSetBackground(bitmap_handle->bitmap_id->display, bitmap_handle->bitmap_id->gc, new_bg_id);
}
                                                             
/* set the background fill color for the viewport */
/*  - a noop in X */
void tml_dp_x_set_fill_backg_id(bitmap_handle, new_fill_bg_id)
    struct bm_info *bitmap_handle;
    unsigned long   new_fill_bg_id;
{
}

/* set the object fill color for the viewport */
/*  - a noop in X */
void tml_dp_x_set_fill_obj_id(bitmap_handle, new_fill_obj_id)
    struct bm_info *bitmap_handle;
    unsigned long   new_fill_obj_id;
{
}

/*set and activate a clipping window in a bitmap*/
void tml_dp_x_set_clip(bitmap_handle, clip_ulx, clip_uly, clip_lrx, clip_lry)
    struct bm_info *bitmap_handle;
    long            clip_ulx,
                    clip_uly,
                    clip_lrx,
                    clip_lry;
{
    XRectangle clip_rect;

    clip_rect.x = clip_ulx;
    clip_rect.y = clip_uly;
    clip_rect.width =  clip_lrx - clip_ulx;
    clip_rect.height = clip_lry - clip_uly;

    /* handles the clip_origin members of the GC too */
    /* the zero values for clip x origin and clip y origin are correct */
    /*   clip_rect.x is clip_ulx which is zero-based */
    /*   clip_rect.y is clip_uly which is zero-based */
    XSetClipRectangles(bitmap_handle->bitmap_id->display, bitmap_handle->bitmap_id->gc, 0, 0, &clip_rect, 1, Unsorted);
}

/*deactivate clipping*/
void tml_dp_x_clear_clip(bitmap_handle)
    struct bm_info *bitmap_handle;
{
    register struct window_info *bm_id;

    bm_id = bitmap_handle->bitmap_id;
    XSetClipMask(bm_id->display, bm_id->gc, None);
    bitmap_handle->exp_nrect = 0;
    bitmap_handle->exp_ulx = bitmap_handle->size_x;
    bitmap_handle->exp_uly = bitmap_handle->size_y;
    bitmap_handle->exp_lrx = 0;
    bitmap_handle->exp_lry = 0;
}

/*clear a rectangular area of the bitmap*/
void tml_dp_x_clear(bitmap_handle, ulx, uly, lrx, lry)
    struct bm_info *bitmap_handle;
    long            ulx,
                    uly,
                    lrx,
                    lry;
{
    register struct window_info *bm_id;
                    long         foreground;

    /* Set GC function to GXcopy */
                     
    bm_id = bitmap_handle->bitmap_id;
    if (GXcopy != bm_id->gc->values.function)
        XSetFunction(bm_id->display, bm_id->gc, GXcopy);

    /* Put the text background color in the GC foreground */

    foreground = bm_id->gc->values.foreground;
    XSetForeground(bm_id->display, bm_id->gc, bm_id->gc->values.background);

    /* Clear specified part of window to background color */
    /* Using XFillRectangle allows TMLib to share windows that */
    /*   may not have text background as window background */

    XFillRectangle(bm_id->display, bm_id->window, bm_id->gc, ulx, uly, lrx - ulx, lry - uly);

    /* Restore the GC foreground */

    XSetForeground(bm_id->display, bm_id->gc, foreground);
}

/*copy a portion of the indicated bitmap from one location to another*/
/* THE GC FUNCTION IS ALTERED TO GXcopy */
void tml_dp_x_copy(bitmap_handle, src_ulx, src_uly, src_lrx, src_lry, dst_ulx, dst_uly)
    struct bm_info *bitmap_handle;
    long            src_ulx,
                    src_uly,
                    src_lrx,
                    src_lry,
                    dst_ulx,
                    dst_uly;
{
    register struct window_info *bm_id;

    bm_id = bitmap_handle->bitmap_id;

    /* Set GC function to GXcopy */
    if (GXcopy != bm_id->gc->values.function)
        XSetFunction(bm_id->display, bm_id->gc, GXcopy);

    /* copy the area to the window */

    XCopyArea(bm_id->display, bm_id->window, bm_id->window,
              bm_id->gc, src_ulx, src_uly, src_lrx - src_ulx, src_lry - src_uly,
              dst_ulx, dst_uly);
}

/*blit a portion of the bitmap to another position, using the indicated rop*/
/*when all six long arguments are 0, set the rop */
/* THE GC FUNCTION REMAINS ALTERED */
void tml_dp_x_blt(bitmap_handle, src_ulx, src_uly, src_lrx, src_lry, dst_ulx, dst_uly, rop_id)
    struct bm_info *bitmap_handle;
    long            src_ulx,
                    src_uly,
                    src_lrx,
                    src_lry,
                    dst_ulx,
                    dst_uly,
                    rop_id;      /* must be a GC function */
{
    register struct window_info *bm_id;
                                  
    bm_id = bitmap_handle->bitmap_id;

    /* Set GC function to new rop */
    if (rop_id != bm_id->gc->values.function)
        XSetFunction(bm_id->display, bm_id->gc, rop_id);

    if (src_ulx | src_uly | src_lrx | src_lry | dst_ulx | dst_uly) {
        XCopyArea(bm_id->display, bm_id->window, bm_id->window, bm_id->gc, 
                  src_ulx, src_uly, 
                  src_lrx - src_ulx, src_lry - src_uly, 
                  dst_ulx, dst_uly);
        }
}

/* draw a sting into the bitmap at the indicated position in the indicated font */
/* --return the length of the string */
/* uses transparent background to avoid overwriting characters that overlap */
/*   the bounding box */
/* THE INDICATED FONT BECOMES THE CURRENT FONT IN THE GC */
/* SETS GC FUNCTION TO GXcopy */
long tml_dp_x_string_draw(bitmap_handle, string, st_count, wherex, wherey, use_font)
    struct bm_info *bitmap_handle;
    char           *string;
    long            st_count,
                    wherex,
                    wherey;
    struct fhandle *use_font;
{
    register struct window_info *bitmap_info;
    unsigned long  valuemask_change;   
    XGCValues      cur_values;   
    XGCValues      new_values;   
    XChar2b       *string16;
    long           i;
    long           icspace;
    long           rop_id;
    XFontStruct   *font_info;
    XTextItem16   *text_items16;
    XTextItem16   *text_item16;
    XTextItem     *text_items8;
    XTextItem     *text_item8;
    XCharStruct   *overall;
  
    /*first, make sure that the font is ok, and that there is a string to draw*/
    if (!use_font || !st_count || !string)
        return -1;

    /*if the font isn't loaded, load it*/
    if (!use_font->font_info)
        use_font = tml_dp_x_get_fhandle(use_font->font_name, use_font->name_len);

    font_info = use_font->font_info;

    /* retrieve current values from GC */

    bitmap_info = bitmap_handle->bitmap_id;
    cur_values = bitmap_info->gc->values;
    valuemask_change = 0;

    /* change to font indicated */

    if (font_info->fid != cur_values.font) {
        valuemask_change |= GCFont;
        new_values.font = font_info->fid;
        }
    if (valuemask_change)
        XChangeGC(bitmap_info->display, bitmap_info->gc, valuemask_change, &new_values);

    /* Set GC function to GXcopy */
    /* The text drawing functions here, XDrawString and XDrawString16, */
    /*   alter foreground pixels only. These calls therefore perform */
    /*   text draw on transparent background without need for any special */
    /*   handling of foreground/background colors or planemask, and allow */
    /*   use of GXcopy. */
    if (GXcopy != (rop_id = bitmap_info->gc->values.function))
        XSetFunction(bitmap_info->display, bitmap_info->gc, GXcopy);

    /* no intercharacter space specified */

    if (0 == use_font->icspace) {

        /* two-byte-character fonts */

        if (use_font->is16) {
            st_count >>= 1;
            XDrawString16(bitmap_info->display, bitmap_info->window, bitmap_info->gc,
                          wherex, wherey, (XChar2b*) string, st_count);
            XTextExtents16(font_info, (XChar2b*) string, st_count,
                           (int*) &(bitmap_handle->last_direction), (int*) &(bitmap_handle->last_ascent),
                           (int*) &(bitmap_handle->last_descent), &(bitmap_handle->last_extent));
            }

        /* one-byte-character fonts */

        else {
            XDrawString(bitmap_info->display, bitmap_info->window, bitmap_info->gc, wherex, wherey,
                        string, st_count);
            XTextExtents(font_info, string, st_count,
                         (int*) &(bitmap_handle->last_direction), (int*) &(bitmap_handle->last_ascent),
                         (int*) &(bitmap_handle->last_descent), &(bitmap_handle->last_extent));
            }
        }

    /* intercharacter space specified */

    else {
        icspace = use_font->icspace;

        /* two-byte-character fonts */

        if (use_font->is16) {
            st_count >>= 1;
            text_items16 = (XTextItem16*) tml_c_tml_os_alloc(st_count, (long) sizeof(XTextItem16));
            string16 = (XChar2b*) string;
            for (i = 0; i < st_count; i++) {
                text_item16 = text_items16 + i;
                text_item16->chars = string16 + i;
                text_item16->nchars = 1;
                if (i)
                    text_item16->delta = icspace;
                text_item16->font = None;
                XDrawText16(bitmap_info->display, bitmap_info->window, bitmap_info->gc,
                    wherex, wherey, text_items16, st_count);
                }
            XTextExtents16(font_info, (XChar2b*) string, st_count,
                           (int*) &(bitmap_handle->last_direction), (int*) &(bitmap_handle->last_ascent),
                           (int*) &(bitmap_handle->last_descent), &(bitmap_handle->last_extent));
            }

        /* one-byte-character fonts */

        else {
            text_items8 = (XTextItem*) tml_c_tml_os_alloc(st_count, (long) sizeof(XTextItem));
            for (i = 0; i < st_count; i++) {
                text_item8 = text_items8 + i;
                text_item8->chars = string + i;
                text_item8->nchars = 1;
                if (i)
                    text_item8->delta = icspace;
                text_item8->font = None;
                XDrawText(bitmap_info->display, bitmap_info->window, bitmap_info->gc,
                    wherex, wherey, text_items8, st_count);
                }
            XTextExtents(font_info, string, st_count,
                         (int*) &(bitmap_handle->last_direction), (int*) &(bitmap_handle->last_ascent),
                         (int*) &(bitmap_handle->last_descent), &(bitmap_handle->last_extent));
            }

        /* add total intercharacter spacing to get true width of string drawn */

        bitmap_handle->last_extent.width += (use_font->icspace * st_count);

        }

    /* restore rop */

    if (GXcopy != rop_id)
        XSetFunction(bitmap_info->display, bitmap_info->gc, rop_id);

    /* report absolute value of width drawn */

    if (0 > bitmap_handle->last_extent.width)
        return -(bitmap_handle->last_extent.width);
    else
        return bitmap_handle->last_extent.width;
}

/*return the size of a string in the indicated font*/
long tml_dp_x_string_size(string, st_count, use_font)
    char           *string;
    long            st_count;
    struct fhandle *use_font;
{
    long         direction;
    long         ascent;
    long         descent;
    long         icspace;
    XChar2b     *char16p;
    XFontStruct *font_info;
    XCharStruct *per_char;
    XCharStruct  overall;

    font_info = use_font->font_info;

    if (use_font->is16)
        st_count >>= 1;

    /*take care of the easy case if we can*/

    if (use_font->is_fixed)
        return((use_font->is_fixed + font_info->max_bounds.width) * st_count);

    /*it's proportional--do it the hard way*/

    icspace = use_font->icspace;
    if (use_font->is16) {
        XTextExtents16(font_info, (XChar2b*) string, st_count,
                       (int*) &direction, (int*) &ascent,
                       (int*) &descent, &overall);
        }
    else {
        XTextExtents(font_info, string, st_count,
                       (int*) &direction, (int*) &ascent,
                       (int*) &descent, &overall);
        }

    return(overall.width += (icspace * st_count));
}

#ifdef not_vanilla_C
/* application calls this when it receives an Expose or GraphicsExpose event */
void tml_dp_add_clip_rect(bitmap_id, ulx, uly, width, height)
    void* bitmap_id;
#else
tml_dp_add_clip_rect(bitmap_id, ulx, uly, width, height)
    char* bitmap_id;
#endif
    long ulx;
    long uly;
    long width;
    long height;
{
    struct bm_info* bitmap_handle = (struct bm_info*) bitmap_id;

    /* still have rectangle space */
    if ((bitmap_handle->exp_nrect != TML_WHOLE_WINDOW)
      && (bitmap_handle->exp_nrect < TML_MAX_X_RECT)) {

        /* add the new rectangle */
        bitmap_handle->exp_rect[bitmap_handle->exp_nrect].x      = ulx;
        bitmap_handle->exp_rect[bitmap_handle->exp_nrect].y      = uly;
        bitmap_handle->exp_rect[bitmap_handle->exp_nrect].width  = width;
        bitmap_handle->exp_rect[bitmap_handle->exp_nrect].height = height;

        /* revise bounding box for redraw */
        if (ulx < bitmap_handle->exp_ulx)
            bitmap_handle->exp_ulx = ulx;
        if (uly < bitmap_handle->exp_uly)
            bitmap_handle->exp_uly = uly;
        if (bitmap_handle->exp_lrx < (ulx + width))
            bitmap_handle->exp_lrx = (ulx + width);
        if (bitmap_handle->exp_lry < (uly + height))
            bitmap_handle->exp_lry = uly + height;

        /* count the new rectangle */
        ++bitmap_handle->exp_nrect;
        }

    /* ran out of rectangle space */
    else {
        bitmap_handle->exp_nrect = TML_WHOLE_WINDOW;
        bitmap_handle->exp_ulx = 0;
        bitmap_handle->exp_uly = 0;
        bitmap_handle->exp_lrx = bitmap_handle->size_x;
        bitmap_handle->exp_lry = bitmap_handle->size_y;
        }
}


/*repair any damage in the specified viewport*/
/* SETS THE GC function to GXcopy */
tml_dp_x_damage_repair(bitmap_handle)
    struct bm_info *bitmap_handle;
{
    Status              status;
    struct window_info *bitmap_info;
    XEvent              event;
    XGCValues           values;
    long                old_sizex, old_sizey;
    tml_c_view_port     vp;

    bitmap_info = bitmap_handle->bitmap_id;
    vp = bitmap_handle->viewport_id;

    /* Set GC function to GXcopy */

    if (GXcopy != bitmap_info->gc->values.function)
        XSetFunction(bitmap_info->display, bitmap_info->gc, GXcopy);

    /* viewport size changed; reformat and  redraw */

    if ((bitmap_handle->new_x != bitmap_handle->size_x) || (bitmap_handle->new_y != bitmap_handle->size_y)) {
        bitmap_handle->size_x = bitmap_handle->new_x;
        bitmap_handle->size_y = bitmap_handle->new_y;
        tml_dp_x_clear_clip(bitmap_handle);
        view_port_redraw(vp);
        }
 
    /* refresh only, through appropriate clip */

    else if (bitmap_handle->exp_nrect) {

        if (bitmap_handle->exp_nrect != TML_WHOLE_WINDOW)
            XSetClipRectangles(bitmap_info->display, bitmap_info->gc, 0, 0, bitmap_handle->exp_rect, bitmap_handle->exp_nrect, Unsorted);
        else
            XSetClipMask(bitmap_info->display, bitmap_info->gc, None);
        /* clear number of rectangles here to prevent infinite loop */
        /*   when tml_c_repaint_screen results in calls to this function */
        bitmap_handle->exp_nrect = 0;
        tml_c_repaint_screen (vp, bitmap_handle->exp_ulx, bitmap_handle->exp_uly, bitmap_handle->exp_lrx, bitmap_handle->exp_lry);
        tml_dp_x_clear_clip(bitmap_handle);
        }
}

#ifdef not_vanilla_C
void* tml_dp_x_get_grafsys_obj(bitmap_handle)
    void* bitmap_handle;
{
    return (void*) ((struct bm_info*) bitmap_handle)->bitmap_id;
}

#else
char* tml_dp_x_get_grafsys_obj(bitmap_handle)
    char* bitmap_handle;
{
    return (char*) ((struct bm_info*) bitmap_handle)->bitmap_id;
}
#endif

/* default cursor-drawing routine - invert a triangle of specified height and width */
/*   with the back edge (lr) on the indicated (x,y) coordinate in the viewport. */
/*   wherex is the x coordinate of the left base vertex of the triangle */
/*   wherey is the y coordinate of the top vertex of the triangle */
/*   SETS THE GC FUNCTION TO GXinvert */
long tml_dp_x_dflt_cursor_draw(draw_to_bitmap, wherex, wherey, height, width, erase)
    struct bm_info *draw_to_bitmap;
    long            wherex, wherey,
                    height, width,
                    erase;
{
    register struct window_info *bm_id;
                    XPoint       points[3];
                    long         foreground, flipground;

    bm_id = draw_to_bitmap->bitmap_id;
    /* Set the GC function to GXxor */
    if (GXxor != bm_id->gc->values.function)
        XSetFunction(bm_id->display, bm_id->gc, GXxor);

    /* Make the cursor polygon */

    points[0].x = wherex - width/2; points[0].y = wherey + height;
    points[1].x = wherex;           points[1].y = wherey;
    points[2].x = wherex + width/2; points[2].y = wherey + height;

    /* Set foreground to XOR of foreground and background */
    /*   to switch those colors within cursor triangle */

    foreground = bm_id->gc->values.foreground;
    flipground = foreground ^ bm_id->gc->values.background;
    XSetForeground(bm_id->display, bm_id->gc, flipground);

    /* - makes restore of exposed cursor automatic */

    XFillPolygon(bm_id->display, bm_id->window, bm_id->gc,
                 points, 3, Convex, CoordModeOrigin);

    /* Restore GC foreground */

    XSetForeground(bm_id->display, bm_id->gc, foreground);

    if (erase)
        return 0;
    else
        return 1;
}

