/**************************************************************************************
    tml_objects.hxx
    header file for the object definitions and assorted type declarations for
      the portable version of the text management library.

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#ifndef tml_objects_hxx
#define tml_objects_hxx

#ifndef char_set_mgr_hxx
#include <tml/char_set_mgr.hxx>
#endif

/**********************************************************************************
    version number of this version of tml
**********************************************************************************/

// 1: base functionality, history, extensibility, extensible property
// 2: add extended argument, variable size heaps, compressed save, magic numbers
// 3: add character set manager
// 4: multiple graphics systems (extensible) per TMLib
// 5: add base support for heterogeneous data types (e.g. graphics)

const long tml_version_number = 5;

/**********************************************************************************
    classes which will be declared in this header (for forward reference)
**********************************************************************************/

class node;
class bnode;
class gnode;
class model;
class model_pos;
class prop_vec;
class property;
class argument;
class int_arg;
class float_arg;
class string_arg;
class subtree_arg;
class char_set_arg;
class union_char_set_arg;
class heap;
class obj_heap;
class squeeze_heap;
class bit_heap;
class write_bit_heap;
class read_bit_heap;
class heap_list;
class oheap_list;
class sqheap_list;
class bit_heap_list;
class prop_func_ent;
class prop_state;
class atom_arg;
class atom_this_arg;
class atom_ptr_arg;
class atom_val_arg;
class atom;
class molecule;
class history;
class format;
class cline;
class phead;
class font_tbl;
class view_pos;
class v_cursor;
class view_port;
class d_model_parts;
class d_model_info;
class d_plid;
class d_node;
class d_prop_vec;
class d_property;
class d_argument;
class tml_ext_type_v1;
class node_v1;
class bnode_v1;
class gnode_v1;
class property_v1;
class argument_v1;
class d_model_info_v1;
class d_model_info_v4;
class vpftbl;

/**********************************************************************************
    constants used in the classes
**********************************************************************************/
/*boolean references for property values*/
const long  tml_true_prop   =   1,
            tml_false_prop  =   0;
                             
/**********************************************************************************
    constants used in the heap_list constructors
**********************************************************************************/

const long obj_heap_incr     =   16384;
const long goo_heap_incr     =   16384;
const long st_heap_incr      =   2048;
const long tml_ao_heap_incr = obj_heap_incr;

/**********************************************************************************
    Enumerated types
**********************************************************************************/

/*enumerated class of directions for tree navigation*/
enum    node_dir  {n_up, n_right, n_left, n_down};

/*enumerated class of TML Version 1 argument types*/
enum    arg_type_v1  {int_val, float_val, string_val, no_val};

/*enumerated class of wrap property values*/
enum    wrap_type {no_wrap, char_wrap, word_wrap};

/*enumerated class of margin locations*/
enum    marg_loc {top_marg, bottom_marg, left_marg, right_marg};

/*enumerated class of types of conceptual lines*/
enum    cl_type {cl_free, cl_text, cl_graphic, cl_blank, cl_gutter, cl_other};

/*enumerated class of the directions the cursor can move*/
enum vp_dir {vp_right, vp_left, vp_up, vp_down,
             vp_begin_line, vp_end_line, vp_top, vp_bottom};
                                                      
/*enumerated class of file types*/
enum file_type {failed, passive, flat_file, tml_file};

/**********************************************************************************
    Declarations establishing independence from pointer/integer conversion
      peculiarities of heterogeneous compiler architectures
**********************************************************************************/

#ifdef ptr_32_int_16
#else
#ifdef ptr_16_int_16
#else
#ifdef ptr_16_int_32
#else                    // ptr_32_int_32
const MAX_HEAP_INCR = 0x0000FFFF;
const MAX_FLOAT_BUF_V2 = 40;
#endif
#endif
#endif

/**********************************************************************************
    preloaded node types
**********************************************************************************/
extern tml_type* const undefined_node;
extern tml_type* const bobj;
extern tml_type* const branch;
extern tml_type* const gobj;
extern tml_type* const pgobj;
extern tml_type* const leaf;
extern tml_type* const is_free;

typedef tml_type *node_type;


/**********************************************************************************
    preloaded property types
**********************************************************************************/
extern tml_type* const undefined_prop;
extern tml_type* const tml_wrap_prop;
extern tml_type* const tml_margin_prop;
extern tml_type* const tml_font_prop;
extern tml_type* const tml_inv_prop;
extern tml_type* const tml_pgsize_prop;
extern tml_type* const tml_header_prop;
extern tml_type* const tml_footer_prop;
extern tml_type* const tml_form_feed_prop;
extern tml_type* const tml_char_set_prop;
extern tml_type* const tml_union_char_set_prop;
extern tml_type* const tml_tab_prop;

typedef tml_type *p_index;            /* index member of property class */


/**********************************************************************************
    preloaded argument types
**********************************************************************************/
extern tml_type* const int_val_arg;
extern tml_type* const float_val_arg;
extern tml_type* const string_val_arg;
extern tml_type* const no_val_arg;
extern tml_type* const subtree_val_arg;
extern tml_type* const char_set_val_arg;
extern tml_type* const union_char_set_val_arg;

typedef tml_type *arg_type;

/**********************************************************************************
    preloaded graphics system types
**********************************************************************************/
extern tml_type* const tml_xwindows_grafsys;
extern tml_type* const tml_gpr_grafsys;

typedef tml_type *grafsys;

/**********************************************************************************
    structures used for general purposes
**********************************************************************************/
/*structure used to pass around x,y coordinates in any space*/
struct xy_coord {
    long    x_coord,
            y_coord;
    };

/*structure containing the basic x,y coordinates (in pixels) of a location
    in a bitmap--should only be used for bitmap space*/
struct screen_xy{
    long    at_x,
            at_y;
};

/**********************************************************************************
    Objects used in the modelling component of tml
**********************************************************************************/

/* macro declarations of heap types for model save and restore -- graphics */
/*   node data heaps are not enumerated as they are saved and restored */
/*   differently from other heaps */
#define OBJ_HEAP                    0

#define FIRST_WRITTEN_HEAP          1

#define GOO_HEAP                    1

#define ST_HEAP                     2

#define NUM_MODEL_HEAP_TYPES        3

#define SQUEEZE_HEAP                3

#define LAST_EPHEMERAL_HEAP_TYPE    3

#define BIT_HEAP                    4

#define NUM_HEAP_TYPES              5

// model information class for disk save and restore - part of file header

class d_model_info {                           // information from class d_model_parts that
public:                                        //   is actually written to disk during
    long           version_number;             //   during disk save.
    d_ptr          root;
    long           num_types;                  // number of universal type identifiers stored on disk file
    long           num_plids;                  // number of used plids in the plid_collection to restore from disk
    long           num_heaps[NUM_HEAP_TYPES];
    long           heap_quanta[NUM_HEAP_TYPES];
    HALF_PTR       cur_heap[NUM_HEAP_TYPES];

    void unscramble(endian* their_endian);
};

/* structures and objects for tracking heaps during model write */
struct heap_key {
    char* start_addr;
    long heap_size;
    long bytes_used;
    };

class d_heap_info {
    heap*      cur_heap[NUM_HEAP_TYPES];
    heap_list* first_list[NUM_HEAP_TYPES];
    heap_key*  sort_array[NUM_HEAP_TYPES];

    friend class d_model_parts;
public:
    d_heap_info(model* write_model);
    void complete_for_heap_type(long heap_type, long num_heaps);
    void prepare_for_destruction(d_model_parts* model_parts);
    };

/* model control structure used during save and restore of model to/from disk */
/*   a portion of this structure serves as a header for a saved tml object */
class d_model_parts {                          // structure used to gather information about
public:                                        //   the model for disk save

/********************************************************/
/* BEGINNING OF PORTION IDENTICAL TO CLASS D_MODEL_INFO */

    long           version_number;             // version number of the tml that wrote a given model to disk
    d_ptr          root;                       // root node of class model in d_ptr form
    long           num_types;                  // number of universal type identifiers stored on disk file
    long           num_plids;                  // number of used plids in the plid_collection to restore from disk
    long           num_heaps[NUM_HEAP_TYPES];
    long           heap_quanta[NUM_HEAP_TYPES];
    HALF_PTR       cur_heap[NUM_HEAP_TYPES];

/* END OF PORTION IDENTICAL TO CLASS D_MODEL_INFO      */
/*******************************************************/

    model*         for_model;
    long           for_write;        // 1 when used for writing, 0 for reading
    tml_ext_type*  types;            // array of universal type identifiers
    tml_type*      tml_types;        // array of objects for restoring virtual function pointers
    d_plid*        d_plid_col;       // array of disk-format plids
    d_heap_info*   heap_info;
    long*          heap_sizes[NUM_HEAP_TYPES];
    long*          bytes_used[NUM_HEAP_TYPES];
    CHARP*         heaps[NUM_HEAP_TYPES];
    long           their_magic;
    endian*        their_endian;
    squeeze_heap*  cur_squeeze_heap;
    sqheap_list*   squeeze_data;
    bit_heap_list* bit_data;

    d_model_parts(model* write_model);
    d_model_parts();
    ~d_model_parts();
    void          from_model_info_v1(d_model_info_v1* model_info_v1);
    void          from_model_info_v4(d_model_info_v4* model_info_v4);
    void          set_model(model* new_model){for_model = new_model;}
    void          load_types();
    long          read_types(void* file_handle);
    long          read_types_v1(void* file_handle);
    long          count_heaps(heap_list* cur_list);
    void          save_pointers(cs_heap*& csm_heaps, gc_hash_tbl*& hash_tbl);
    void          complete_for_save();
    long          save(void* file_handle, cs_heap*& csm_heaps, gc_hash_tbl*& hash_tbl);
    long          prepare_for_restore(void* file_handle, long max_heap_types);
    long          restore_heap_array(void* file_handle, long heap_type);
    void          restore_pointers(gc_hash_tbl& hash_tbl);
    void          restore_pointers_v1(gc_hash_tbl& hash_tbl);
    squeeze_heap* get_current_squeeze_heap(){return(cur_squeeze_heap);}
    char*         obj_alloc(long count);
    char*         obj_realloc(long count);
    char*         obj_noalloc();
    void          q_object_data(squeeze_heap* data);
    void          q_bit_data(void* data, long count);
    void          q_bit_data(read_bit_heap* new_heap);
};

/* persistent location identifiers are once removed references to a location
in a model which are updated whenever a model changes. They are part of the
model structure and may be refererred to only by that model. See the model
constructor. */

const PLID_INIT_SIZE = 0x4;             // use 4 for debug

/* define the object, associated with the model, that will define the location to be saved */
class plid{
    node       *with_node;              // associated with what node
    long        offset;                 // at the offset
    long        ref_count;              // how many references to this plid exist

    friend      class  plid_collection;
};

/* mechanism to trace how many plid objects point back to a node. a node class will
have a pointer to this object */
class plid_trace{
    model      *in_model;
    plid_trace *next;                   // address of first member in trace list for node
    long        plid_index;             // points back to refering plid

    friend class obj_heap;
};

/* the table (collection) of plids, associated with a model,
defined by the user, to create hypertext links */
class plid_collection{
protected:
    model  *in_model;
    long    num_entries;                // number of entries in plid table
    long    top_free;                   // index to next available entry in tabl\e
    plid   *first_tbl_entry;            // address of where the table lives
};

class model_pos{
public:
    node     *in_node;
    long      char_index;
    
    model_pos();
    void      normalize(); /* convert so char_index falls within scope of in_node */
    long      operator== (model_pos comp_pos);
    long      operator!= (model_pos comp_pos);
    long      operator<  (model_pos comp_pos);
    long      operator<= (model_pos comp_pos);
    long      operator>  (model_pos comp_pos);
    long      operator>= (model_pos comp_pos);
    model_pos operator+  (long inc);
    model_pos operator-  (long dec);
    long      operator-  (model_pos dec);
    model_pos operator+= (long inc) {*this = *this + inc; return *this;}
    model_pos operator-= (long dec) {*this = *this - dec; return *this;}
    model_pos operator++ ();
    model_pos operator-- ();
    model_pos get_next();
    model_pos get_prior();
    void      mark_changed();
    long      can_format();
    void      in_place_redraw();
};
              
class model_range{
    model_pos   start,
                end;

public:
    model_range();
    
    void        set_start(model_pos new_start){start = new_start;}
    void        set_end(model_pos new_end){ end = new_end;}
    model_pos   get_start(){return start;}
    model_pos   get_end(){return end;}
    node*       isolate();
    long        contains(model_pos target_pos);
};

/*class for nodes, the basic objects in a text structure*/
class node {
    model           *in_model;
protected:
    node            *up,
                    *right,
                    *left;
    prop_vec        *properties;            //first element in the property vector
    long             sib_order;             //ordering among siblings
    node_type        n_type;
    plid_trace      *node_plid_trace;       //this field indicates what plids
                                            //  refer to this node

    friend  void    d_model_parts::restore_pointers(gc_hash_tbl& hash_tbl);
    friend  void    node_set_vptr(node*);   // for restore from disk
    friend  class   subtree_arg;
    friend  class   plid_trace;
    friend  class   obj_heap;
    friend  class   model;
public:
    /*functions dealing with the general node data*/
    node(long node_size, model* to_model = 0);
    ~node();
      
    /*functions allowing inquiry and setting of the data contained in the class*/
    model*      get_model(){return (in_model);}
    void        set_model(model *new_model){in_model = new_model;}
    node*       get_up(){return up;}
    void        set_up(node* new_up){up = new_up;}
    node*       get_left(){return left;}
    void        set_left(node* new_left){left = new_left;}
    node*       get_right(){return right;}
    void        set_right(node* new_right){right = new_right;}
    node_type   get_type(){return n_type;}
    void        set_type(node_type this_type){n_type = this_type;} // must not be virtual or save/restore will fail
    long        get_level();                              //find the level of node 'this' in the tree by counting
    long        get_sib_order(){return sib_order;}
    void        update_sib_vec(node* new_up, node* new_left); //update up and sib_order of this and right siblings
    prop_vec*   get_props(){return properties;}
    model_range get_span();
    long        is_ancestor(node *target);

    /*functions for placing and moving nodes in a tree*/
    node*       connect_node (node* to_connect, node_dir dir);   //connect a node to a tree - if to_connect and node models differ, copy to_connect
    void        detach();                              //take a node out of a tree
    void        abdicate(node* heir);                  //replace a node with a vector of its left siblings, heir, right siblings
    node*       move (node *move_to, node_dir dir);    //move a node

    /*functions dealing with properties of nodes*/
    void        prop_attach(prop_vec *add_pvec);
    void        prop_detach();
    void        exit_props();    //deactivates all of the properties of the node;
    void        enter_props();  //activates all of the properties of the node;

    /* functions for dealing with plids */
    void        set_node_plid_trace(plid_trace *first_trace);
    long        search_trace(long look_for, plid_trace **found, plid_trace **near_trace);
    plid_trace* get_node_plid_trace(){return node_plid_trace;}
    long        count_plid_traces();     //counts number of plid traces attached to a node, used for save/restore
    void        detach_traces(); //deletes all traces, called from ~node

    /*functions for saving and restoring from disk, trash compaction*/
    virtual void    save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_node* savee = 0);
    virtual void    restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_node* restoree = 0);
    virtual void    restore_v1(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, node_v1* node_ver_1); // restores nodes saved in tml version 1 format
    virtual long    is_terminal(){return 0;}
    virtual long    has_goo(){return 0;}
    virtual long    combine_data(node* with_node);
                    // do nothing except in subclass with combinable data

/*virtual functions--actual implementation may differ for various sub-classes*/
    virtual long    get_osize(){return (long) sizeof(node);}
    virtual long    can_format(){return 0;}
    virtual long    get_max_index(){return 0;}
    virtual void*   get_contents(){return 0;}
    virtual long    is_pgnum(){return 0;} // identifies pgnode hierarchy

    /*the following functions are illdefined for all but the derived classes*/
    virtual node*   copy(model *to_model = 0);
    virtual node*   copy_all(model *to_model = 0){return(copy(to_model));}
    virtual node*   clone(model*new_model, gc_hash_tbl& hash_tbl, node* return_node = 0);
    virtual node*   dup();
    virtual node*   dup_all(){return(dup());}
    virtual node*   split(long split_point);
                
    /*navigation functions*/
    virtual node*   get_next(node_dir& dir);                //get next node in traversal order
    virtual node*   get_prev(node_dir& dir);                //get previous node in traversal order
    virtual node*   get_node(node_dir dir);                 //get the node in the specified direction from 'this'
    virtual void    set_node(node *to_node, node_dir dir);
    virtual node*   get_last(){return this;}           //meaningful only for branch and bnodes

    /*input functions*/
    virtual long        attach_data(void *data_start, long data_count);
    virtual model_pos   insert(long after_index, char *add_data, long add_count = 1, char_set* in_char_set = 0);
    virtual model_pos   del_cont(long delete_index, long delete_num = 1);
    virtual model_pos   overstrike(long begin_index, char *rep_with, long over_num = 1);

    /*formatting functions*/
    virtual long    get_format_hi(format *in_format);
    virtual cline*  format_line(format *in_format, long start_index, long space_hi, long space_wide, long first_pixel = 0);
    virtual cline*  tamrof_line(format *in_format, long start_index, long space_hi, long space_wide, long first_pixel = 0);
};

/*bnodes--derived class for non-terminal nodes in the structure tree*/
class bnode:public node {
    node        *down_ptr;
                                                      
public:
    /*constructor for the class bnode*/
    bnode(model *to_model = 0, long size = 0);  //size argument for extensions

    /*return the size (in bytes) of this object--used when the object is freed*/
    long        get_osize(){return((long) sizeof(bnode));}
                                              
    /*inquiry functions concerning contents and formattability*/
    long        can_format(){return 0;}
    long        get_max_index(){return 0;}

    /*functions for saving and restoring from disk, trash compaction*/
    void        save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_node* savee = 0);
    void        restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_node* restoree = 0);
    void        restore_v1(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, node_v1* node_ver_1); // restores bnodes saved in tml version 1 format

    /*navigation function--all can allow descending into a sub-tree*/
    node*       get_next(node_dir& dir);
    node*       get_prev(node_dir& dir);
    node*       get_node(node_dir  dir);
    void        set_node(node *to_node, node_dir dir);
    node*       get_last();

    /*copy functions for this sort of node*/
    node*       copy(model* to_model = 0);  //make a copy of this, perhaps to a different model*/
    node*       copy_all(model* to_model = 0);//copy the tree with root node 'this', perhaps to a different model
    node*       clone(model* new_model, gc_hash_tbl& hash_tbl, node* return_node = 0);
    node*       dup();                               //duplicate a node --must be to same model
    node*       dup_all();                           //duplicate a tree with 'this' as its root in the specified mode
    node*       split(long split_point);
};

/* file node: derived class allowing incorporation of external models into a model */
class fnode : public bnode {                                                               
    unsigned char* filename;    // terminated string
    file_type      f_type;
};

/*goo nodes--derived class for the terminal object which points to characters in the goo heap*/
class gnode:public node{
    char        *goo_ptr;
    long         count;

    friend void gc_linearize(model*);
public:
    /*constructors for the class gnode*/
    gnode(model *to_model = 0, long size = 0); // size argument for extensions

    /* overload of = operator, allowing direct assignment of strings to nodes */
    gnode&      operator= (char *s);

    /*return the size (in bytes) of this object--used when the object is freed*/
    long        get_osize(){return((long) sizeof(gnode));}

    /*functions returning information on the contents of a gnode*/
    long        can_format(){return 1;}
    long        get_max_index(){return(count - 1);}
    void*       get_contents(){return goo_ptr;}  //get the non-null terminated string dominated by the node
    long        get_char_count(){return count;}
    char*       get_0t_contents();                         //get the null-terminated string dominated by the node
    node*       split(long new_count);                   //split a leaf node into two nodes
    node*       get_last();                             //gets the last sibling in the tree
                                             
    /*copy and duplication functions*/
    node*       copy(model* to_model = 0); //copy gnode 'this' to a different model
    node*       copy_all(model* to_model = 0){ return(copy(to_model));}
    node*       clone(model* new_model, gc_hash_tbl& hash_tbl, node* return_node = 0);
    node*       dup();                   //duplicate a gnode
    node*       dup_all(){return(dup());}

    /*functions used for saving and restoring from disk, trash compaction*/
    void        save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_node* savee = 0);
    void        restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_node* restoree = 0);
    void        restore_v1(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, node_v1* node_ver_1); //restores gnodes saved in tml version 1 format
    void        set_goo_d_ptr(d_ptr *d_ptr_tmp);
    void        set_char_count(long new_count){count = new_count;}
    void        set_goo_ptr(char* nnt){goo_ptr = nnt;}  // CAREFUL! should be a non-null-terminated string
                                                        //   in an existing gheap. Used in garbage collection by gc_copy_tree()
    long        is_terminal(){return(-1);}
    long        has_goo(){return(-1);}
    long        combine_data(node* with_node);

    /*input functions*/
    long        attach_data(void *data_start, long data_count);
    model_pos   insert(long after_index, char *insert, long insert_num = 1, char_set* in_char_set = 0);
    model_pos   del_cont(long delete_index, long delete_num = 1);
    model_pos   overstrike(long begin_index, char *rep_with, long over_num = 1);

    /*formatting functions*/
    long        get_format_hi(format *in_format);
    cline*      format_line(format* in_format, long start_index, long space_hi, long space_wide, long first_pixel = 0);
    cline*      tamrof_line(format* in_format, long start_index, long space_hi, long space_wide, long first_pixel = 0);
};

/*page number nodes--derived class for the terminal object which will contain the current page number*/
class pgnode:public gnode{
public:
    /*constructors for the class pgnode*/
    pgnode(model *to_model = 0, long size = 0); // size argument for extensions

    /*return the size (in bytes) of this object--used when the object is freed*/
    long  get_osize(){return((long) sizeof(pgnode));}

    node* clone(model* new_model, gc_hash_tbl& hash_tbl, node* return_node = 0);
    node*       copy(model* to_model = 0);
    node*       copy_all(model* to_model = 0){ return(copy(to_model));}

    /* identify members of this hierarchy, allowing for extensions */
    long  is_pgnum(){return 1;}

    /*formatting functions*/
    cline*      format_line(format* in_format, long start_index, long space_hi, long space_wide, long first_pixel = 0);
    cline*      tamrof_line(format* in_format, long start_index, long space_hi, long space_wide, long first_pixel = 0);
};

typedef struct
{
   xy_coord   xy_size;
   long       num_sects, sect_size, pix_size, *cmap;
} bmap_size_data;


class prop_vec {
    prop_vec        *next;
    property        *prop;
    model           *in_model;
    long             ref_count;

    friend class model;
    friend class obj_heap;
    friend void node::prop_attach(prop_vec *add_pvec);
    friend void node::prop_detach();

public:
    prop_vec(model *to_model = 0);
    ~prop_vec();
    prop_vec*   get_next(){return(next);}
    prop_vec*   detach(prop_vec* from_vec);
    prop_vec*   prepend(prop_vec* add_element);
    void        append(prop_vec* add_vec);
    prop_vec*   copy(model* to_model = 0);
    prop_vec*   copy_all(model* to_model = 0);
    prop_vec*   dup();
    prop_vec*   dup_all();
    void        attach(property *add_prop);
    void        detach();
    void        activate();
    void        deactivate();
    long        get_ref_count() { return(ref_count);}
    void        init_ref_count() { ref_count = 1; }
    void        incr_ref_count() { ++ref_count; }
    property*   get_prop() { return(prop);}
    void        save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl);
    void        restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl);
    void        restore_v1(d_model_parts& model_parts, gc_hash_tbl& hash_tbl); // restores prop_vecs saved in tml version 1 format
    void        unscramble_v1(endian* their_endian); // for restore of TML Version 1 files only
    void        set_m_ptr(prop_vec *next_vec){next = next_vec;} // set next field in history of prop_vec::append()
    prop_vec*   clone(model* to_model, gc_hash_tbl& hash_tbl);
    model*      get_model(){return(in_model);}
    void        set_model(model* new_model){in_model = new_model;}
};

const PROP_HASH_SHIFT = 0x04;
const PROP_HASH_MASK = (unsigned) 0x3F;

class property{
protected:
    long                 ref_count;  //how many vectors use this property
    p_index              index;      //index into the property function table
    union{
        model           *in_model;   //what model is this in--used when alloced
        property        *next;       //next free property slot--used when free
        } locator;
    argument            *entry_arg,  //head of the entry argument list
                        *exit_arg;   //head of the exit argument list
    void                *act_value;  //ptr to the activation value
                     
    friend class obj_heap;
    friend class model;
    friend void prop_vec::attach(property *add_prop);
    friend void prop_vec::detach();
    friend void prop_set_vptr(property* new_prop);

public:
    property(model *to_model = 0, long property_size = 0);
    ~property();
    p_index           get_index(){return(index);}
    void              set_index(p_index i){index = i;}  // must not be virtual or save/restore will fail
    virtual property* copy(model *to_model = 0);
    virtual property* dup();
    argument*         get_entry_av(){return (entry_arg);}
    argument*         get_exit_av(){return(exit_arg);}
    void              attach_entry(argument *entry_vec);
    void              attach_exit(argument *exit_vec);
    void              detach_entry();
    void              detach_exit();
    virtual void      activate();
    virtual void      deactivate();
    long              get_ref_count() { return(ref_count);}
    void              init_ref_count() { ref_count = 1; }
    void              incr_ref_count() { ++ref_count; }
    virtual property* clone(model* to_model, gc_hash_tbl& hash_tbl, property* return_prop = 0);
    virtual void      save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_property* savee = 0);
    virtual void      restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_property* restoree = 0);
    virtual void      restore_v1(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, property_v1* prop_v1); // restores properties saved in tml version 1 format
    property*         get_locator_next(){return locator.next;}
    void              set_locator_next(property* next_prop){locator.next = next_prop;}
    model*            get_model(){return(locator.in_model);}
    void              set_model(model *new_model){locator.in_model = new_model;}
    virtual long      get_osize(){return((long) sizeof(property));}
};

class font_property:public property {
    vpftbl* ftbl;
public:
    vpftbl*      get_ftbl();
    void         set_ftbl(vpftbl* new_ftbl){ftbl = new_ftbl;}
    grafsys      get_graphics_system();
    void         set_graphics_system(grafsys new_grafsys);
    property*    clone(model* to_model, gc_hash_tbl& hash_tbl, property* return_prop = 0);
    property*    copy(model *to_model = 0);
    property*    dup();
    void         save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_property* savee = 0);
    void         restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_property* restoree = 0);
    virtual void restore_v3(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, property* prop_v3);
    long         get_osize(){return((long) sizeof(font_property));}
};

class argument{
    model              *in_model;
    arg_type            a_type;
    long                ref_count;
protected:
    argument           *next;       
    void               *arg_val;
    
    friend class model;
    friend class obj_heap;
    friend void property::attach_entry(argument *entry_vec);
    friend void property::attach_exit(argument *exit_vec);
    friend void property::detach_entry();
    friend void property::detach_exit(); 
    friend void arg_set_vptr(argument* arg);

public:
    argument(long arg_size = 0, model* to_model = 0);
    ~argument();
    model*               get_model(){return(in_model);}
    argument*            get_next(){return(next);}
    void*                get_val(){return(arg_val);}
    arg_type             get_vtype(){return(a_type);}
    long                 get_ival();   // Soon to be obsolete
    float                get_fval();   // Soon to be obsolete
    char*                get_sval();   // Soon to be obsolete
    long                 get_scount(); // Soon to be obsolete
    argument*            detach(argument *from_vec);
    void                 append(argument *add_arg);
    argument*            prepend(argument *add_arg);
    long                 get_ref_count() { return(ref_count);}
    void                 init_ref_count() { ref_count = 1; }
    void                 incr_ref_count() { ++ref_count; }
    virtual long         get_osize () {return (long) sizeof(argument);}
    virtual argument*    copy(model* to_model = 0);
    virtual argument*    copy_all(model* to_model = 0);
    virtual argument*    clone(model* to_model, gc_hash_tbl& hash_tbl, argument* return_arg = 0);
    virtual void         save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* savee = 0);
    virtual void         restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* restoree = 0);
    void                 set_m_ptr(argument *next_arg){next = next_arg;}  // set next field in history of argument::append()
    void                 set_model(model *new_model){in_model = new_model;}
    void                 set_vtype(arg_type new_type) {a_type = new_type;}  // must not be virtual or save/restore will fail
    void                 set_ref_count(long new_ref) {ref_count = new_ref;}
};


class int_arg:public argument
{
public:
    int_arg(model* to_model = 0, long size = 0);
    ~int_arg();
    long        get_osize(){return (long) sizeof(int_arg);}
    void        set_val(long value);
    argument*   copy(model* to_model = 0);
    argument*   clone(model* to_model, gc_hash_tbl& hash_tbl, argument* return_arg = 0);
    void        save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* savee = 0);
    void        restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* restoree = 0);
};


class float_arg:public argument
{
public:
    float_arg(model* to_model = 0, long size = 0);
    ~float_arg();
    long        get_osize(){return (long) sizeof(float_arg);}
    void        set_val(float value);
    argument*   copy(model* to_model = 0);
    argument*   clone(model* to_model, gc_hash_tbl& hash_tbl, argument* return_arg = 0);
    void        save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* savee = 0);
    void        restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* restoree = 0);
};


class string_arg:public argument
{
    long   count;
public:
    string_arg(model* to_model = 0, long size = 0);
    ~string_arg();
    long        get_osize(){return (long) sizeof(string_arg);}
    long        get_scount(){return count;}
    void        set_val(char* value, long val_count);
    argument*   copy(model* to_model = 0);
    argument*   clone(model* to_model, gc_hash_tbl& hash_tbl, argument* return_arg = 0);
    void        save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* savee = 0);
    void        restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* restoree = 0);
};


class subtree_arg:public argument
{
public:
    subtree_arg(model* to_model = 0, long size = 0);
    ~subtree_arg();
    long        get_osize(){return (long) sizeof(subtree_arg);}
    void        set_val (node* new_subtree);
    argument*   copy(model* to_model = 0);
    argument*   clone(model* to_model, gc_hash_tbl& hash_tbl, argument* return_arg = 0);
    void        save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* savee = 0);
    void        restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* restoree = 0);
};


class char_set_arg:public argument
{
public:
    char_set_arg(model* to_model = 0, long size = 0);
    ~char_set_arg();
    long        get_osize(){return (long) sizeof (char_set_arg);}
    void        set_val(char_set* new_char_set);
    argument*   copy(model* to_model = 0);
    argument*   clone(model* to_model, gc_hash_tbl& hash_tbl, argument* return_arg = 0);
    void        save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* savee = 0);
    void        restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* restoree = 0);
};


class union_char_set_arg:public argument
{
public:
    union_char_set_arg(model* to_model = 0, long size = 0);
    ~union_char_set_arg();
    long        get_osize(){return (long) sizeof(union_char_set_arg);}
    void        set_val(union_char_set* new_union_char_set);
    argument*   copy(model* to_model = 0);
    argument*   clone(model* to_model, gc_hash_tbl& hash_tbl, argument* return_arg = 0);
    void        save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* savee = 0);
    void        restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* restoree = 0);
};

class heap{
protected:
    long             next_byte;
    long             size;
    char            *heap_start;
    model           *in_model;

    friend class obj_heap;
    friend class goo_heap;
    friend class st_heap;
public:
    heap::heap();
    ~heap();
    char* get_start(){return(heap_start);}
    long  get_byte(){return(next_byte);}
    long  get_size(){return(size);}
    void  attach(model* to_model){in_model = to_model;}
    void  detach(){in_model = 0;}
    void  set_byte(long num_bytes){next_byte = num_bytes;}
    void  set_start(char *first_byte_p){heap_start = first_byte_p;}
    virtual long restore(void* file_handle, d_model_parts* model_parts, long index);
};


class obj_heap : public heap {
    node            *tf_node;
    prop_vec        *tf_pvec;
    property        *tf_prop;
    argument        *tf_arg;
    plid_trace      *tf_p_trace;
      
    friend class model;
    friend class node;
    friend class prop_vec;
    friend class property;
    friend class argument;
    friend class plid_trace;
    friend class additive_property;
    friend class gnode;
    friend class pgnode;
    friend class bnode;
    friend class int_arg;
    friend class float_arg;
    friend string_arg;
    friend subtree_arg;
    friend char_set_arg;
    friend union_char_set_arg;
public:
    obj_heap(long oheap_size = 0);
    ~obj_heap();
    char*       alloc(long count);
    void        free(node*     free_node);
    void        free(prop_vec* free_pvec);
    void        free(property* free_prop);
    void        free(argument* free_arg);
    node*       get_tf_node(long node_size);
    prop_vec*   get_tf_pvec(){return tf_pvec;}
    property*   get_tf_prop(long property_size, property*& prev_prop);
    argument*   get_tf_arg(long arg_size);
    long        restore(void* file_handle, d_model_parts* model_parts, long index);
};


class squeeze_heap : public heap {
public:
    squeeze_heap(long sqheap_size = 0);
    ~squeeze_heap();
    char* alloc(long count);
    long   restore(void* file_handle, d_model_parts* model_parts, long index);
};


class goo_heap : public heap {
public:
    goo_heap(long gheap_size = 0);
    ~goo_heap();
    char* string_alloc(long count, char* copy_string = 0);
    char* char_alloc(char* copy_char, long char_size = 1);
    long  restore(void* file_handle, d_model_parts* model_parts, long index);
};


class st_heap : public heap {
public:
    st_heap(long sheap_size = 0);
    ~st_heap();
    char* alloc(char* copy_string, long count);
    long  restore(void* file_handle, d_model_parts* model_parts, long index);
};


class bit_heap : public heap {
    csm_lock* lock;
public:
    bit_heap();
    ~bit_heap();
    csm_lock* get_lock(){return lock;}
};


class write_bit_heap : public bit_heap {
public:
    write_bit_heap(void* data, long count);
};


class read_bit_heap : public bit_heap {
public:
    read_bit_heap(long count);
    long restore(void* file_handle, d_model_parts* model_parts, long index);
};


class heap_list {
protected:
    class heap_list *next;
    heap            *this_heap;
public:
    heap_list();
    ~heap_list();
    heap_list* get_next(){return next;}
    void       set_next(heap_list* next_list){next = next_list;}
    heap*      get_heap(){return this_heap;}
    void       set_heap(heap* new_heap){this_heap = new_heap;}
};


class oheap_list : public heap_list{
public:
    oheap_list(){;}
    ~oheap_list();
    oheap_list* get_next(){return((oheap_list*) next);}
    void        set_next(oheap_list* next_list){next = next_list;}
    obj_heap*   get_heap(){return((obj_heap*) this_heap);}
    void        set_heap(obj_heap* oheap){this_heap = oheap;}
};


class gheap_list : public heap_list{
public:
    gheap_list(){;}
    ~gheap_list();
    gheap_list* get_next(){return((gheap_list*) next);}
    void        set_next(gheap_list* next_list){next = next_list;}
    goo_heap*   get_heap(){return((goo_heap*) this_heap);}
    void        set_heap(goo_heap* gheap){this_heap = gheap;}
};
    
                                   
class sheap_list : public heap_list{
public:
    sheap_list(){;}
    ~sheap_list();
    sheap_list* get_next(){return((sheap_list*) next);}
    void        set_next(sheap_list* next_list){next = next_list;}
    st_heap*    get_heap(){return((st_heap*) this_heap);}
    void        set_heap(st_heap* sheap){this_heap = sheap;}
};


class sqheap_list : public heap_list{
public:
    sqheap_list(){;}
    ~sqheap_list();
    sqheap_list*  get_next(){return((sqheap_list*) next);}
    void          set_next(sqheap_list* next_list){next = next_list;}
    squeeze_heap* get_heap(){return((squeeze_heap*) this_heap);}
    void          set_heap(squeeze_heap* sqheap){this_heap = sqheap;}
};


class bit_heap_list : public heap_list {
public:
    bit_heap_list* get_next(){return((bit_heap_list*) next);}
    void           set_next(bit_heap_list* next_list){next = next_list;}
    bit_heap*      get_heap(){return((bit_heap*) this_heap);}
    void           set_heap(bit_heap* heap){this_heap = heap;}
};


class prop_t_ent{
    prop_t_ent  *prior,
                *next;
    p_index      prop_index;
    void        *value;
    short        is_default;
    
    friend class prop_state;

public:
    prop_t_ent();
    ~prop_t_ent();
    void    set_prop_index(p_index new_i){prop_index = new_i;}
    p_index get_prop_index(){return prop_index;}
    void    set_value(void *new_value){value = new_value;}
    void*   get_value(){return value;}
};

const   long prop_tbl_size   =   64;
const   long prop_tbl_mask   =   0x3f;

class prop_state{
    short   lock;
    prop_t_ent  *state_tbl[prop_tbl_size];
    prop_t_ent  *tf_entry;

    prop_t_ent* find_entry(p_index in_dex);
    prop_t_ent* find_default(p_index in_dex);

public:
    prop_state();
    ~prop_state();
    short        inq_lock(){return lock;}
    void         set_lock(){while (lock); lock = 1;}
    void         release_lock(){lock = 0;}
    prop_t_ent  *get_entry();
    void         free_entry(prop_t_ent *to_free);
    long         push_prop(p_index prop_ind, void *prop_value);
    long         pop_prop(p_index prop_ind);
    void*        inq_state(p_index prop_ind);
    void*        get_default(p_index prop_ind);
    long         set_default(p_index prop_ind, void *def_value);
};

typedef void* (*GEN_FUNCP)(...);

enum atom_arg_type { this_arg, ptr_arg, val_arg };

class atom_arg {
    void*     origin;
    atom_arg* next;

    friend class atom;
public:
    atom_arg(history* to_history, atom_arg_type type);
    ~atom_arg();
    void*     get_contents(){return(origin);}
    void      set_contents(void* source){origin = source;}
    atom_arg* get_next(){return(next);}
    void      set_next(atom_arg* arg){next = arg;}
    virtual   void empty(){};
    virtual   void set_top_free(history* from_history){from_history = 0;}
};

class atom_this_arg:public atom_arg {
public:
    atom_this_arg(history* to_history, void* source);
    ~atom_this_arg();
    void set_top_free(history *from_history);
};

class atom_ptr_arg:public atom_arg {
public:
    atom_ptr_arg(history* to_history, void* source);
    ~atom_ptr_arg();
    void empty(){delete get_contents();}
    void set_top_free(history *from_history);
};

class atom_val_arg:public atom_arg {
public:
    atom_val_arg(history* to_history, void* source);
    atom_val_arg(history* to_history, long  source);
    atom_val_arg(history* to_history, float source);
    ~atom_val_arg();
    void set_top_free(history *from_history);
};

class atom {
    GEN_FUNCP      undo_func;
    GEN_FUNCP      redo_func;
    long           undo_args;  //number of arguments for undo function
    long           redo_args;  //number of arguments for redo function
    atom_arg*     *args;
    long           count;      // count of atom_arg array elements
    atom          *next;
    atom          *prev;

    friend class model;
    friend class history;
public:
    atom(history *to_history, WHOLE_PTR undo_funcp, WHOLE_PTR redo_funcp,
         atom_arg* arg1  = 0, atom_arg* arg2  = 0, atom_arg* arg3  = 0, atom_arg* arg4  = 0,
         atom_arg* arg5  = 0, atom_arg* arg6  = 0, atom_arg* arg7  = 0, atom_arg* arg8  = 0,
         atom_arg* arg9  = 0, atom_arg* arg10 = 0, atom_arg* arg11 = 0, atom_arg* arg12 = 0,
         atom_arg* arg13 = 0, atom_arg* arg14 = 0, atom_arg* arg15 = 0, atom_arg* arg16 = 0,
         atom_arg* arg17 = 0, atom_arg* arg18 = 0, atom_arg* arg19 = 0, atom_arg* arg20 = 0,
         atom_arg* arg21 = 0, atom_arg* arg22 = 0, atom_arg* arg23 = 0, atom_arg* arg24 = 0,
         atom_arg* arg25 = 0, atom_arg* arg26 = 0);
    ~atom();
    void   free(history *from_history);
    long   get_count(){return count;}
    atom  *get_next(){return next;}
    atom  *set_next(atom *new_atom){next = new_atom; return new_atom;}
    atom  *get_prev(){return prev;}
    atom  *set_prev(atom *new_atom){prev = new_atom; return new_atom;}
    void   undo();
    void   redo();
};

class atom_obj_heap {
    char          *bytes;
    unsigned long  next_byte;
    unsigned long  heap_size;

    friend class history;
    friend class ao_list;

public:
    atom_obj_heap(unsigned long new_heap_size = tml_ao_heap_incr);
    ~atom_obj_heap();
    char*         get_bytes(){return bytes;}
    unsigned long get_next_byte(){return next_byte;}
    unsigned long get_heap_size(){return heap_size;}
    void          set_next_byte(long next){next_byte = next;}
};

class ao_list {
    atom_obj_heap *this_heap;
    ao_list       *next;

    friend class history;
public:
    ao_list(unsigned long new_heap_size = tml_ao_heap_incr);
    ~ao_list();
    atom_obj_heap* get_heap(){return this_heap;}
    ao_list*       get_next(){return next;}
    void           set_next(ao_list *list){next = list;}
};

class molecule {
    atom     *head_atom;
    atom     *tail_atom;
    history  *in_history;
    molecule *next;             // for chaining on free list

    friend class history;
public:
    molecule::molecule(history* to_history);
    ~molecule();
    void set(atom* head, atom* tail){head_atom = head; tail_atom = tail;}
    void set_next(molecule *top_free){next = top_free;}
    void undo();
    void redo();
    void copy_free(molecule* from_molecule);
    void free();
};

class history {
    ao_list         *ao_store;          //atom object cache (history)
    molecule        *top_free_molecule; //free list of molecules
    atom            *top_free_atom;     //free list of atoms
    atom_this_arg   *top_free_this_arg; //free list of atom_this_arg
    atom_ptr_arg    *top_free_ptr_arg;  //free list of atom_ptr_arg
    atom_val_arg    *top_free_val_arg;  //free list of atom_val_arg
    atom            *head_atom;         //head of current molecule
    atom            *tail_atom;         //tail of current molecule
    molecule*       *hist;              //array of molecular operations
    long             hist_size;         //effective size of above array
    long             hist_array_size;   //actual size of above array
    long             active;            //undo or redo in progress; prevents recursive
                                        //  recording of history
    long             recursing;         //recursive call to a member function
                                        //  that records history is in progress
    long             next_op;           //index for next molecular operation to be added
    long             undo_op;           //index of next operation to be undone
    long             in_effect;         //number of undoable operations
    long             filled;            //number of history slots filled
                                        // (molecules in effect + molecules undone)
    long             uninitialized;     //no molecules yet allocated
public:
    /* functions used for history, undo, redo */
    history();
    ~history();
    void      hist_free();
    long      has_hist(){return((hist) ? -1 : 0);}
    long      is_active(){return(active && !recursing ? -1 : 0);}
    void      deactivate(){recursing = -1;}
    void      reactivate(){recursing = 0;}
    long      set_hist_size(long size);
    long      get_hist_size(){return hist_size;}
    ao_list*  get_ao_store(){return ao_store;}
    void      set_ao_store(ao_list* list){ao_store = list;}
    char*     alloc_obj_bytes(unsigned long nbytes);
    molecule* get_top_free_molecule(){return(top_free_molecule);}
    atom*     get_top_free_atom(){return(top_free_atom);}
    void      set_top_free_molecule(molecule* new_top){top_free_molecule = new_top;}
    void      set_top_free_atom(atom* new_top){top_free_atom = new_top;}
    atom_this_arg* get_top_free_this_arg(){return(top_free_this_arg);}
    atom_ptr_arg*  get_top_free_ptr_arg(){return(top_free_ptr_arg);}
    atom_val_arg*  get_top_free_val_arg(){return(top_free_val_arg);}
    void      set_top_free_this_arg(atom_this_arg *arg){top_free_this_arg = arg;}
    void      set_top_free_ptr_arg(atom_ptr_arg *arg){top_free_ptr_arg = arg;}
    void      set_top_free_val_arg(atom_val_arg *arg){top_free_val_arg = arg;}
    long      add_atom(atom *new_atom);
    void      end_molecule();
    long      add_molecule(atom *head, atom *tail);
    long      undo_molecule();
    long      redo_molecule();
};

class mf_list {
    mf_list         *next,
                    *prior;
    format          *with_format;

public:
    mf_list*        inq_next(){return next;}
    mf_list*        inq_prior(){return prior;}
    format*         inq_format(){return with_format;}

    void            set_next(mf_list *new_next){next = new_next;}
    void            set_prior(mf_list *new_prior){prior = new_prior;}
    void            set_format(format *new_format){with_format = new_format;}
};

class ft_list {
    ft_list         *next;
    font_tbl        *this_font;

public:
    ft_list*        inq_next(){return next;}
    font_tbl*       inq_font(){return this_font;}

    void            set_next(ft_list *new_next){next = new_next;}
    void            set_font(font_tbl *new_font){this_font = new_font;}
};
                                                 
class mark_list{
    mark_list       *next,
                    *prior;
    model_range     *mark_range;
    long             is_visible;

public:
    mark_list();
    ~mark_list();
    mark_list*      get_next(){return next;}
    mark_list*      get_prior(){return prior;}
    model_range*    get_range(){return mark_range;}

    void            set_next(mark_list *next_ml){next = next_ml;}
    void            set_prior(mark_list *prior_ml){prior = prior_ml;}
    model_range*    set_range(model_range new_range);
    long            extend_mark(model_pos new_start, model_pos new_end);
    long            draw();
    long            erase();
};
                                    
typedef long  (*search_func)(model_range over_range, model_range& where_found, void *pattern, long forward);
    /*model_range = range to search over; model_range& = where found;
      void * = pattern returned by compile_func; long = forward of != 0, backward otherwise)*/

typedef long  (*compile_func)(char *, long , void *, long, long);
    /*char *= pointer to beginning of string with search pattern; long =
        length of pattern to compile; void * = returned pattern; long = max_size of
        returned pattern; long = case insensitivity flag (0 = case sensitive, != 0 case insens)
        */

typedef char* (*search_nav_func)(model_pos&);

class search_env{
    search_env      *next;
    search_func      search_rtn;
    compile_func     compile_rtn;
    search_nav_func  next_char_func,   // return current character and push ptr to next one
                     prior_char_func,  // return current character and push ptr to prev one
                     this_char_func;   // return current character and do not change ptr

public:
    search_env(search_func     use_search,
               compile_func    use_compile,
               search_nav_func use_for_next,
               search_nav_func use_for_prior,
               search_nav_func use_for_this);
           
    /*inquiry functions*/
    search_env*     get_next(){return next;}
    void            set_next(search_env *new_next){next = new_next;}
    search_func     get_search(){return search_rtn;}
    compile_func    get_compile(){return compile_rtn;}
    search_nav_func get_next_func(){return next_char_func;}
    search_nav_func get_prior_func(){return prior_char_func;}
    search_nav_func get_this_func(){return this_char_func;}
};

class model{
    long             ref_count;
    node            *root_node;
    obj_heap        *current_oheap;
    goo_heap        *current_gheap;
    st_heap         *current_sheap;
    oheap_list      *first_oheap;
    gheap_list      *first_gheap;
    sheap_list      *first_sheap;
    long             oheap_size;
    long             gheap_size;
    long             sheap_size;
    plid_collection *model_plid_table;
    prop_state      *model_state;
    node            *prop_pos;
    history         *past;
    mf_list         *attached_formats;
    ft_list         *attached_fonts;
    mark_list       *attached_marks;
    search_env      *search_stack;
    union_char_set  *char_sets;
    csm_lock        *model_lock;
    csm_lock        *heaps_lock;

    friend class node;
    friend class prop_vec;
    friend class property;
    friend class argument;
    friend class obj_heap;
    friend class goo_heap;
    friend class st_heap;
    friend class plid_collection;
    friend class plid_trace;
    friend void gc_linearize(model*);
public:
    /*constructor and destructor for a model*/
    model(long obj_heap_size = 0, long goo_heap_size = 0, long st_heap_size = 0);
    ~model();

    /*inquiry functions for the model data*/
    node*            get_root(){return root_node;}
    oheap_list*      get_first_oheap(){return(first_oheap);}
    gheap_list*      get_first_gheap(){return(first_gheap);}
    sheap_list*      get_first_sheap(){return(first_sheap);}
    obj_heap*        get_current_oheap(){return(current_oheap);}
    goo_heap*        get_current_gheap(){return(current_gheap);}
    st_heap*         get_current_sheap(){return(current_sheap);}
    long             get_oheap_size(){return oheap_size;}
    long             get_gheap_size(){return gheap_size;}
    long             get_sheap_size(){return sheap_size;}
    node*            get_prop_pos(){return(prop_pos);}
    mf_list*         get_formats(){return attached_formats;}
    ft_list*         get_fonts(){return attached_fonts;}
    mark_list*       get_marks(){return attached_marks;}
    union_char_set*  get_char_sets(){return char_sets;}
    csm_lock*        get_model_lock(){return model_lock;}
    csm_lock*        get_heaps_lock(){return heaps_lock;}

    /*functions to directly manipulate the model data*/
    node*       make_root();
    node*       set_root(node* new_root);
    void        attach(obj_heap *new_heap);
    void        attach(goo_heap *new_heap);
    void        attach(st_heap *new_heap);
    void        attach(union_char_set* new_union_set);
    void        detach(obj_heap *old_heap);
    void        detach(goo_heap *old_heap);
    void        detach(st_heap *old_heap);
    void        detach(union_char_set* old_union_set);
    void        set_current_oheap(obj_heap *cur_oheap){current_oheap = cur_oheap;}
    void        set_current_gheap(goo_heap* cur_gheap){current_gheap = cur_gheap;}
    void        set_current_sheap(st_heap *cur_sheap){current_sheap = cur_sheap;}
    void        set_oheap_size(long obj_heap_size){oheap_size = obj_heap_size;}
    void        set_gheap_size(long goo_heap_size){gheap_size = goo_heap_size;}
    void        set_sheap_size(long st_heap_size){sheap_size = st_heap_size;}
    void        add_format(format *new_format);
    void        delete_format(format *del_format);
    void        set_font_tbl(ft_list *new_entry){attached_fonts = new_entry;}
    void        set_char_sets(union_char_set* new_char_sets){char_sets = new_char_sets;}

    /*functions manipulating the property state of a model*/
    prop_state* get_state(){return model_state;}
    void*     inq_state(p_index prop_ind){return model_state->inq_state(prop_ind);}
    void*     inq_state(node *at_node, p_index prop_ind);
    node*     move_prop(node *to_node);
    void      set_prop_pos(node *new_prop_pos){prop_pos = new_prop_pos;}

    /*functions which allocate space in a model*/
    node*     make_node(long node_size);
    prop_vec* make_pvec();
    property* make_prop(long property_size = 0);
    argument* make_arg(long new_value);
    argument* make_arg(float new_value);
    argument* make_arg(char* new_value, long new_count);
    argument* make_arg(node *subtree);
    argument* make_arg(char_set *new_char_set);
    argument* make_arg(union_char_set *new_union_char_set);
    char*     make_goo(long count, char *contents = 0);
    char*     make_obj(long obj_size);

    /*functions used to save, restore, and compact a model*/
    void      make_model_lock(){model_lock = new csm_lock(this);}
    void      lock_busy(){if (model_lock) model_lock->p_busy();}
    void      lock_suspend(){if (model_lock) model_lock->p_suspend();}
    long      lock_bounce();
    void      unlock(){if (model_lock) model_lock->v();}
    void      make_heaps_lock(){heaps_lock = new csm_lock(this);}
    void      lock_heaps_busy(){if (heaps_lock) heaps_lock->p_busy();}
    void      lock_heaps_suspend(){if (heaps_lock) heaps_lock->p_suspend();}
    long      lock_heaps_bounce();
    void      unlock_heaps(){if (heaps_lock) model_lock->v();}
    long      d_write_model(void *file_handle);   // save model on disk to file indicated
                                                  //   by file handle returned by
                                                  //   tml_os_open()
    long      d_write_model(char* pathname, long name_len); // save model on disk to
                                                           //   file indicated by
                                                           //   pathname and name length
    void      move_gheaps(model *to_model);

    /*functions used for history, undo, redo, journalling*/
    history*  get_past(){return past;}

    /*cut and paste functions*/
    node*     cut(node *cut_node);
    node*     paste(node *paste_node, node *paste_to, node_dir paste_dir);
                                                                      
    /*mark functions*/
    mark_list*  get_first_mark(){return attached_marks;}
    mark_list*  make_mark(model_range init_range);
    void        delete_mark(mark_list *to_delete);
    long        mark_draw(mark_list *to_draw);
    long        mark_erase(mark_list *to_erase);
                                  
    /*search functions*/
    void        return_to_default_search_env();
    void        push_search_env(search_env *new_envir);
    void        pop_search_env();
    search_env *get_search_env() {return search_stack;}
    model_range search_exp(void *pattern, model_range search_range, long forward = 1);
    long        compile_exp(char *search_exp, long search_exp_len, void *compiled_exp,
                            long max_comp_len, long case_sens = 1);
    model_range search(char *search_exp, long search_exp_len, model_range search_range,
                       long forward = 1, long case_sens = 1);

    /*function to update any display of a model after the model has been changed*/
    void        update_display();
};
                                               
typedef void* (*prop_func_pt)(argument* func_arg);

extern prop_func_ent *get_prop_func_tbl_entry(long tbl_index);
extern void set_prop_func_tbl_entry(long tbl_index, prop_func_ent *entry);

class prop_func_ent{
    prop_func_ent   *next;
    p_index          prop_index;
    prop_func_pt     entry,
                     exit;
    short            loaded;
public:
    prop_func_ent();
    ~prop_func_ent();
    void            set_next(prop_func_ent *next_ent){next = next_ent;}
    prop_func_ent*  get_next(){return next;}
    void            set_index(p_index i){prop_index = i;}
    p_index         get_index(){return prop_index;}
    void            load_entry(prop_func_pt new_entry){ entry = new_entry;}
    void            load_exit(prop_func_pt new_exit){exit = new_exit;}
    prop_func_pt    get_entry(){return entry;}
    prop_func_pt    get_exit(){return exit;}
    short           get_loaded(){return loaded;}
    void            set_loaded(short i){loaded = i;}
    void            load();

};


/**********************************************************************************
    Objects used in writing/reading tml models to/from disk
**********************************************************************************/

class d_plid {
    long    index;
    node    *with_node;              // associated with what node
    long    offset;                  // at the offset
    long    ref_count;               // how many references to this plid exist


    friend class d_model_parts;
};

class d_node {
    node*      right;
    prop_vec*  properties;
    node_type  n_type;
    long       num_plid_traces;
public:
    void         set_right(d_ptr *d_ptr_tmp);
    void         set_pv_d_ptr(d_ptr *d_ptr_tmp);
    void         set_type_index(node* save_node);
    node*        get_right(){return(right);}
    prop_vec*    get_props(){return(properties);}
    node_type    get_type_index(){return(n_type);}
    void         unscramble(endian* their_endian);
};

class d_bnode : public d_node {
    node* down_ptr;
public:
    void  set_down_ptr(d_ptr *d_ptr_tmp);
    node* get_down_ptr(){return(down_ptr);}
    void  unscramble(endian* their_endian);
};

class d_fnode : public d_node {
    unsigned char* filename;
    long           count;
};

class d_gnode : public d_node {
    char* goo_ptr;
    long  count;
public:
    void  set_count(long new_count){count = new_count;}
    void  set_goo_d_ptr(d_ptr* d_ptr_tmp);
    long  get_count(){return(count);}
    char* get_goo_ptr(){return(goo_ptr);}
    void  unscramble(endian* their_endian);
};

class d_prop_vec {
    prop_vec* next;
    property* prop;
public:
    void      set_d_ptr(d_ptr* d_ptr_tmp);       // set next field to pseudo-pointer for disk save
    void      set_prop_d_ptr(d_ptr* d_ptr_tmp);  // set prop field to pseudo-pointer for disk save
    prop_vec* get_next(){return(next);}
    property* get_prop(){return(prop);}
    void      unscramble(endian* their_endian);
};

class d_property {
    argument* entry_arg;
    argument* exit_arg;
    p_index   index;
public:
    void         set_entry_d_ptr(d_ptr *d_ptr_tmp);
    void         set_exit_d_ptr(d_ptr *d_ptr_tmp);
    void         set_type_index(property* save_prop);
    argument*    get_entry_arg(){return(entry_arg);}
    argument*    get_exit_arg(){return(exit_arg);}
    p_index      get_type_index(){return(index);}
    void         unscramble(endian* their_endian);
};

class d_argument {
    arg_type  vtype;
    argument* next;
    void*     arg_val;
public:
    void           set_d_ptr(d_ptr *d_ptr_tmp);     // set next field to pseudo-pointer for disk save
    void           set_val_d_ptr(d_ptr *d_ptr_tmp); // set value pointer field to pseudo-pointer for disk save
    void           set_type_index(argument* save_arg);
    argument*      get_next(){return(next);}
    void*          get_val(){return(arg_val);}
    arg_type       get_type_index(){return(vtype);}
    void           unscramble(endian* their_endian);
};


class d_string_arg : public d_argument {
    long count;
public:
    void set_count(long num_bytes){count = num_bytes;}
    long get_count(){return count;}
    void unscramble(endian* their_endian);
};

/**********************************************************************************
    Objects used in reading TML version 1 tml models from disk
**********************************************************************************/

class tml_ext_type_v1 { // members public as statically initialized
public:
    tml_uint32 high;
    tml_uint32 low;

    friend long operator==(tml_ext_type_v1 t1, tml_ext_type_v1 t2)
        {return (t1.high == t2.high)&&(t1.low == t2.low);}
    void unscramble(endian* their_endian){high = (tml_uint32) their_endian->my_long((unsigned long) high);
                                          low  = (tml_uint32) their_endian->my_long((unsigned long) low );}
};

class type_trans_v1 { // members public as statically initialized
public:
    tml_ext_type_v1 *type_v1;
    tml_type        *type;
    GEN_FUNCP       make_func;  // returns pointer to an object
                                //   of the appropriate type
                                //   must be a member function
                                //   of class model or general
                                //   function call taking model
                                //   as first argument
    type_trans_v1   *next;
};

class type_trans_queue_v1 { // members public as statically initialized
public:
    type_trans_v1 *head;
    type_trans_v1 *tail;

    void           queue_type_trans(type_trans_v1* new_trans);
};

class d_model_info_v1 {
public:
    long          version_number;
    d_ptr         root;
    long          num_types;
    long          num_oheaps,
                  num_gheaps,
                  num_sheaps;
    long          last_o,
                  last_g,
                  last_s;
    HALF_PTR      cur_oheap,
                  cur_gheap,
                  cur_sheap;
    HALF_PTR      first_oheap;
    HALF_PTR      first_gheap;
    HALF_PTR      first_sheap;

    void unscramble(endian* their_endian);
};

class node_v1 {
    model           *in_model;
    node_v1         *up,
                    *right,
                    *left;
    prop_vec        *properties;
    short            sib_order;  // sibling order member was short
    node_type        n_type;     //   at TML Version 1
    void            *dummy_vptr; // this class has no virtual functions,
                                 //   but TML Version 1 node did

    friend class d_model_parts;
    friend class node;
public:
    node_v1(long size, model *to_model = 0); // size for derived classes
    node_type get_type(){return(n_type);}
    void      set_type(node_type new_type){n_type = new_type;}
    void      unscramble(endian* their_endian);
};

class bnode_v1 : public node_v1{
    node_v1     *down_ptr;

    friend class bnode;
public:
    bnode_v1(model *to_model = 0, long size = 0); // size for extensions paralleling possible
                                                  //   user extensions to TML Version 1 gnode
    void unscramble(endian* their_endian);
};

class gnode_v1 : public node_v1{
    char        *goo_ptr;
    short        count;  // count was short at TML Version 1

    friend class gnode;
public:
    gnode_v1(model *to_model = 0, long size = 0); // size for extensions paralleling possible
                                                  //   user extensions to TML Version 1 gnode
    void unscramble(endian* their_endian);
};

class property_v1 {
    union{
        model           *in_model;   //what model is this in--used when alloced
        property_v1     *next;       //next free property slot--used when free
        } locator;
    short                ref_count;  //how many vectors use this property
    p_index              index;      //index into the property function table
    argument_v1         *entry_arg,  //head of the entry argument list
                        *exit_arg;   //head of the exit argument list

    friend class prop_vec;
    friend class property;
public:
    void unscramble(endian* their_endian);
};
                 

class argument_v1 {
    argument_v1        *next;
    model              *in_model;
    arg_type_v1         a_type;
    long                ref_count;
    union{
        long            ival;
        float           fval;
        struct{
            char       *s_start;
            long        s_count;
            }       sval;
        }       arg_val;

    friend class property;
    friend class argument;
public:
    void unscramble(endian* their_endian);
};


/**********************************************************************************
    Objects used in reading from disk models written by TML versions 2 through 4
**********************************************************************************/

class d_model_info_v4 {                        // information from object d_model_parts that
public:                                        //   is actually written to disk during
    long           version_number;             //   during disk save.
    d_ptr          root;
    long           num_types;                  // number of universal type identifiers stored on disk file
    long           num_plids;                  // number of used plids in the plid_collection to restore from disk
    long           num_heaps[SQUEEZE_HEAP + 1];
    long           heap_quanta[SQUEEZE_HEAP + 1];
    HALF_PTR       cur_heap[SQUEEZE_HEAP + 1];

    void unscramble(endian* their_endian);
};


/**********************************************************************************
    utility function to return head of queue of TML Version 1 
      preloaded types and user extensions dervied from them
**********************************************************************************/

type_trans_v1*   tml_type_trans_q_v1_get_head();


/**********************************************************************************
    Objects used in the formatting component of tml
**********************************************************************************/

/*the conceptual line base class*/
class cline{
    format      *in_format;
    phead       *on_phead;
    cline       *next,
                *prior;
    cl_type      type;
    short        ulx, uly,
                 width, height;
    
public:
    cline(format *to_format = 0);
    ~cline();

    /*traits to access conceptual line data*/
    cl_type     get_type(){return type;}
    short       get_ulx(){return ulx;}
    short       get_uly(){return uly;}
    short       get_lrx(){return (ulx + width);}
    short       get_lry(){return (uly + height);}
    short       get_hi(){return height;}
    short       get_wide(){return width;}
    phead*      get_phead(){return on_phead;}
    format*     get_format(){return in_format;}
    cline*      inq_next(){return next;}
    cline*      inq_prior(){return prior;}
    cline*      get_down();
    cline*      get_up();

    /*traits to manipulate conceptual line data*/
    void        set_phead(phead* on_page){on_phead = on_page;}
    void        set_type(cl_type new_type){type = new_type;}
    void        set_wide(short new_width){width = new_width;}
    void        set_hi(short new_hi){height = new_hi;}

    /*traits to do things to a conceptual line*/
    cline*      split(short new_size, short from_top = 0, short direction = 0 );
    long        resize(short hi_delta);
    void        attach(cline *target_line, long before);
    void        detach();

    /*virtual functions for this class*/
    virtual cline*      get_next();
    virtual void        set_next(cline* next_line){next = next_line;}
    virtual void        set_prior(cline* prior_line){prior = prior_line;}
    virtual cline*      get_prior();
    virtual model_pos   get_start();
    virtual model_pos   get_end();
    virtual void        set_ul_coord(short upper_x, short upper_y){ulx = upper_x; uly = upper_y;}
    virtual void        set_start_x_off(long new_x_off);
    virtual long        get_start_x_off();
    virtual long        get_contents_width(){return 5;}
    virtual void        adjust_xy(long deltax, long deltay){ ulx += (short) deltax; uly += (short) deltay;}
    virtual long        contains(model_pos target_pos);
    virtual long        can_draw(){return((type == cl_gutter) ? 1 : 0);}
    virtual void        replace(cline *replace_line);
    virtual void        release(){delete this;}
    virtual cline*      alloc_space(cline *fill_with, long at_bottom = 0);
    virtual xy_coord    mpos_2_xy(model_pos find_pos);
    virtual model_pos   xy_2_mpos(xy_coord where);
    virtual view_pos*   draw_line(view_port *in_viewport, screen_xy where, long start_index = -1);
    virtual cline*      reformat(model_pos new_start, long& next_ok, long force_start = 0);
};

class tline:public cline {
    tline       *right,
                *left;
    font_tbl    *use_font;
    model_pos    start_pos;
    long         count,
                 start_x_offset;
    short        baseline_y,
                 text_width;

    void restore_left_lines(view_port* in_viewport, screen_xy where, char* this_draw_st, unsigned long this_draw_len);
    void restore_left_fragment(view_port* in_viewport, screen_xy where, char* this_draw_st, long this_start_index);

public:
    /*constructor and destructor for derived class tline*/
    tline(format *to_format = 0);
    ~tline();

    /*traits to access the text line data*/
    cline*      get_right(){return right;}
    cline*      get_left(){return left;}
    font_tbl*   get_font(){return use_font;}
    short       get_baseline(){return baseline_y;}
    long        get_count(){return count;}
    long        get_start_x_off(){return start_x_offset;}

    /*traits to manipulate the text line data*/
    long        adjust_baseline(long new_baseline_y);
    void        set_baseline(long new_baseline){baseline_y = (short) new_baseline;}
    void        set_font(font_tbl *new_font){use_font = new_font;}
    void        set_start(model_pos new_start){start_pos = new_start;}
    void        set_count(long new_count){count = new_count;}
    void        set_text_width(long new_width){text_width = (short) new_width;}
    long        get_text_width(){return text_width;}
    long        get_full_text_width();  //gets the width of this and all right siblings
    void        set_right(tline *new_right);
    void        set_left(tline  *new_left);
    void        set_start_x_off(long new_off){start_x_offset = new_off;}
                                             
    long        trunc_line (long max_x);

    /*implementations of the virtual functions*/
    cline*      get_next();
    cline*      get_prior(){return (left ? left : cline::get_prior());}
    void        set_next(cline *next_line);
    void        set_prior(cline *prior_line);
    model_pos   get_start(){return start_pos;}
    model_pos   get_end();
    long        get_contents_width(){return get_full_text_width();}
    void        set_ul_coord(short upper_x, short upper_y);
    void        adjust_xy(long deltax, long deltay);
    long        contains(model_pos target_pos);
    long        can_draw(){return 1;}
    void        replace(cline *replace_line);
    void        release();
    xy_coord    mpos_2_xy(model_pos find_pos);
    model_pos   xy_2_mpos(xy_coord where);
    view_pos*   draw_line(view_port *in_viewport, screen_xy where, long start_index = -1);
    cline*      reformat(model_pos new_start, long& next_ok, long force_start = 0);
};

class phead{
    phead       *next,
                *prior;
    short        dx,
                 dy;
    format      *in_format;
    long         page_num,
                 fline_no;
    long         max_line_width, max_used_ht;
    short        format_ok,     //boolean value--0 = page needs reformatting, 1 = format valid
                 line_cnt;
    model_pos    start_pos,     //model position of the first character on the page
                 end_pos;       //model position of the last character on the page
    cline       *top_line,
                *bottom_line,
                *top_free,
                *bottom_free,
                *top_used,
                *bottom_used;

    /*private functions used by the formatting function*/
    void    set_margins();
    long    set_header_footer();
    cline*  make_blank(cline *blank_line);

    friend class format;
    friend long   cline::resize(short hi_delta);
    friend cline* cline::split(short new_size, short from_top, short direction);
    friend void   cline::replace(cline *replace_line);
    friend cline* cline::alloc_space(cline *fill_with, long at_bottom);
    friend cline* tline::reformat(model_pos new_start, long& next_ok, long force_start);

public:
    phead(format *to_format = 0);
    ~phead();
    /*traits allowing access to the phead data fields*/
    phead*      get_next(){return next;}
    phead*      get_prior(){return prior;}
    long        get_page_no(){return page_num;}
    long        get_fline_no(){return fline_no;}
    long        get_max_line_width(){return max_line_width;}
    long        get_max_used_ht(){return max_used_ht;}
    long        get_line_cnt(){return line_cnt;}
    model_pos   get_start(){return start_pos;}
    model_pos   get_end(){return end_pos;}
    format*     get_format(){return in_format;}
    long        get_format_ok(){return format_ok;}
    long        get_hi(){return dy;}
    long        get_wide(){return dx;}
    cline*      get_freeline(short height, short from_top = 1);
    cline*      get_tline();
    cline*      get_bline();
    cline*      get_tfree();
    cline*      get_bfree();
    cline*      get_tused();
    cline*      get_bused();
    long        is_visible();

    /*traits which alter the phead data fields directly*/
    void        set_format(format* new_format){in_format = new_format;}
    void        set_size(short hi, short wide);
    void        set_start_pos(model_pos new_start){start_pos = new_start;}
    void        set_end_pos(model_pos new_end){end_pos = new_end;}
    void        mark_reformat(){format_ok = 0;}
    void        set_top_line(cline *new_top){top_line = new_top;}
    void        set_bottom_line(cline *new_bottom){bottom_line = new_bottom;}
    void        set_top_free(cline *new_tfree){top_free = new_tfree;}
    void        set_bottom_free(cline *new_bfree){bottom_free = new_bfree;}
    void        set_top_used(cline *new_tused){top_used = new_tused;}
    void        set_bottom_used(cline *new_bused){bottom_used = new_bused;}
    void        adjust_line_cnt(long adjust_amt);
    void        set_max_line_width(long new_width);
    void        calc_max_line_width();
    void        set_max_used_ht(long new_ht);
    void        calc_max_used_ht ();

    /*navigation and creation functions*/
    phead*      next_page(long reformat = 0);
    phead*      prior_page(long reformat = 0);
    phead*      insert_page(model_pos start, long after = 1);
    cline*      mpos_2_cline(model_pos target);
    cline*      mpos_2_hcline(model_pos target);

    /*traits that manipulate the conceptual lines on the page*/
    void        free_clines(cline *start_line = 0);
    long        line_place(cline *place_line, short at_x, short at_y);
    void        adjust_xy(cline *start_line, long deltax, long deltay);

    /*format the page*/
    model_pos   format_page(model_pos start);
    model_pos   tamrof_page(model_pos start);
    model_pos   reformat_page(model_pos start);
};
                                                       
class fv_list{
    fv_list     *next,
                *prior;
    view_port   *with_vp;
    
public:
    fv_list*    inq_next(){return next;}
    fv_list*    inq_prior(){return prior;}
    view_port*  inq_vp(){return with_vp;}

    void        set_next(fv_list *new_next){next = new_next;}
    void        set_prior(fv_list *new_prior){prior = new_prior;}
    void        set_vp(view_port *new_vp){with_vp = new_vp;}
};

class rp_list {
    rp_list     *next,
                *prior;
    phead       *rt_page;

public:
    rp_list*    get_next(){return next;}
    rp_list*    get_prior(){return prior;}
    phead*      get_page(){return rt_page;}

    void        set_next(rp_list* new_next){next = new_next;}
    void        set_prior(rp_list* new_prior){prior = new_prior;}
    void        set_page(phead *new_page){rt_page = new_page;}
};

class format{
    phead       *first_phead,
                *last_phead;
    short        ref_count,
                 retain_pages;
    prop_state  *format_state;
    model       *this_model;
    fv_list     *attached_vps;
    rp_list     *retained_top,
                *retained_bottom,
                *top_free_rt_list;
    long         page_wide,
                 page_hi,
                 max_line_width,
                 max_used_ht;
    friend model_pos phead::format_page(model_pos start);

public:
    format();
    ~format();

    /*traits which allow access to the format data*/
    long        get_ref_count(){return ref_count;}
    model*      get_model(){return this_model;}
    phead*      get_first_page(){return first_phead;}
    phead*      get_last_page(){return last_phead;}
    void*       inq_state(p_index prop_ind);
    void*       inq_state(node *at_node, p_index prop_ind);
    void        add_viewport(view_port *add_vp);
    void        delete_viewport(view_port *del_vp);
    fv_list*    get_fv_list(){return attached_vps;}
    long        get_page_hi(){return page_hi;}
    long        get_page_wide(){return page_wide;}
    long        get_max_line_width(){return max_line_width;}
    long        get_max_used_ht(){return max_used_ht;}

    /*traits which allow manipulation of the format data*/
    void    inc_ref_count(){ref_count++;}
    void    dec_ref_count(){ref_count--;}
    void    set_page_size(short hi, short wide);
    void    set_pages_saved(short save_num){retain_pages = save_num;}
    void    set_first_page(phead *new_first){first_phead = new_first;}
    void    set_last_page(phead *new_last){last_phead = new_last;}
    long    set_state(p_index prop_ind, void* value);
    long    unset_state(p_index prop_ind);
    void    make_rp_entry(phead *new_page);
    void    purge_rp_list();
    void    set_max_line_width(long new_width){max_line_width = new_width;}
    void    calc_max_line_width();
    void    set_max_used_ht(long new_ht){max_used_ht = new_ht;}
    void    calc_max_used_ht();

    /*traits which manipulate the format*/
    void    attach(model *to_model);
    void    detach();
                                          
    /*navigation traits*/
    phead*  mpos_2_phead(model_pos find_pos);
    phead*  get_page(long page_no);
};


class width_tbl{
    unsigned long   start_index;
    unsigned long   max_index;
    unsigned long   dflt_char;
             long   dflt_width;
             short* widths;
             short* lbearings;
             short* rbearings;

public:
    width_tbl(unsigned long default_char, unsigned long start, unsigned long max, short** width_tabs);
    ~width_tbl();
    unsigned long   get_num_entries(){return 1 + max_index - start_index;}
             short* get_widths(){return widths;}
             short* get_lbearings(){return lbearings;}
             short* get_rbearings(){return rbearings;}
             long   get_width(unsigned long charac)   {return ((start_index <= charac) && (charac <= max_index))
                                                              ? *(widths    + charac - start_index)
                                                              : get_width(dflt_char);}
             long   get_lbearing(unsigned long charac){return ((start_index <= charac) && (charac <= max_index))
                                                              ? *(lbearings + charac - start_index)
                                                              : get_lbearing(dflt_char);}
             long   get_rbearing(unsigned long charac){return ((start_index <= charac) && (charac <= max_index))
                                                              ? *(rbearings + charac - start_index)
                                                              : get_rbearing(dflt_char);}
};

class font_tbl{
    char      *name;           //full pathname to font
    void      *font_handle;    //handle to device dependent font id
    long       name_count,     //number of characters in name
               char_size,      //number of bytes per character
               num_chars,      //number of characters in font
               first_char,     //index of the first character in font
               font_hi,        //height of the font
               icspace,        //inter-character spacing
               baseline,       //baseline offset from top of font matrix
               fixed,          //0=proportional font, otherwise width in fixed width font
               is_loaded;      //0=font not loaded, otherwise loaded
    width_tbl *width;          //server object for escapement table - proportional fonts only
    vpftbl    *ftbl;           //table of functions dependent on graphics system
public:
    font_tbl(vpftbl* func_tab = 0);
    ~font_tbl();
    
    /*member functions to access the data*/
    char*      get_name(){return name;}
    long       get_name_count(){return name_count;}
    void*      get_handle(){return font_handle;}
    long       get_hi(){return font_hi;}
    long       get_baseline(){return baseline;}
    long       get_width(unsigned long charac){return(fixed ? fixed : width->get_width(charac));};
    long       get_lbearing(unsigned long charac){return(fixed ? 0 : width->get_lbearing(charac));};
    long       get_rbearing(unsigned long charac){return(fixed ? 0 : width->get_rbearing(charac));};
    long       get_char_size(){return char_size;}
    long       get_icspace(){return icspace;}
    long       get_leading();
    long       get_point_size();
    width_tbl* get_width_table(unsigned long* num_chars);
    vpftbl*    get_ftbl();
            
    /*member functions to set the data*/
    long    set_info(char *fname, long flen);
    long    load_font();
    long    set_leading(long lead_pixels);
    void    set_baseline (long new_bl) {baseline = new_bl;}
    void    set_hi (long new_hi) {font_hi = new_hi;}
    void    set_icspace(long new_icspace){icspace = new_icspace;}
    void    set_width_table (width_tbl* new_widths){width = new_widths;}
    void    set_fbtl(vpftbl* new_table){ftbl = new_table;};
};

/**********************************************************************************
    Objects used in the rendering component of tml
**********************************************************************************/


extern const screen_xy zero_zero;

/*typedef for a pointer to the function which draws the cursor*/
typedef long (*cursor_draw)(void *in_bitmap, long wherex, long wherey,
                            long cursor_hi, long cursor_wide, long erase = 0);
                                                                 
/*the viewport cursor class*/
class v_cursor{
    view_port   *in_viewport;   //viewport in which the cursor occurs
    screen_xy    raw_loc,       //the x,y coordinate of the cursor uncaged to character location
                 caged_loc;     //the x,y coordinate of the cursor caged to character location
    view_pos    *in_view_pos;   //screen location in which the cursor resides
    cline       *in_cline;      //the conceptual line the cursor is on
    model_pos    after_char;    //character the cursor is behind
    long         is_visible;    //< 1 cursor not drawn; otherwise cursor drawn
    long         is_enabled;    //< 1 cursor can't be drawn; otherwise draw when visible
    short        height,        //dimensions of the current cursor graphic
                 width;
    cursor_draw  draw_func;     //cursor drawing routine
      
public:
    v_cursor();
    /*traits to access the data*/
    view_port*  get_viewport(){return in_viewport;}
    screen_xy   get_raw_loc(){return raw_loc;}
    screen_xy   get_caged_loc(){return caged_loc;}
    view_pos*   get_view_pos(){return in_view_pos;}
    cline*      get_cline(){return in_cline;}
    model_pos   get_model_pos(){return after_char;}
    long        get_visible(){return is_visible;}
    cursor_draw get_draw_func(){return draw_func;}

    /*four virtual data access functions allow alternate cursors*/
    virtual     long  get_height(){return (long) height;}
    virtual     long  get_width(){return (long) width;}
    virtual     long  get_origin_x(){return caged_loc.at_x;}
    virtual     long  get_origin_y(){return caged_loc.at_y;}

    /*traits to manipulate the data*/
    void        attach(view_port* to_viewport){in_viewport = to_viewport;}
    void        detach();
    void        set_model_pos(model_pos new_pos){after_char = new_pos;}
    long        enable(long force = 0);
    long        disable(long force = 0);
    void        set_visible(long new_vis){is_visible = new_vis;}

    /*traits to draw and move the cursor*/
    long        draw();
    long        erase();
    void        set_draw_func(cursor_draw new_func);
    screen_xy   move(vp_dir move_dir);
    long        set(screen_xy position);
    long        set(model_pos position);
};


typedef void (*core_m_draw)(void* bitmap_handle, long src_ulx, long src_uly, 
                                                 long src_lrx, long src_lry, 
                                                 long dst_ulx, long dst_uly, 
                                                 long rop);

typedef long (*draw_m_draw)(view_port *in_viewport, core_m_draw core_func, 
                            model_pos start_draw, model_pos end_draw, 
                            long erase = 0);

/*the class of marks drawn in the viewport*/
class draw_mark{
    draw_mark       *next,
                    *prior;
    mark_list       *model_mark;
    long             is_visible,
                     is_enabled;
    view_port       *in_viewport;
    core_m_draw      core_func;
    draw_m_draw      draw_func;

public:
    draw_mark();
    ~draw_mark();
    draw_mark*      get_next(){return next;}
    draw_mark*      get_prior(){return prior;}
    mark_list*      get_model_mark(){return model_mark;}
    view_port*      get_viewport(){return in_viewport;}
    long            get_enabled(){return is_enabled;}
    draw_m_draw     get_draw_func(){return draw_func;}
    core_m_draw     get_core_func(){return core_func;}

    void            set_next(draw_mark *new_next){next = new_next;}
    void            set_prior(draw_mark *new_prior){prior = new_prior;}
    void            set_model_mark(mark_list *new_mod_mark){model_mark = new_mod_mark;}
    void            set_view_port(view_port *new_vp){in_viewport = new_vp;}
    void            set_draw_func(draw_m_draw new_draw);
    void            set_core_func(core_m_draw new_draw);

    void            enable(long force = 0)
                      {if (force) is_enabled = 1; else is_enabled++;}
    void            disable(long force = 0)
                      {if (force) is_enabled = 0; else is_enabled--;}
    
    long            extend(model_pos new_start, model_pos new_end);
    long            draw();
    long            erase();
};


/* view_port function table -- table of rendering functions*/
class vpftbl {
public:
    char*   (*ptr_2_tml_dp_get_default_font)(long& name_len);
    void*   (*ptr_2_tml_dp_get_fhandle)(char *font_name, long name_len);
    long    (*ptr_2_tml_dp_load_font)(void *font_handle);
    long    (*ptr_2_tml_dp_get_fcharsize)(void *font_handle);
    long    (*ptr_2_tml_dp_get_fcharnum)(void *font_handle);
    long    (*ptr_2_tml_dp_get_fchar1)(void *font_handle);
    long    (*ptr_2_tml_dp_get_fhi)(void *font_handle);
    long    (*ptr_2_tml_dp_get_fpoint)(void *font_handle);
    long    (*ptr_2_tml_dp_get_fbline)(void *font_handle);
    long    (*ptr_2_tml_dp_is_fixed)(void *font_handle);
    short** (*ptr_2_tml_dp_get_fwidths)(void *font_handle);
    long    (*ptr_2_tml_dp_get_ficspace)(void *font_handle);
    long    (*ptr_2_tml_dp_set_ficspace)(void *font_handle, long new_icspace);
    long    (*ptr_2_tml_dp_get_fspace)(void *font_handle);
    void    (*ptr_2_tml_dp_set_fspace)(void *font_handle, long new_space);
    long    (*ptr_2_tml_dp_funload)(void *font_handle);
    void    (*ptr_2_tml_dp_frelease)(void *font_handle);
    void*   (*ptr_2_tml_dp_make_graphic_handle)(void *bitmap_id, void *viewport_id);
    void    (*ptr_2_tml_dp_delete_graphic_handle)(void *bitmap_id);
    long    (*ptr_2_tml_dp_get_hi)(void *bitmap_id);
    long    (*ptr_2_tml_dp_get_wide)(void *bitmap_id);
    long    (*ptr_2_tml_dp_get_foreg_id)(void *bitmap_id);
    long    (*ptr_2_tml_dp_get_backg_id)(void *bitmap_id);
    long    (*ptr_2_tml_dp_get_fill_backg_id)(void *bitmap_id);
    long    (*ptr_2_tml_dp_get_fill_obj_id)(void *bitmap_id);
    void    (*ptr_2_tml_dp_set_foreg_id)(void *bitmap_id, long new_fg_id);
    void    (*ptr_2_tml_dp_set_backg_id)(void *bitmap_id, long new_bg_id);
    void    (*ptr_2_tml_dp_set_fill_backg_id)(void *bitmap_id, long new_fill_bg_id);
    void    (*ptr_2_tml_dp_set_fill_obj_id)(void *bitmap_id, long new_fill_obj_id);
    void    (*ptr_2_tml_dp_set_clip)(void *bitmap_id, long clip_ulx, long clip_uly, long clip_lrx, long clip_lry);
    void    (*ptr_2_tml_dp_clear_clip)(void *bitmap_id);
    void    (*ptr_2_tml_dp_clear)(void *bitmap_id, long ulx, long uly, long lrx, long lry);
    void    (*ptr_2_tml_dp_copy)(void *bitmap_id, long src_ulx, long src_uly, long src_lrx, long src_lry, long dst_ulx, long dst_uly);
    void    (*ptr_2_tml_dp_blt)(void *bitmap_id, long src_ulx, long src_uly, long src_lrx, long src_lry, long dst_ulx, long dst_uly, long rop_id);
    long    (*ptr_2_tml_dp_string_draw)(void *bitmap_id, char *string, long st_count, long wherex, long wherey, void *use_font);
    long    (*ptr_2_tml_dp_string_size)(char *string, long st_count, void *use_font);
    void    (*ptr_2_tml_dp_damage_repair)(void *bitmap_id);
    long    (*ptr_2_tml_dp_dflt_cursor_draw)(void *bitmap_id, long wherex, long wherey,
                                             long height, long width, long erase = 0);
    void*   (*ptr_2_tml_dp_get_grafsys_obj)(void* bitmap_id);
    char*   (*ptr_2_get_default_font_name)(...);
    char*   (*ptr_2_set_default_font_name)(...);

    tml_type* grafsys_type;

};

/*the viewport class*/
class view_port{
    void        *bitmap_handle;     //pointer to identifier for the bitmap
    long         dx,                //width (in pixels) of viewport
                 dy,                //height (in pixels) of the viewport
                 h_offset;          //horizontal scrolling offset
    view_pos    *top_line,          //top line visible in the viewport
                *bottom_line;       //bottom line visible in the viewport
    format      *this_format;       //what format is being displayed
    v_cursor    *cursor;            //the cursor in this viewport
    long         foreg_id,          //id of foreground color
                 backg_id;          //id of the background color
    prop_state  *view_state;        //state of the properties overriden by the bitmap
    view_pos    *tf_vp;             //most recently freed view_pos
    draw_mark   *first_mark;        //the head of the drawn marks list
    vpftbl      *ftbl;              //table of rendering functions
                
    void        disable_all_marks(long force = 0);
    void        enable_all_marks(long force = 0);
    void        erase_all_marks();
    void        draw_all_marks();
public:
    view_port();
    ~view_port();

    /*traits to access the data*/
    void*       get_handle(){return bitmap_handle;}
    long        get_width(){return dx;}
    long        get_hi(){return dy;}
    long        get_hoffset(){return h_offset;}
    view_pos*   get_top_line(){return top_line;}
    view_pos*   get_bottom_line(){return bottom_line;}
    format*     get_format(){return this_format;}
    v_cursor*   get_cursor(){return cursor;}
    long        get_foreg_id(){return foreg_id;}
    long        get_backg_id(){return backg_id;}
    void*       inq_state(p_index prop_ind);
    void*       inq_state(node* at_node, p_index prop_ind);
    draw_mark*  get_marks(){return first_mark;}
    vpftbl*     get_ftbl(){return ftbl;}

    /*traits to set data*/
    /*set_handle takes a pre-allocated bitmap, assigns it to the viewport, and
        fills in the width and height*/
    void        set_handle(void* bitmap_id);
    long        attach(format* to_format, long fx_size = 0, long fy_size = 0);
    void        detach_format();
    long        attach(v_cursor* to_cursor);
    long        attach(v_cursor* to_cursor, screen_xy set_to);
    void        detach_cursor();
    void        set_top_line(view_pos *new_top){top_line = new_top;}
    void        set_bottom_line(view_pos *new_bottom){bottom_line = new_bottom;}
    void        set_foreg_id(long new_id);
    void        set_backg_id(long new_id);
    long        set_state(p_index prop_ind, void* value);
    long        unset_state(p_index prop_ind);
    void        add_mark(draw_mark *new_mark);
    void        detach_mark(draw_mark *del_mark);
    void        set_ftbl(vpftbl* new_table){ftbl = new_table;}

    /*routines which allocate and free view positions in the viewport*/
    view_pos*   get_pos();
    void        free_pos(view_pos* to_free);
    void        free_all_pos();

    /*routines that actually draw things*/
    void        clear();
    void        clear(screen_xy ul);
    void        clear(screen_xy ul, screen_xy lr);
    void        draw(cline* first_line, screen_xy where);
    void        draw(model_pos draw_char, screen_xy where);
    void        adjust(long delta_x, long delta_y);
    void        redraw();
    void        redraw(model_pos ref_start_mpos, model_pos rend_start_mpos, view_pos *start_vpos);
    void        draw_to_bottom(screen_xy where, cline *start_line);
                                                  
    /*navigation routines*/
    view_pos*   xy_2_vpos(screen_xy where);
    cline*      xy_2_cline(screen_xy where);
    model_pos   xy_2_mpos(screen_xy where);
    view_pos*   cline_2_vpos(cline *find_line);
    screen_xy   cline_2_xy(cline *find_line);
    view_pos*   mpos_2_vpos(model_pos find_char);
    screen_xy   mpos_2_xy(model_pos find_char, view_pos *hint = 0);
    long        is_visible(phead *page);

    /*editing routines*/
    long        append_string(char* in_string, long count, model_pos after, char_set* in_char_set = 0);
    long        prepend_string(char* in_string, long count, model_pos before, char_set* in_char_set = 0);
    long        delete_char(model_pos delete_pos);
    long        ins_before_cursor(char* in_string, long count, char_set* in_char_set = 0);
    long        ins_after_cursor(char* in_string, long count, char_set* in_char_set = 0);
    long        insert_prop(node *to_node, property *add_prop);
    long        insert_prop(model_pos first_char, model_pos last_char, property *add_prop);
    long        delete_prop(property *del_prop, node *from_node);
    draw_mark*  mark_region(model_pos start_mark, model_pos end_mark);
};

class xw_view_port : public view_port {
public:
    xw_view_port();
};

#ifdef apollo
class gpr_view_port : public view_port {
public:
    gpr_view_port();
};
#endif

/*the view position class*/
class view_pos{
    view_pos        *prior,         //links to surrounding view_positions
                    *next;
    view_port       *in_viewport;   //what viewport is it in?
    long             ulx,           //upper left corner of the position
                     uly;
    cline           *this_line;     //pointer to the formatting line contained

    friend           view_port::view_port();
    friend view_pos* view_port::get_pos();
    friend      void view_port::free_pos(view_pos *to_free);
    friend      void view_port::adjust(long delta_x, long delta_y);

public:
    view_pos();
    ~view_pos();

    /*traits to access the data*/
    view_pos*   get_prior(){return prior;}
    view_pos*   get_next(){return next;}
    view_pos*   get_up();
    view_pos*   get_down();
    view_port*  get_viewport(){return in_viewport;}
    long        get_x(){return ulx;}
    long        get_y(){return uly;}
    cline*      get_cline(){return this_line;}
    model_pos   get_start();
    model_pos   get_end();

    /*traits to manipulate the data*/
    void        set_x(long new_x){ulx = new_x;}
    void        set_y(long new_y){uly = new_y;}
    void        set_line(cline *new_line){this_line = new_line;}

    /*misc. functions*/
    void        detach();   //remove from the lists
    void        attach(view_port *to_viewport){in_viewport = to_viewport;}
    long        attach(view_pos *to_pos, vp_dir dir);
    view_pos*   add(vp_dir dir);
    void        adjust(long x_delta, long y_delta);
    long        contains(model_pos find_pos)
                {return ((find_pos >= this_line->get_start()) && (find_pos <= this_line->get_end()));}
};


#endif   
// tml_objects_hxx
