/**************************************************************************************
    view_pos_traits.cxx
    code for the basic traits for the view_pos object. This object is the
    tml++ analog of the tml screen position, and is used to navigate
    through the various coordinate systems given the visual rendition in
    a viewport.

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <tml/tml_objects.hxx>
#include <tml/tml_dp_graph.hxx>

/*basic constructor for a view position object*/
view_pos::view_pos()
{
    prior = next = 0;
    in_viewport = 0;
    ulx = uly = -1;
    this_line = 0;
}

/*basic destructor for a view position object*/
view_pos::~view_pos()
{
    prior = next = 0;
    in_viewport = 0;
    ulx = uly = -1;
    this_line = 0;
}

/*find the first view position above this one on the screen*/
view_pos* view_pos::get_up()
{
    return prior;
}

/*find the first view position below this one on the screen*/
view_pos* view_pos::get_down()
{
    return next;
}

/*return the model position beginning the conceptual line contained in the view_pos*/
model_pos view_pos::get_start()
{
    model_pos return_pos;

    if (!this_line){
        return_pos.in_node = 0;
        return_pos.char_index = -1;
        }
    else
        return_pos = this_line->get_start();

    return return_pos;
}

/*return the model position ending the conceptual line contained in the view_pos*/
model_pos view_pos::get_end()
{
    model_pos return_pos;

    if (!this_line){
        return_pos.in_node = 0;
        return_pos.char_index = -1;
        }
    else
        return_pos = this_line->get_end();

    return return_pos;
}
/*disassociate a view_pos from any others it is attached to*/
void view_pos::detach()
{
    if (prior)
        prior->next = next;
    else
        if (in_viewport->get_top_line() == this)
            in_viewport->set_top_line(next);
    if (next)
        next->prior = prior;
    else
        if (in_viewport->get_bottom_line() == this)
            in_viewport->set_bottom_line(prior);
    prior = next = 0;
}
                                                                                    
/*attach a view_pos to another in the direction indicated--
   returns 0 if successful, -1 if unsuccessful*/
long view_pos::attach(view_pos *to_pos, vp_dir dir)
{
     
    if (!to_pos)
        if ((!in_viewport) || (in_viewport->get_top_line()) || (in_viewport->get_bottom_line()))
            return -1;
        else {
            in_viewport->set_top_line(this);
            in_viewport->set_bottom_line(this);
            return 0;
            }

    if (prior || next) detach();

    switch (dir){
      case vp_right:
      case vp_up:
        if (to_pos->prior)
            to_pos->prior->next = this;
        else
            if (in_viewport->get_top_line() == to_pos)
                in_viewport->set_top_line(this);
        prior = to_pos->prior;
        next = to_pos;
        to_pos->prior = this;
        break;

      case vp_left:
      case vp_down:
        if (to_pos->next)
            to_pos->next->prior = this;
        else
            if (in_viewport->get_bottom_line() == to_pos)
                in_viewport->set_bottom_line(this);
        next = to_pos->next;
        prior = to_pos;
        to_pos->next = this;
        break;

      default:
        return -1;
    }

    return 0;
}

/*add a new view_pos, attached as indicated by dir. Returns a
  pointer to the new view_pos if successful, 0 otherwise*/
view_pos* view_pos::add(vp_dir dir)
{
    register view_pos   *new_pos;
                                    
    if (!in_viewport) return 0;
    new_pos = in_viewport->get_pos();
    if (!new_pos) return 0;

    if (new_pos->attach(this, dir)) {
        in_viewport->free_pos(new_pos);
        return 0;
        }

    return new_pos;
}

                            
/*adjust the y position of a set of view_positions by delta_y*/
void view_pos::adjust(long x_delta, long y_delta)
{

    register view_pos   *hold_pos,
                        *next_line;
                                                          
    register view_port  *this_viewport;
         
    /*first, take care of the case where we have fallen off the viewport*/
    if ((uly + this_line->get_hi() + y_delta) > in_viewport->get_hi()) {
        /*make the prior line of the first off the screen the new bottom line*/
        this_viewport = in_viewport;
        this_viewport->set_bottom_line(prior);
        for (hold_pos = this; hold_pos; hold_pos = next_line){
            next_line = hold_pos->next;
            if ((hold_pos->uly + hold_pos->this_line->get_hi() + y_delta) > in_viewport->get_hi())
                this_viewport->free_pos(hold_pos);
            else
                this_viewport->set_bottom_line(hold_pos);
            }
        if (this_viewport->get_bottom_line())
            this_viewport->get_bottom_line()->next = 0;
        return;
        }

    /* adjust all lines below this one by that amount*/
    if (next)
        next->adjust(x_delta, y_delta);

    uly += y_delta;
    ulx += x_delta;

    /*finally, if the line has gone off the top of the screen, get rid of it*/
    if (uly < 0)
        in_viewport->free_pos(this);

    return;
}



